/***********************************************************************************
 * File: alg1_board.cpp
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#include "belofte.h"

#include "board.h"
#include "usercmd.h"

#include "alg1_moves.h"
#include "alg1_board.h"

// --------------------------------------------------------------------

t_alg1_board m1_board;
t_alg1_board m1_boardmapper;

// --------------------------------------------------------------------

extern board_t m_board;
extern game_t game;
extern uistate_t uistate;

extern hashkey_t hashmultiplexer[HASH_PIECELIST_SIZE][8][8];
extern int hashPieceindex[127];

// --------------------------------------------------------------------

static t_boolean b1_piecelist(char *szPieces);

// --------------------------------------------------------------------

static const int centervalues[8][8] = {
  { 0, 1, 2, 2, 2, 2, 1, 0},
  { 1, 3, 4, 5, 5, 4, 3, 1},
  { 2, 4, 6, 7, 7, 6, 4, 2},
  { 3, 5, 7, 9, 9, 7, 5, 3},
  { 3, 5, 7, 9, 9, 7, 5, 3},
  { 2, 4, 6, 7, 7, 6, 4, 2},
  { 1, 3, 4, 5, 5, 4, 3, 1},
  { 0, 1, 2, 3, 3, 2, 1, 0}
};

t_score m1_centervalues[BOARDARRAY_SIZE];

/**
 * copy 8x8 board into 144 array
 * and fill from user interface into engine data
 * called once on game init or when doing undo move
 */
void b1_init(void) {
  int iRow, iColumn, i;
  posid iKingPos;

  // store flags for begin position, used in static pos evaluation
  // TODO: correct because we will not always start at begin
  for (i = 0; i < 8; i++) {
    m1_board.m_whitebegin[i] = 1;
    m1_board.m_blackbegin[i] = 1;
  }

  // copy from ui board
  for (iRow = 0; iRow < 8; iRow++) {
    for (iColumn = 0; iColumn < 8; iColumn++) {
      m1_board.m_fields[B1_64TO144(iColumn, iRow)] = m_board.m_fields[iColumn][iRow];
    }
  }

  // speedup evaluation later on
  for (iKingPos = STARTPOS; iKingPos < ENDPOS; iKingPos++) {
    if (m1_board.m_fields[iKingPos] == PIECE_KING) {
      m1_board.m_whiteking = iKingPos;
      break;
    }
  }

  for (iKingPos = STARTPOS; iKingPos < ENDPOS; iKingPos++) {
    if (m1_board.m_fields[iKingPos] == BLACK_KING) {
      m1_board.m_blackking = iKingPos;
      break;
    }
  }

  // set castle flag to allowed
  m1_board.m_whitecastle[CASTLE_LONG] = myFALSE;
  m1_board.m_whitecastle[CASTLE_SHORT] = myFALSE;
  m1_board.m_blackcastle[CASTLE_LONG] = myFALSE;
  m1_board.m_blackcastle[CASTLE_SHORT] = myFALSE;

  // now we will base ourselves on the positions of the pieces
  if ((m1_board.m_whiteking == KINGWHITESTART) && (m1_board.m_fields[26] == PIECE_ROOK)) {
    m1_board.m_whitecastle[CASTLE_LONG] = myTRUE;
  }
  if ((m1_board.m_whiteking == KINGWHITESTART) && (m1_board.m_fields[33] == PIECE_ROOK)) {
    m1_board.m_whitecastle[CASTLE_SHORT] = myTRUE;
  }
  if ((m1_board.m_blackking == KINGBLACKSTART) && (m1_board.m_fields[110] == BLACK_ROOK)) {
    m1_board.m_blackcastle[CASTLE_LONG] = myTRUE;
  }
  if ((m1_board.m_blackking == KINGBLACKSTART) && (m1_board.m_fields[117] == BLACK_ROOK)) {
    m1_board.m_blackcastle[CASTLE_SHORT] = myTRUE;
  }

  // now we correct with the rook positions from the fen string
  if (m_board.m_piecemoved[COLOUR_WHITE][0]) {
    m1_board.m_whitecastle[CASTLE_LONG] = myFALSE;
  }
  if (m_board.m_piecemoved[COLOUR_WHITE][1]) {
    m1_board.m_whitecastle[CASTLE_SHORT] = myFALSE;
  }
  if (m_board.m_piecemoved[COLOUR_BLACK][0]) {
    m1_board.m_blackcastle[CASTLE_LONG] = myFALSE;
  }
  if (m_board.m_piecemoved[COLOUR_BLACK][1]) {
    m1_board.m_blackcastle[CASTLE_SHORT] = myFALSE;
  }

  // to keep scores, castling allowed or done increases the score
  // initially, castling is not done
  m1_board.m_whitecastledone = myFALSE;
  m1_board.m_blackcastledone = myFALSE;

  // carry over ep data from ui board
  m1_board.m_epfield = M1_LIJN(m_board.m_epfield) + 2 + (M1_RIJ(m_board.m_epfield) + 2)*12;

  m1_board.m_rule50 = m_board.m_rule50;
  m1_board.m_hash = m_board.m_hash;
}

void b1_constructor(void) {
}

/**
 * Fill in array for faster calculation
 */
void b1_delayedconstructor(void) {
  int iRow, iColumn;
  for (iColumn = 0; iColumn < 8; iColumn++)
    for (iRow = 0; iRow < 8; iRow++) {
      // inversion of column/row is not important as values are symmetric
      m1_centervalues[B1_64TO144(iColumn, iRow)] = centervalues[iColumn][iRow] / 100.0;
    }
}

/**
 * Calculate hash based on board position
 * @return hash of current board
 */
hashkey_t b1_calchash(void) {
  int iRow, iColumn;
  int iCountPieces = -2;
  hashkey_t hash = HASH_INIT;
  for (iColumn = 0; iColumn < 8; iColumn++)
    for (iRow = 0; iRow < 8; iRow++) {
      if (m1_board.m_fields[B1_64TO144(iColumn, iRow)] != FIELD_EMPTY) {
        hash = hash ^ hashmultiplexer[hashPieceindex[m1_board.m_fields[B1_64TO144(iColumn, iRow)]]]
                [iColumn][iRow];
        iCountPieces++;
      }
    }

  if (m1_board.m_whitecastle[CASTLE_SHORT] == myTRUE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_WHITECASTLE_S];
  if (m1_board.m_whitecastle[CASTLE_LONG] == myTRUE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_WHITECASTLE_L];
  if (m1_board.m_blackcastle[CASTLE_SHORT] == myTRUE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_BLACKCASTLE_S];
  if (m1_board.m_blackcastle[CASTLE_LONG] == myTRUE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_BLACKCASTLE_L];
  if (m1_board.m_epfield != 0)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_EPCASE];

  // add higher 8 bits
  if (m1_board.m_sidetomove == COLOUR_BLACK)
    hash = hash | 0x0100000000000000ULL;

  hash = hash | (((hashkey_t) (30 - iCountPieces)) << 57);

  return hash;
}

/**
 * perfrom the move on the board
 * and update board statistics
 */
void b1_apply_move(const int iCurPly, const posid iFrom, const posid iTo,
        const t_promotion cFlag) {
  uistate.searchnodes++;
  if (iCurPly & 1) {
    // curply is odd for white
    b1_apply_movewhite(iFrom, iTo, cFlag);
  } else {
    b1_apply_moveblack(iFrom, iTo, cFlag);
  }
}

/**
 * pefrom the move no the board
 * and update board statistics
 */
void b1_apply_movewhite(const posid iFrom, const posid iTo, const t_promotion cFlag) {
  // handling for ep
  m1_board.m_epfield = 0;
  if (m1_board.m_fields[iFrom] == PIECE_PAWN) {
    if ((iTo - iFrom) == 24) {
      // set epfield for next move from this position
      m1_board.m_epfield = iTo - 12;
    } else {
      // test if this is an ep move (not a double and no real capture)
      if ((m1_board.m_fields[iTo] == FIELD_EMPTY) && ((iTo - iFrom) != 12)) {
        // clear out captured piece
        m1_board.m_fields[iTo - 12] = FIELD_EMPTY;
      }
    }
    // clear out 50 moves rule in case of pawn move
    m1_board.m_rule50 = 0;
  } else {
    m1_board.m_rule50++;
  }

  // clear out 50 moves rule in case of capture
  if (m1_board.m_fields[iTo] != FIELD_EMPTY) m1_board.m_rule50 = 0;

  // apply move
  if (cFlag != PROMOTION_NONE) {
    // promotion
    m1_board.m_fields[iTo] = (t_piece) toupper(cFlag);
  } else {
    m1_board.m_fields[iTo] = m1_board.m_fields[iFrom];
  }
  m1_board.m_fields[iFrom] = FIELD_EMPTY;

  // update king castling flags and do castling
  if (iFrom == m1_board.m_whiteking) {
    m1_board.m_whiteking = iTo;
    if ((iTo - iFrom) == 2) {
      // short castle
      m1_board.m_fields[iTo - 1] = m1_board.m_fields[iTo + 1];
      m1_board.m_fields[iTo + 1] = FIELD_EMPTY;
      // score flag
      m1_board.m_whitecastledone = myTRUE;
      // begin pos for rook cleared
      m1_board.m_whitebegin[COL144TO64(iTo + 1)] = 0;
    } else if ((iTo - iFrom) == -2) {
      // long castle
      m1_board.m_fields[iTo + 1] = m1_board.m_fields[iTo - 2];
      m1_board.m_fields[iTo - 2] = FIELD_EMPTY;
      // score flag
      m1_board.m_whitecastledone = myTRUE;
      // begin pos for rook cleared
      m1_board.m_whitebegin[COL144TO64(iTo - 2)] = 0;
    }
    m1_board.m_whitecastle[CASTLE_LONG] = myFALSE;
    m1_board.m_whitecastle[CASTLE_SHORT] = myFALSE;
  }

  // clear begin flag
  if (ROW144TO64(iFrom) == 0) {
    m1_board.m_whitebegin[COL144TO64(iFrom)] = 0;

    if (COL144TO64(iFrom) == 0) m1_board.m_whitecastle[CASTLE_LONG] = myFALSE;
    if (COL144TO64(iFrom) == 7) m1_board.m_whitecastle[CASTLE_SHORT] = myFALSE;
  }

  // capture, so clear other flag
  if (ROW144TO64(iTo) == 7) {
    m1_board.m_blackbegin[COL144TO64(iTo)] = 0;

    if (COL144TO64(iFrom) == 0) m1_board.m_blackcastle[CASTLE_LONG] = myFALSE;
    if (COL144TO64(iFrom) == 7) m1_board.m_blackcastle[CASTLE_SHORT] = myFALSE;
  }

  m1_board.m_hash = b1_calchash();
}

// --------------------------------------------------------------------

void b1_apply_moveblack(const posid iFrom, const posid iTo, const t_promotion cFlag) {
  // handling for ep
  m1_board.m_epfield = 0;
  if (m1_board.m_fields[iFrom] == BLACK_PAWN) {
    if ((iFrom - iTo) == 24) {
      // set epfield for next move from this position
      m1_board.m_epfield = iTo + 12;
    } else {
      // test if this is an ep move (not a double and no real capture)
      if ((m1_board.m_fields[iTo] == FIELD_EMPTY) && ((iFrom - iTo) != 12)) {
        // clear out captured piece
        m1_board.m_fields[iTo + 12] = FIELD_EMPTY;
      }
    }
    // clear out 50 moves rule in case of pawn move
    m1_board.m_rule50 = 0;
  } else {
    m1_board.m_rule50++;
  }

  // clear out 50 moves rule in case of capture
  if (m1_board.m_fields[iTo] != FIELD_EMPTY) m1_board.m_rule50 = 0;

  // apply move
  if (cFlag != PROMOTION_NONE) {
    // promotion
    m1_board.m_fields[iTo] = (t_piece) tolower(cFlag);
  } else {
    m1_board.m_fields[iTo] = m1_board.m_fields[iFrom];
  }
  m1_board.m_fields[iFrom] = FIELD_EMPTY;

  // update king castling flags and do castling
  if (iFrom == m1_board.m_blackking) {
    m1_board.m_blackking = iTo;
    if ((iTo - iFrom) == 2) {
      // short castle
      m1_board.m_fields[iTo - 1] = m1_board.m_fields[iTo + 1];
      m1_board.m_fields[iTo + 1] = FIELD_EMPTY;
      // score flag
      m1_board.m_blackcastledone = myTRUE;
      // begin pos for rook cleared
      m1_board.m_blackbegin[COL144TO64(iTo + 1)] = 0;
    } else if ((iTo - iFrom) == -2) {
      // long castle
      m1_board.m_fields[iTo + 1] = m1_board.m_fields[iTo - 2];
      m1_board.m_fields[iTo - 2] = FIELD_EMPTY;
      // score flag
      m1_board.m_blackcastledone = myTRUE;
      // begin pos for rook cleared
      m1_board.m_blackbegin[COL144TO64(iTo - 2)] = 0;
    }
    m1_board.m_blackcastle[CASTLE_LONG] = myFALSE;
    m1_board.m_blackcastle[CASTLE_SHORT] = myFALSE;
  }

  // clear begin flag
  if (ROW144TO64(iFrom) == 7) {
    m1_board.m_blackbegin[COL144TO64(iFrom)] = 0;

    if (COL144TO64(iFrom) == 0) m1_board.m_blackcastle[CASTLE_LONG] = myFALSE;
    if (COL144TO64(iFrom) == 7) m1_board.m_blackcastle[CASTLE_SHORT] = myFALSE;
  }

  // capture, so clear other flag
  if (ROW144TO64(iTo) == 0) {
    m1_board.m_whitebegin[COL144TO64(iTo)] = 0;

    if (COL144TO64(iFrom) == 0) m1_board.m_whitecastle[CASTLE_LONG] = myFALSE;
    if (COL144TO64(iFrom) == 7) m1_board.m_whitecastle[CASTLE_SHORT] = myFALSE;
  }

  m1_board.m_hash = b1_calchash();
}

/**
 * Evaluate if position is theoretic draw
 * @return myTRUE if theoretic draw
 */
t_boolean b1_theoreticdraw(void) {
  char szPieces[MAX_PIECES_ON_BOARD];

  if (b1_piecelist(szPieces) == myTRUE) return myFALSE;

  if (strcmp(szPieces, "KNK") == 0) return myTRUE;
  if (strcmp(szPieces, "KKN") == 0) return myTRUE;
  if (strcmp(szPieces, "KBK") == 0) return myTRUE;
  if (strcmp(szPieces, "KKB") == 0) return myTRUE;
  if (strcmp(szPieces, "KNNK") == 0) return myTRUE;
  if (strcmp(szPieces, "KKNN") == 0) return myTRUE;
  if (strcmp(szPieces, "KNKN") == 0) return myTRUE;
  if (strcmp(szPieces, "KNKB") == 0) return myTRUE;
  if (strcmp(szPieces, "KBKN") == 0) return myTRUE;
  if (strcmp(szPieces, "KBKB") == 0) return myTRUE;

  return myTRUE;
}

bool b1_draw50moves(const int nMovesAlert) {
  return (m1_board.m_rule50 > (nMovesAlert * 2) - 1);
}

/**
 * Fill in piecelist in normalized manner
 * @param szPieces to store result string
 * @return myTRUE in case of sufficient material to win (piecelist incomplete)
 */
static t_boolean b1_piecelist(char *szPieces) {
  int iPos;
  t_piece cPiece;
  int iPieces = 0;

  char szPieceWN[MAX_PIECES_ON_BOARD] = "";
  char szPieceBN[MAX_PIECES_ON_BOARD] = "";
  char szPieceWB[MAX_PIECES_ON_BOARD] = "";
  char szPieceBB[MAX_PIECES_ON_BOARD] = "";

  strcpy(szPieces, "K");

  // calculate piece evaluation
  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    cPiece = m1_board.m_fields[iPos];
    switch (cPiece) {
      case PIECE_QUEEN: case BLACK_QUEEN:
      case PIECE_ROOK: case BLACK_ROOK:
      case PIECE_PAWN: case BLACK_PAWN:
        return myTRUE;
      case PIECE_KNIGHT:
        strcat(szPieceWN, "N");
        iPieces++;
        break;
      case BLACK_KNIGHT:
        strcat(szPieceBN, "N");
        iPieces++;
        break;
      case PIECE_BISHOP:
        strcat(szPieceWB, "B");
        iPieces++;
        break;
      case BLACK_BISHOP:
        strcat(szPieceBB, "B");
        iPieces++;
        break;
        // added to avoid compile warning, -wall
      case PIECE_KING: case BLACK_KING:
      case FIELD_EMPTY: case FIELD_BORDER:
        break;
    }
  }

  if (iPieces > 3) return myTRUE;

  strcat(szPieces, szPieceWN);
  strcat(szPieces, szPieceWB);
  strcat(szPieces, "K");
  strcat(szPieces, szPieceBN);
  strcat(szPieces, szPieceBN);

  return myFALSE;
}
// eof
