/***********************************************************************************
 * File: alg1_book.cpp
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#include "belofte.h"

#include "board.h"
#include "search.h"
#include "usercmd.h"
#include "YUtils.h"

#include "alg1_moves.h"
#include "alg1_book.h"
#include "alg1_search.h"

// --------------------------------------------------------------------

extern game_t game;
extern uimode_t uimode;

// --------------------------------------------------------------------

int busy_book_creating;
struct st_book book;

// --------------------------------------------------------------------

static void book_readheader(FILE *fpPGN, const char *szFirstLine, int *iResult);
static int book_addmoves(char *szLine, const int iResult);
static int book_addsinglemove(char *szMove, const int iResult);
static void book_addposition(const char *szFen, char *szMove, const int iResult);
static void book_addpositionnormalized(const char *szFen, char *szMove, const int iResult);
static void book_FenNormalize(char *szDest, const char *szSrc);

#if defined(FEATURE_FENINVERT)
static void book_FenInvert(char *szDest, const char *szSrc);
#endif

/**
 * load opening book
 * it will do some checking to verify that the binary book is in the correct
 * binary format
 */
void book_load(const char *szBinFileName) {
  FILE *fpBook;
  char szBuffer[255];
  size_t nRead;

  fpBook = fopen(szBinFileName, "rb");
  if (fpBook != NULL) {
    // we can read the binary version of the book, so do this because it is faster
    nRead = fread(&book, sizeof (book), 1, fpBook);
    fclose(fpBook);
    if (!nRead) return;

    if (book.szBookID != BOOKID) {
      snprintf(szBuffer, sizeof (szBuffer),
              "Opening %s in invalid endian format!", szBinFileName);
      cmdresp_snd(szBuffer);
    } else if (book.iBookVersion != BOOKVERSION) {
      snprintf(szBuffer, sizeof (szBuffer),
              "Opening %s version not supported!", szBinFileName);
      cmdresp_snd(szBuffer);
    } else if (book.iBookSize != SIZE_BOOK) {
      snprintf(szBuffer, sizeof (szBuffer),
              "Opening %s different sizes not supported!", szBinFileName);
      cmdresp_snd(szBuffer);
    } else {
      snprintf(szBuffer, sizeof (szBuffer),
              "Opening book %s loaded.", szBinFileName);
      cmdresp_snd_info(szBuffer);
      return;
    }
    // clear in memory book when something wrong found
    memset(&book, 0, sizeof (struct st_book));
  }
}

/**
 * save in memory representation of book into a binary file
 * will be called upon book creation or when ascii version of book is present and binary not
 */
void book_save(const char *szBinFileName) {
  char szBuffer[MAX_DATANAME_LENGTH];
  FILE *fpBook;

  // following parametes are used for loading of the binary book
  // if there is a mismatch, the binary book will not be loaded
  book.szBookID = BOOKID;
  book.iBookVersion = BOOKVERSION;
  book.iBookSize = SIZE_BOOK;

  if (strlen(szBinFileName) > 0) {
    fpBook = fopen(szBinFileName, "wb");
    if (!fpBook) {
      snprintf(szBuffer, sizeof (szBuffer),
              "Could not create book: %s\n", szBinFileName);
      cmdresp_snd(szBuffer);
      return;
    }
    fwrite(&book, sizeof (book), 1, fpBook);
    fclose(fpBook);
  }

  snprintf(szBuffer, sizeof (szBuffer), "Book %s saved", szBinFileName);
  cmdresp_snd_info(szBuffer);
}

/**
 * create a book from a pgn file
 */
void book_create(const char *szInputFileName, const char *szBinFileName) {
  FILE *fpPGN;
  int bEnd;
  char szLine[MAX_FEN_LENGTH];
  char szBuffer[MAX_DATANAME_LENGTH];
  int iResult = 0;

  busy_book_creating = 1;
  cmd_new("", "nobd");

  fpPGN = fopen(szInputFileName, "r");
  if (fpPGN == NULL) {
    snprintf(szBuffer, sizeof (szBuffer),
            "PGN file not found: %s\n", szInputFileName);
    cmdresp_snd(szBuffer);
    busy_book_creating = 0;
    return;
  }

  while (!feof(fpPGN)) {
    bEnd = file_readline(szLine, fpPGN, sizeof (szLine));
    if (bEnd) break;
    if (strlen(szLine) && (szLine[0] == '[')) {
      book_readheader(fpPGN, szLine, &iResult);
      if (iResult != BOOKRESULT_UNKNOWN) {
        // do not store moves when there is no result known
        // because the flag is used to store intermediate moves
        cmd_new("", "nobd");
        cmd_force("", "");
      }
    } else if (strlen(szLine) && (szLine[0] == '%')) {
      // line of comments, skip
    } else if (iResult != BOOKRESULT_UNKNOWN) {
      // there is a result, it is worth reading the game and storing it
      if (book_addmoves(szLine, iResult) == MOVERESULT_ERROR) {
        // if there is any invalid move, then abort adding this game
        iResult = BOOKRESULT_UNKNOWN;
      }
    }
    // if (bEnd == -1) break;
  }
  fclose(fpPGN);

  snprintf(szBuffer, sizeof (szBuffer), "Book created from %s", szInputFileName);
  cmdresp_snd_info(szBuffer);

  book_save(szBinFileName);

  busy_book_creating = 0;

  cmd_new("", "nobd");
}

/**
 * read the game header until a result tag is found, if none is found
 * do not use this game and skip to the next game
 */
static void book_readheader(FILE *fpPGN, const char *szFirstLine, int *iResult) {
  char szLine[MAX_FEN_LENGTH];
  *iResult = BOOKRESULT_UNKNOWN;

  strncpy(szLine, szFirstLine, sizeof (szLine));
  do {
    if (strlen(szLine) && (szLine[0] == '%')) {
      // line of comments, skip
    } else if (strlen(szLine) && (szLine[0] == '[')) {
      // line starts with header
      if (strncmp(szLine, "[Result ", 8) == 0) {
        if (strncmp(szLine, "[Result \"1/2-1/2", 16) == 0) {
          *iResult = BOOKRESULT_DRAW;
        } else if (strncmp(szLine, "[Result \"1-0", 12) == 0) {
          *iResult = BOOKRESULT_WHITEWON;
        } else if (strncmp(szLine, "[Result \"0-1", 12) == 0) {
          *iResult = BOOKRESULT_BLACKWON;
        }
      }
      cmdresp_snd(szLine);
    } else {
      // no header line anymore
      break;
    }
    file_readline(szLine, fpPGN, sizeof (szLine));
    ;
  } while (!feof(fpPGN));

}

/**
 * add a line of moves, do some line interpretation
 */
static int book_addmoves(char *szLine, const int iResult) {
  char szMove[MAX_FEN_LENGTH];
  char *szBuffer = szLine;

  while (EOF != sscanf(szBuffer, "%[^ \n\r]", szMove)) {
    if (strlen(szMove) > 0) {
      if (book_addsinglemove(szMove, iResult) == MOVERESULT_ERROR)
        // if there is an error reading the move, skip to next game
        return MOVERESULT_ERROR;
    }
    szBuffer += strlen(szMove) + 1;
    szMove[0] = '\0'; // needed because multiple terminators cause the szMove not to be cleared
    if (strlen(szLine) <= (size_t) (szBuffer - szLine))
      break;
  }

  return MOVERESULT_OK;
}

/**
 * add a single move to the book
 */
static int book_addsinglemove(char *szMove, const int iResult) {
  char szBuffer[MAX_DATANAME_LENGTH];
  char szCoordMove[10];

  // we do not add score indications to the book
  // the score is carried over from the header
  if (strcmp(szMove, "1-0") == 0) return 0;
  if (strcmp(szMove, "0-1") == 0) return 0;
  if (strcmp(szMove, "1/2-1/2") == 0) return 0;

  // move numbers are skipped
  if (szMove[strlen(szMove) - 1] == '.') return 0;

  // here we will correct for having a dot in the move, in case there is no
  // space in between the move number and the move
  // this code calls on the implementation of the function calling this
  // we know that the buffer will be moved only with the lenght of this move found
  if (szMove[1] == '.') {
    szMove[1] = '\0';
    return 0;
  }
  if (szMove[2] == '.') {
    szMove[2] = '\0';
    return 0;
  }

  if (game.curply > MAX_GAME_LENGTH - 1) {
    snprintf(szBuffer, sizeof (szBuffer),
            "Cannot store more than %d moves in a game. "
            "Skipping remaining moves.", MAX_GAME_LENGTH - 1);
    cmdresp_snd(szBuffer);
    return -1;
  }

  if (game.curply > BOOK_DEPTH) {
    snprintf(szBuffer, sizeof (szBuffer),
            "We only store %d half-moves. Skipping remaining moves.", BOOK_DEPTH);
    cmdresp_snd_debug(szBuffer);
    return -1;
  }

  if (board_sanapplymove(szMove, szCoordMove) == MOVERESULT_NIL) {
    // invalid move, so abort storing this move
    snprintf(szBuffer, sizeof (szBuffer), "Invalid move %s, skipping game", szMove);
    cmdresp_snd(szBuffer);
    return MOVERESULT_ERROR;
  }

  // move is valid, temporarily undo it so the position can be added correctly
  // TODO: check if this is realy needed, maybe we could just retrieve the previous move
  board_unapplymove();
  book_addposition(game.m_boardhistory[game.curply].szFEN, szCoordMove, iResult);
  board_applymove(szCoordMove);

  return 1;
}

/**
 * add a single position to the book, some parsing to be done
 */
static void book_addposition(const char *szFen, char *szCoordMove, const int iResult) {
  char szNormalizedEpd[MAX_FEN_LENGTH];

  book_FenNormalize(szNormalizedEpd, szFen);
  book_addpositionnormalized(szNormalizedEpd, szCoordMove, iResult);
}

/**
 * create the epd string based on the standard Fen string
 * the normalized is needed because we do not need some additional flags
 * at the end for specifying book boves, we will thus delete them
 */
static void book_FenNormalize(char *szDest, const char *szSrc) {
  int iCount = 0;
  unsigned int i;

  szDest[0] = '\0';
  strncat(szDest, szSrc, MAX_FEN_LENGTH - 1);

  // correct by removing two last ply indicators in the move
  for (i = 0; i < strlen(szSrc); i++) {
    if (szDest[i] == ' ') {
      iCount++;
      if (iCount == 3) {
        // normally, this code was 4, meaning after the last ep
        // possible move.  Now we reduce it to 3 to increase
        // the hits in case of inversion of colours
        szDest[i] = '\0';
        break;
      }
    }
  }

  if (iCount != 3) {
    // invalid FEN string
    szDest[0] = '\0';
  }
}

#if defined(FEATURE_FENINVERT)

/**
 * adapt the Fen in such a way that the colours are inverted
 * as if a white player would play black
 */
static void book_FenInvert(char *szDest, const char *szSrc) {
  char *szTmpFen; // working array
  char szTmpBuffer[MAX_FEN_LENGTH];
  void *pMalloc;
  char *aMainParts[3];
  char *aSubParts[8];
  char szDummy[1] = "";
  unsigned int iToken = 0;
  const char s1[2] = " ";
  const char s2[2] = "/";

  szDest[0] = '\0';

  // validate input string
  if ((szSrc == NULL) || (strlen(szSrc) == 0)) {
    // do nothing if the original string is empty
    return;
  }

  pMalloc = malloc(MAX_FEN_LENGTH);
  szTmpFen = pMalloc;

  szTmpFen[0] = '\0';
  strncat(szTmpFen, szSrc, MAX_FEN_LENGTH - 1);

  aMainParts[iToken] = strtok(szTmpFen, s1);
  if (aMainParts[iToken] == NULL) {
    // do nothing if the original string is not valid
    free(pMalloc);
    return;
  }

  for (iToken = 1; iToken < 3; iToken++) {
    aMainParts[iToken] = strtok(NULL, s1);
    if (aMainParts[iToken] == NULL) {
      // do nothing if the string is not valid, there must be at least 3 parts
      free(pMalloc);
      return;
    }
  }

  // now the aMainParts array points to all but the first array
  /// we will now construct the first array

  iToken = 0;
  szTmpBuffer[0] = '\0';
  strncat(szTmpBuffer, aMainParts[0], MAX_FEN_LENGTH - 1);
  aSubParts[iToken] = strtok(szTmpBuffer, s2);
  if (aSubParts[iToken] == NULL) {
    // do nothing if the original string is not valid
    free(pMalloc);
    return;
  }

  for (iToken = 1; iToken < 8; iToken++) {
    aSubParts[iToken] = strtok(NULL, s2);
    if (aSubParts[iToken] == NULL) {
      break;
    }
  }

  for (iToken = 0; iToken < 8; iToken++) {
    if (aSubParts[iToken] == NULL) {
      aSubParts[iToken] = szDummy;
    }
  }

  // now we have aSubParts pointing to all rows
  // and aMainParts pointing to the other parts of the Fen string

  for (iToken = 7; iToken > 0; iToken--) {
    strcat(szDest, aSubParts[iToken]);
    strcat(szDest, "/");
  }
  strcat(szDest, aSubParts[0]);

  // switch colours by inverting casing
  for (iToken = 0; iToken < strlen(szDest); iToken++) {
    if (isalpha(szDest[iToken])) szDest[iToken] = (szDest[iToken]) ^ (LOWERCASEBIT);
  }

  // switch side to move
  if (aMainParts[1][0] == 'w')
    strcat(szDest, " b ");
  else
    strcat(szDest, " w ");

  // castling flags switched - this could be done differently, but as the number of combinations is limited
  // still doable this way
  if (strcmp(aMainParts[2], "KQk") == 0)
    strcat(szDest, "Kkq");
  else if (strcmp(aMainParts[2], "KQq") == 0)
    strcat(szDest, "Qkq");
  else if (strcmp(aMainParts[2], "KQ") == 0)
    strcat(szDest, "kq");
  else if (strcmp(aMainParts[2], "Kkq") == 0)
    strcat(szDest, "KQk");
  else if (strcmp(aMainParts[2], "Kq") == 0)
    strcat(szDest, "Qk");
  else if (strcmp(aMainParts[2], "K") == 0)
    strcat(szDest, "k");
  else if (strcmp(aMainParts[2], "Qkq") == 0)
    strcat(szDest, "KQq");
  else if (strcmp(aMainParts[2], "Qk") == 0)
    strcat(szDest, "KQ");
  else if (strcmp(aMainParts[2], "Q") == 0)
    strcat(szDest, "q");
  else if (strcmp(aMainParts[2], "kq") == 0)
    strcat(szDest, "KQ");
  else if (strcmp(aMainParts[2], "k") == 0)
    strcat(szDest, "K");
  else if (strcmp(aMainParts[2], "q") == 0)
    strcat(szDest, "Q");
  else {
    strcat(szDest, aMainParts[2]);
    //strcat(szDest, "");
  }

  /*
  // add ep case
  strcat(szDest,aMainParts[3]);
  if (strcmp(aMainParts[3],"-") != 0) {
      szInvertedFen[strlen(szDest)-1] = '9' - aMainParts[3][1] + '0';
  }
   */
  free(pMalloc);
}
#endif

/**
 * Add a position to the book
 */
static void book_addpositionnormalized(const char *szNormalizedEpd, char *szMove, const int iResult) {
  int iAdded = myFALSE;
  int iMoveAlreadyInBook = myFALSE;
  int iBookPositions = 0;
  int iMovesInPosition = 0;

  while ((iBookPositions < book.liBookUsedPositions) && (!iMoveAlreadyInBook)) {
    if (strcmp(book.pos[iBookPositions].szFen, szNormalizedEpd) == 0) {
      // the FEN position is already in the book, look for adding the move
      iAdded = myTRUE;
      iMovesInPosition = 0;
      while (iMovesInPosition < book.pos[iBookPositions].iMoves) {
        if (strcmp(book.pos[iBookPositions].m_move[iMovesInPosition].szMove, szMove) == 0) {
          // the move is already in the book, so no need to add it again
          iMoveAlreadyInBook = myTRUE; // set flag for breaking out outer loop and not adding it
          book.pos[iBookPositions].m_move[iMovesInPosition].nOccurence++;
          if (iResult == BOOKRESULT_WHITEWON)
            book.pos[iBookPositions].m_move[iMovesInPosition].nWhiteWin++;
          else if (iResult == BOOKRESULT_BLACKWON)
            book.pos[iBookPositions].m_move[iMovesInPosition].nBlackWin++;
          break; // break out inner loop
        }
        iMovesInPosition++;
      }
      if (!iMoveAlreadyInBook) {
        if (book.pos[iBookPositions].iMoves < SIZE_BRANCH - 1) {
          // FEN position already in book, but move itself not
          iMovesInPosition = book.pos[iBookPositions].iMoves;
          strncpy(book.pos[iBookPositions].m_move[iMovesInPosition].szMove, szMove, MAX_SANMOVE_LENGTH);
          book.pos[iBookPositions].m_move[iMovesInPosition].nOccurence = 1;
          if (iResult == BOOKRESULT_WHITEWON) {
            book.pos[iBookPositions].m_move[iMovesInPosition].nWhiteWin = 1;
            book.pos[iBookPositions].m_move[iMovesInPosition].nBlackWin = 0;
          } else if (iResult == BOOKRESULT_BLACKWON) {
            book.pos[iBookPositions].m_move[iMovesInPosition].nWhiteWin = 0;
            book.pos[iBookPositions].m_move[iMovesInPosition].nBlackWin = 1;
          }
          book.pos[iBookPositions].iMoves++;
          book.iLeaves++;
          break;
        }
      }
    }
    iBookPositions++;
  }

  if (!iAdded) {
    if (book.liBookUsedPositions < SIZE_BOOK - 1) {
      strncpy(book.pos[book.liBookUsedPositions].szFen, szNormalizedEpd, MAX_FEN_LENGTH);
      book.pos[book.liBookUsedPositions].iMoves = 1;
      strncpy(book.pos[book.liBookUsedPositions].m_move[0].szMove, szMove, MAX_SANMOVE_LENGTH);
      book.pos[book.liBookUsedPositions].m_move[0].nOccurence = 1;
      if (iResult == BOOKRESULT_WHITEWON) {
        book.pos[iBookPositions].m_move[iMovesInPosition].nWhiteWin = 1;
        book.pos[iBookPositions].m_move[iMovesInPosition].nBlackWin = 0;
      } else if (iResult == BOOKRESULT_BLACKWON) {
        book.pos[iBookPositions].m_move[iMovesInPosition].nWhiteWin = 0;
        book.pos[iBookPositions].m_move[iMovesInPosition].nBlackWin = 1;
      }
      book.liBookUsedPositions++;
      book.iLeaves++;
    }
  }
}

/**
 * select a book move in current position
 * return 0 if not a book move, 1 if my book, 2 if generic book
 * it will also check for inverted colours
 */
int book_move(const char *szFen, char *szMove) {
  int i;
  int j;
  int iMoves;
  int iResult = 0;
#if defined(FEATURE_FENINVERT)
  char szInvertedFen[MAX_FEN_LENGTH];
#endif
  char szNormalizedEpd[MAX_FEN_LENGTH];
  char szBuffer[MAX_DATANAME_LENGTH];
  int iScoreMinimum[8] = {70, 60, 55, 50, 45, 40, 35, 30};

  book_FenNormalize(szNormalizedEpd, szFen);
#if defined(FEATURE_FENINVERT)
  book_FenInvert(szInvertedFen, (const char *) szNormalizedEpd);
#endif

  for (j = 0; j < 8; j++) {
    // make sure we only select good scores
    if ((iResult == 0) && (book.liBookUsedPositions > 0)) {
      for (i = 0; i < book.liBookUsedPositions; i++) {
        if (strcmp(book.pos[i].szFen, szNormalizedEpd) == 0) {
          // there is a book move, get one of the possible moves from the list
          iResult = book_selectbestmove(i, szMove, iScoreMinimum[j]);
          break;
#if defined(FEATURE_FENINVERT)
        } else if (strcmp(book.pos[i].szFen, szInvertedFen) == 0) {
          // there is a book move, get one of the possible moves from the list
          iResult = book_selectbestmove(i, szMove, iScoreMinimum[j]);
          break;
#endif
        }
      }
    }
  }

  // verify that the book move can be played
  if (iResult == 2) {
    iResult = 0; // reset back to zero, we cannot play unless we find it
    iMoves = search_fill_movelist((t_colour) (game.curply & 1));
    for (i = 0; i < iMoves; i++) {
      alg1_search_printablemove(szBuffer, i);
      if (strcmp(szMove, szBuffer) == 0) {
        // among the list of possible moves in this position, play it
        iResult = 2;
        break;
      }
    }
  }

  return iResult;
}

/**
 * select best move according to frequency played
 * the moves with the highest frequency has the highest chance to be played
 * TODO: add weighting factor according to result, the number of games won
 * can come into weight
 */

int book_selectbestmove(const int iPos, char *szMove, const int iReqScore) {
  int iSelected;
  int i;
  int iTotalOccurance = 0;
  int iWeightedTotalOccurance = 0;
  int iMoveIndex = -1;
  char szBuffer[MAX_DATANAME_LENGTH];

  szMove[0] = '\0';

  // sum total occurances, e.g. 1,5,4,1 = 11
  for (i = 0; i < book.pos[iPos].iMoves; i++)
    iTotalOccurance += book.pos[iPos].m_move[i].nOccurence;

  if (iReqScore >= 50) {
    // good scores, only take major moves into account

    // threshold, below this there is no move taken into account for the
    // high scores
    int iMinimalOccurance = iTotalOccurance / 20;

    // sum weighted total occurances, e.g. 1,5,4,1 = 11
    for (i = 0; i < book.pos[iPos].iMoves; i++)
      if (book.pos[iPos].m_move[i].nOccurence > iMinimalOccurance)
        iWeightedTotalOccurance += book.pos[iPos].m_move[i].nOccurence;

    // random one of them 0-10;
    iSelected = (int) ((double) iWeightedTotalOccurance * ((double) rand() / (double) RAND_MAX));
    // now loop through all moves and select
    for (i = 0; i < book.pos[iPos].iMoves; i++) {
      if (book.pos[iPos].m_move[i].nOccurence > iMinimalOccurance) {
        iSelected -= book.pos[iPos].m_move[i].nOccurence;
        if (iSelected <= 0) {
          iMoveIndex = i;
          break;
        }
      }
    }
  } else {
    // bad scores, allow all opening books
    // random one of them 0-10;
    iSelected = (int) ((double) iTotalOccurance * ((double) rand() / (double) RAND_MAX));

    // now loop through all moves and select
    for (i = 0; i < book.pos[iPos].iMoves; i++) {
      iSelected -= book.pos[iPos].m_move[i].nOccurence;
      if (iSelected <= 0) {
        iMoveIndex = i;
        break;
      }
    }

  }

  if (iMoveIndex != -1) {
    // make sure we blank out if the score is not ok
    int nScore;
    int iOccurance = book.pos[iPos].m_move[iMoveIndex].nOccurence;
    int iWhiteWin = book.pos[iPos].m_move[iMoveIndex].nWhiteWin;
    int iBlackWin = book.pos[iPos].m_move[iMoveIndex].nBlackWin;
    int iDraw = iOccurance - (iWhiteWin + iBlackWin);

    // curply is even for white
    if (!(game.curply & 1)) {
      nScore = (iWhiteWin * 100 + iDraw * 50) / iOccurance;
    } else {
      nScore = (iBlackWin * 100 + iDraw * 50) / iOccurance;
    }

    snprintf(szBuffer, sizeof (szBuffer),
            "Book %s - %d wh:%d bl:%d score:%d%% (min:%d%%)",
            book.pos[iPos].m_move[iMoveIndex].szMove,
            iOccurance, iWhiteWin, iBlackWin, nScore, iReqScore);
    cmdresp_snd_info(szBuffer);

    if (nScore >= iReqScore)
      strncpy(szMove, book.pos[iPos].m_move[iMoveIndex].szMove, MAX_SANMOVE_LENGTH);

  }

  return (strlen(szMove) == 0 ? 0 : 2);

}

/**
 * print moves in current posistion
 * it will also check a swap colour
 */
void book_listmoves(const char *szFen, const t_boolean bShowFen) {
  int i, j;
  char szNormalizedEpd[MAX_FEN_LENGTH];
  //char szInvertedFen[MAX_FEN_LENGTH];
  char szBuffer[MAX_DATANAME_LENGTH];
  int iPrinted;

  book_FenNormalize(szNormalizedEpd, szFen);
  //book_FenInvert(szInvertedFen, (const char *) szNormalizedEpd);

  if (bShowFen) {
    snprintf(szBuffer, sizeof (szBuffer), "Nor: %s\n"
            , szNormalizedEpd);
    cmdresp_snd(szBuffer);
  }

  if (book.liBookUsedPositions > 0) {
    iPrinted = 0;
    for (i = 0; i < book.liBookUsedPositions; i++) {
      if (strcmp(book.pos[i].szFen, szNormalizedEpd) == 0) {
        if (bShowFen) {
          snprintf(szBuffer, sizeof (szBuffer),
                  "Book moves : (out of %ld total moves and %ld positions stored)",
                  book.iLeaves, book.liBookUsedPositions);
          cmdresp_snd(szBuffer);
        }
        for (j = 0; j < book.pos[i].iMoves; j++) {
          iPrinted = 1;
          snprintf(szBuffer, sizeof (szBuffer), "%s (%d)(%d/%d)",
                  book.pos[i].m_move[j].szMove,
                  book.pos[i].m_move[j].nOccurence,
                  book.pos[i].m_move[j].nWhiteWin,
                  book.pos[i].m_move[j].nBlackWin);
          cmdresp_snd(szBuffer);
        }
      }
      /*
      if (strcmp(book.pos[i].szFen, szInvertedFen) == 0) {
        if (bShowFen) {
          snprintf(szBuffer, sizeof (szBuffer),
                  "Book moves : (out of %ld total moves and %ld positions stored) (inv)\n",
                  book.iLeaves, book.liBookUsedPositions);
          cmdresp_snd(szBuffer);
        }
        for (j = 0; j < book.pos[i].iMoves; j++) {
          iPrinted = 1;
          snprintf(szBuffer, sizeof (szBuffer), "%s (%d)(%d/%d)\n",
                  book.pos[i].m_move[j].szMove,
                  book.pos[i].m_move[j].nOccurence,
                  book.pos[i].m_move[j].nWhiteWin,
                  book.pos[i].m_move[j].nBlackWin);
          cmdresp_snd(szBuffer);
        }
      }
       */
    }
    if (!iPrinted) {
      snprintf(szBuffer, sizeof (szBuffer),
              "No matching book move found : (out of %ld total moves and %ld positions stored)\n",
              book.iLeaves, book.liBookUsedPositions);
      cmdresp_snd(szBuffer);
    }
  }

}

// --------------------------------------------------------------------
// eof
