/***********************************************************************************
 * File: alg1_book.h
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#if !defined(__ALG1_BOOK_H__)
#define __ALG1_BOOK_H__

#include "belofte.h"

/** CHBK
 * following string is represented as HEX, if the book is created on a little-endian machine,
 * it will be invalid on a big-endian machine because the byte ordering will be different
 */

#define BOOKID 0x4348424B

/** each time we alter the book size, we change the version
 */
#define BOOKVERSION 2

/**
 * the branch size is 30 bytes times the branching factor (20)
 * ( the book size is the 136 plus the branch size ) times the book size
 * a value of book-size of 700000 will occupy appoximately 512 MB
 */
#define BOOK_DEPTH      30
#define SIZE_BRANCH     20
#define SIZE_BOOK       200000

// a lot is to be gained in size by storing the move as 2 bytes (from and to)
// and limiting the counters to 24 bits (16 million positions)
// and limiting the number of branches to e.g. 10-12
// then the size of a position can be brought down to one third at least

// --------------------------------------------------------------------

struct st_book_move { // single book move entry
  char szMove[MAX_SANMOVE_LENGTH]; // the move
  int nOccurence; // number of times this is played
  int nWhiteWin; // number of times white won
  int nBlackWin; // number of times black won
  int nDraw; // number of times the score was draw
  int nMovesAfter; // the number of moves that were played after this move
  // this is total, all games together, to indicate the purity of the score
};

struct st_book_pos { // Opening book entry
  char szFen[MAX_FEN_LENGTH]; // the identifier
  int nOccurence; // number of times this is in the opening book
  int iMoves; // number of moves stored
  struct st_book_move m_move[SIZE_BRANCH]; // each move
};

struct st_book {
  long szBookID; // some magic number to identify BIG or LITTLE ENDIAN when loading binary book
  long iBookVersion; // support different versions
  long iBookSize; // control digit for the number of book positions
  char szBookDate[14]; // date of last modification in the book YYYYMMDDHHMMSS
  long liBookUsedPositions; // number of used positions in the book
  long iLeaves; // total number of leaves
  struct st_book_pos pos[SIZE_BOOK]; // array of positions
};

// --------------------------------------------------------------------

#define BOOKRESULT_UNKNOWN -1
#define BOOKRESULT_DRAW 0
#define BOOKRESULT_WHITEWON 1
#define BOOKRESULT_BLACKWON 2

// --------------------------------------------------------------------

void book_load(const char *szBinFileName);
void book_save(const char *szBinFileName);
void book_create(const char *szInputFileName, const char *szBinFileName);
int book_move(const char *szFen, char *szMove);
void book_listmoves(const char *szFen, const t_boolean bShowFen);
int book_selectbestmove(const int iPos, char *szMove, const int iReqScore);

// --------------------------------------------------------------------

#endif
