/***********************************************************************************
 * File: alg1_eval.cpp
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#include "belofte.h"

#include "board.h"
#include "search.h"

#include "alg1_moves.h"
#include "alg1_board.h"
#include "alg1_eval.h"
#include "alg1_search.h"

// --------------------------------------------------------------------

static int eval_ifposfinished(const t_boolean isWhite,
        int * iMovesWhite, int * iMovesBlack);
static t_score eval_mobilitymv(const t_boolean isWhite,
        const int iWhiteMove, const int iBlackMoves);
static t_score eval_materialwhite(void);

static int countDoublePawn(const t_boolean bEvalForWhite);
static int whiteKingProtected(void);
static int blackKingProtected(void);
static int whitePawns(void);
static int blackPawns(void);
static int onOpenDirection(const posid iPos, const int iColour, const int iDir);
static int onOpenLine(const posid iPos, const int iColour);
static int offensiveBishop(const posid iPos, t_colour iColour);
static int offensiveRook(const posid iPos, t_colour iColour);
static int onSameDir(const posid iPos, const int iDir, const t_piece cPieceToLookFor, int *iPieces);

// --------------------------------------------------------------------

extern t_alg1_board m1_board;
extern t_alg1_board m1_boardmapper;
extern t_alg1_move m1_movelist;
extern level_t level;
extern game_t game;

// --------------------------------------------------------------------

int m1_aPieceValues[FIELDVALUE_SIZE];
t_boolean isEndGame;

// --------------------------------------------------------------------
// all positions below are seen with white from bottom

static const int opening_kingvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 2, 0, -5, -5, 0, 2, 0},
  { 8, 10, 10, -6, 5, -6, 10, 10}
};

static const int opening_rookvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 1, 0, 0, 0, 0, 0, 0, 1},
  { 0, 0, 0, 12, 12, 0, 0, 0},
  { 0, 1, 3, 10, 10, 5, 1, 0}
};

static const int rookvalues[8][8] = {
  { 4, 6, 6, 6, 6, 6, 6, 4},
  { 8, 10, 10, 10, 10, 10, 10, 8},
  { 0, 6, 6, 6, 6, 6, 6, 0},
  { 0, 0, 3, 3, 3, 3, 0, 0},
  { 0, 0, 3, 3, 3, 3, 0, 0},
  { 1, 3, 3, 3, 3, 3, 3, 1},
  { 0, 3, 6, 12, 12, 6, 3, 0},
  { 0, 3, 3, 10, 10, 5, 3, 0}
};

static const int opening_knightvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 3, 0, 0, 3, 0, 0},
  { 0, 0, 0, 2, 2, 0, 0, 0},
  {-1, 0, 5, 3, 3, 5, 0, -1},
  {-1, 0, 2, 3, 3, 2, 0, -1},
  {-2, -1, -1, 0, 0, -1, -1, -2}
};

static const int knightvalues[8][8] = {
  {-4, -1, 0, 0, 0, 0, -1, -4},
  {-1, 0, 1, 1, 1, 1, 0, -1},
  { 0, 2, 3, 4, 4, 3, 2, 0},
  { 1, 2, 4, 4, 4, 4, 2, 1},
  { 1, 3, 4, 4, 4, 4, 3, 1},
  { 0, 2, 3, 3, 3, 3, 2, 0},
  {-1, 0, 2, 3, 3, 2, 0, -1},
  {-4, -1, -1, 0, 0, -1, -1, -4}
};

static const int opening_bishopvalues[8][8] = {
  {-2, 0, 0, 0, 0, 0, 0, -2},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 3, 0, 0, 0, 0, 3, 0},
  { 0, 0, 4, 4, 3, 4, 0, 0},
  { 0, 1, 1, 3, 3, 1, 1, 0},
  { 0, 3, 0, 3, 3, 0, 3, 0},
  {-2, -1, 0, 0, 0, 0, -1, -2}
};

static const int bishopvalues[8][8] = {
  {-2, 0, 0, 0, 0, 0, 0, -2},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 3, 0, 0, 0, 0, 3, 0},
  { 0, 0, 4, 4, 3, 4, 0, 0},
  { 0, 1, 1, 3, 3, 1, 1, 0},
  { 0, 3, 0, 3, 3, 0, 3, 0},
  {-2, -1, 0, 0, 0, 0, -1, -2}
};

static const int opening_queenvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 4, 0, 0, 0, 0, 0, 0},
  {-1, -1, 1, 2, 2, 0, -1, -1},
  {-2, -1, 0, 5, 0, 0, -1, -2}
};

static const int opening_pawnvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 1, 2, 2, 1, 0, 0},
  { 0, 0, 3, 6, 6, 0, 0, 0},
  { 2, 0, 2, 2, 2, 1, 1, 1},
  { 1, 1, 1, 0, 0, 2, 1, 1},
  { 0, 0, 0, 0, 0, 0, 0, 0}
};

static const int pawnvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0},
  { 0, 0, 1, 2, 2, 1, 0, 0},
  { 0, 0, 3, 6, 6, 0, 0, 0},
  { 2, 0, 2, 2, 2, 1, 1, 1},
  { 1, 1, 1, 0, 0, 2, 1, 1},
  { 0, 0, 0, 0, 0, 0, 0, 0}
};

static const int endgame_pawnvalues[8][8] = {
  { 0, 0, 0, 0, 0, 0, 0, 0},
  {50, 50, 60, 70, 70, 60, 50, 50},
  {15, 20, 25, 30, 30, 25, 20, 15},
  { 6, 10, 15, 20, 20, 15, 10, 6},
  { 2, 3, 4, 4, 4, 4, 3, 2},
  { 1, 2, -5, -5, -5, -5, 2, 1},
  {-5, -5, -5, -9, -9, -9, -5, -5},
  { 0, 0, 0, 0, 0, 0, 0, 0}
};

/*----------------------------------------------------------------------+
|      Position evaluation functions                                    |
+----------------------------------------------------------------------*/

/**
 * calculate the current search depth evaluation
 */
t_score eval_pos(const t_boolean isWhite) {
  int iMovesWhite;
  int iMovesBlack;
  t_score fMaterial;
  t_score fMobility;
  t_score fKingPos;
  t_score fPosStatic;
  t_score fPosDynamic;

  int iEvaluation = eval_ifposfinished(isWhite, &iMovesWhite, &iMovesBlack);
  if (iEvaluation != SCORE_UNDEFINED) return (t_score) iEvaluation;

  fMaterial = eval_material(isWhite);

  isEndGame = myFALSE;
  if (eval_countcaptures() > 20) isEndGame = myTRUE;

  fMobility = eval_mobilitymv(isWhite, iMovesWhite, iMovesBlack);
  fKingPos = eval_kingpos(isWhite);
  fPosStatic = eval_staticpos(isWhite);
  fPosDynamic = eval_dynamicpos(isWhite);

  return fMaterial + fMobility + fKingPos + fPosStatic + fPosDynamic;
}

/**
 * calculate the current search depth evaluation
 * return only material evaluation in case of large difference
 */
t_score eval_pos_failmaterial(const t_boolean isWhite, const t_score fMaterialCut) {
  int iMovesWhite;
  int iMovesBlack;
  t_score fMaterial;
  t_score fMobility;
  t_score fKingPos;
  t_score fPosStatic;
  t_score fPosDynamic;

  int iEvaluation = eval_ifposfinished(isWhite, &iMovesWhite, &iMovesBlack);
  if (iEvaluation != SCORE_UNDEFINED) return (t_score) iEvaluation;

  // exit if in exchange variation and loss of piece
  fMaterial = eval_material(isWhite);
  if (isWhite == myTRUE) {
    if (fMaterial > fMaterialCut + 3) return fMaterial;
  } else {
    if (fMaterial < fMaterialCut - 3) return fMaterial;
  }

  isEndGame = myFALSE;
  if (eval_countcaptures() > 20) isEndGame = myTRUE;

  fMobility = eval_mobilitymv(isWhite, iMovesWhite, iMovesBlack);
  fKingPos = eval_kingpos(isWhite);
  fPosStatic = eval_staticpos(isWhite);
  fPosDynamic = eval_dynamicpos(isWhite);

  return fMaterial + fMobility + fKingPos + fPosStatic + fPosDynamic;
}

int eval_countcaptures(void) {
  int iRow, iColumn;
  int iCountCaptures = 32;

  for (iColumn = 0; iColumn < 8; iColumn++) {
    for (iRow = 0; iRow < 8; iRow++) {
      if (m1_board.m_fields[B1_64TO144(iColumn, iRow)] != FIELD_EMPTY) {
        iCountCaptures--;
      }
    }
  }

  return iCountCaptures;
}

/**
 * Return position globals such as stalemate or draw
 * @return SCORE_UNDEFINED, +-SCORE_ABORT, +-SCORE_WON, SCORE_DRAW
 */
int eval_stale_pos(const t_boolean isWhite) {
  int iMovesWhite;
  int iMovesBlack;
  return eval_ifposfinished(isWhite, &iMovesWhite, &iMovesBlack);
}

/**
 * First part of position evaluation, position specials such as stalemate
 * or mate or theoretic draw
 * @param iMovesWhite int* to move counter for white in position
 * @param iMovesBlack int* to move counter for black in position
 * @return SCORE_UNDEFINED in case of more evaluation required
 */
static int eval_ifposfinished(const t_boolean isWhite,
        int * iMovesWhite, int * iMovesBlack) {
  posid iEpField;
  t_alg1_moves m1_movesbuffer;

  // generate for white

  // TODO: this test should not be here, should be in search modules
  if (game.curply > MAX_GAME_LENGTH - 1) {
    // if we cannot store moves anymore, consider game lost
    if (isWhite) return -SCORE_ABORT;
    else return SCORE_ABORT;
  }

  // we do not bother about possible moves if score is way off normal score
  memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));
  iEpField = m1_board.m_epfield;
  m1_board.m_epfield = 0;

  // if row 5 is set, we know that white possibly can take ep
  if (ROW144TO64(iEpField) == 5)
    m1_board.m_epfield = iEpField;
  *iMovesWhite = count_movelist(COLOUR_WHITE);

  // if 5 was set, and it should have been 2, we need to undo the clear
  if (ROW144TO64(iEpField) == 2)
    m1_board.m_epfield = iEpField;
  else
    m1_board.m_epfield = 0;
  *iMovesBlack = count_movelist(COLOUR_BLACK);
  m1_board.m_epfield = iEpField; // TODO: check if required
  memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));

  // last part of evaluation, invert if needed
  // verify if mate of stalemate position
  if (isWhite) {
    if (*iMovesBlack == 0) {
      if (alg1_isInCheck(myFALSE)) {
        return SCORE_WON;
      }
    }
    if (*iMovesWhite == 0) {
      if (alg1_isInCheck(myTRUE)) {
        return -SCORE_WON;
      } else {
        return SCORE_DRAW;
      }
    }
  } else {
    if (*iMovesWhite == 0) {
      if (alg1_isInCheck(myTRUE)) {
        return SCORE_WON;
      }
    }
    if (*iMovesBlack == 0) {
      if (alg1_isInCheck(myFALSE)) {
        return -SCORE_WON;
      } else {
        return SCORE_DRAW;
      }
    }
  }

  if (b1_theoreticdraw() == myTRUE) return SCORE_DRAW;
  if (board_positionrepeats(m1_board.m_hash) > 0) return SCORE_DRAW;
  if (b1_draw50moves(40)) return SCORE_DRAW;

  return SCORE_UNDEFINED;
}

/** Calculate mobility value
 */
t_score eval_mobility(const t_boolean isWhite) {
  int iMovesWhite;
  int iMovesBlack;

  int iEvaluation = eval_ifposfinished(isWhite, &iMovesWhite, &iMovesBlack);
  if (iEvaluation != SCORE_UNDEFINED) return iEvaluation;

  return eval_mobilitymv(isWhite, iMovesWhite, iMovesBlack);
}

/** Calculate mobility value with number of moves already available
 */
static t_score eval_mobilitymv(const t_boolean isWhite,
        const int iMovesWhite, const int iMovesBlack) {
  int iMoves;

  if (isWhite) {
    iMoves = iMovesWhite - iMovesBlack;
  } else {
    iMoves = iMovesBlack - iMovesWhite;
  }
  return ((t_score) iMoves / MOBILITY_DIVIDER);
}

/** evaluate white and black material, return score for playing colour
 */
t_score eval_material(const t_boolean isWhite) {
  return isWhite ? eval_materialwhite() : -eval_materialwhite();
}

/** evaluate white and black material, return score for white
 */
static t_score eval_materialwhite(void) {
  int iWhitePiecePos = 0;
  int iBlackPiecePos = 0;
  int iPieces = 0;

  int iPos;
  t_piece cPiece;

  // calculate piece evaluation
  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    cPiece = m1_board.m_fields[iPos];
    if ((cPiece != FIELD_BORDER) && (cPiece != FIELD_EMPTY)) {
      // calculate piece evaluation
      if (!(cPiece & (LOWERCASEBIT))) {
        // white piece
        iWhitePiecePos += m1_aPieceValues[(int) cPiece];
      } else {
        // black piece
        iBlackPiecePos += m1_aPieceValues[(int) cPiece];
      }
      iPieces++;
    }
  }

  // black pos is negative
  if (iWhitePiecePos == -iBlackPiecePos) return 0;

  // no relative calculation for last 6 pieces
  if (iPieces < 7) return iWhitePiecePos + iBlackPiecePos;

  // increase advantage in case material decreases
  return ((t_score) (iWhitePiecePos + iBlackPiecePos)) *
          (128.0 - ((t_score) (iWhitePiecePos - iBlackPiecePos))) / 64.0;
}

/** calculate king safety
 */
t_score eval_kingpos(const t_boolean isWhite) {
  int fKingSafety = 0;

  // calculate malus if king has not moved after the first moves
  if (game.curply >= OPENING_LENGTH) {
    fKingSafety -= (m1_board.m_whitebegin[4] - m1_board.m_blackbegin[4]) * 2;
  }

  if (isEndGame == myFALSE) {
    fKingSafety += whiteKingProtected();
    fKingSafety -= blackKingProtected();
  }

  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, -12) * 2;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, 12) * 4;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, -1) * 3;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, 1) * 3;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, 11) * 4;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, 13) * 4;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, -11) * 3;
  fKingSafety -= onOpenDirection(m1_board.m_whiteking, COLOUR_WHITE, -13) * 3;

  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, -12) * 4;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, 12) * 2;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, -1) * 3;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, 1) * 3;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, 11) * 3;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, 13) * 3;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, -11) * 4;
  fKingSafety += onOpenDirection(m1_board.m_blackking, COLOUR_BLACK, -13) * 4;

  if (!isWhite) {
    fKingSafety = -fKingSafety;
  }

  return (t_score) ((t_score) fKingSafety / (t_score) KING_DIVIDER);
}

/**
 * return the static evaluation of the board
 */
t_score eval_staticpos(const t_boolean isWhite) {
  posid iPos;
  t_piece cPiece;
  t_score fBoardValue = 0;

  // calculate piece evaluation and boardposition values
  // also calculate protection rate of own position
  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    cPiece = m1_board.m_fields[iPos];
    if ((cPiece != FIELD_BORDER) && (cPiece != FIELD_EMPTY)) {
      // calculate piece evaluation and boardposition values
      if (!(cPiece & (LOWERCASEBIT))) {
        // white piece
        if (game.curply <= OPENING_LENGTH) {
          // do not attribute special field values in the endgame
          if (cPiece == PIECE_PAWN) {
            fBoardValue += opening_pawnvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_KING) {
            fBoardValue += opening_kingvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_QUEEN) {
            fBoardValue += opening_queenvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_ROOK) {
            fBoardValue += opening_rookvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_BISHOP) {
            fBoardValue += opening_bishopvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_KNIGHT) {
            fBoardValue += opening_knightvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          }
        } else if (isEndGame == myTRUE) {
          if (cPiece == PIECE_PAWN) {
            fBoardValue += endgame_pawnvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          }
        } else {
          if (cPiece == PIECE_PAWN) {
            fBoardValue += pawnvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_ROOK) {
            fBoardValue += rookvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_KNIGHT) {
            fBoardValue += knightvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == PIECE_BISHOP) {
            fBoardValue += bishopvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          }
        }
      } else {
        // black piece
        if (game.curply <= OPENING_LENGTH) {
          // do not attribute special field values in the endgame
          if (cPiece == BLACK_PAWN) {
            fBoardValue -= opening_pawnvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_KING) {
            fBoardValue -= opening_kingvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_QUEEN) {
            fBoardValue -= opening_queenvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_ROOK) {
            fBoardValue -= opening_rookvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_BISHOP) {
            fBoardValue -= opening_bishopvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_KNIGHT) {
            fBoardValue -= opening_knightvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          }
        } else if (isEndGame == myTRUE) {
          if (cPiece == BLACK_PAWN) {
            fBoardValue -= endgame_pawnvalues[7 - ROW144TO64(iPos)][COL144TO64(iPos)];
          }
        } else {
          if (cPiece == BLACK_PAWN) {
            fBoardValue -= pawnvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_ROOK) {
            fBoardValue -= rookvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_KNIGHT) {
            fBoardValue -= knightvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          } else if (cPiece == BLACK_BISHOP) {
            fBoardValue -= bishopvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
          }
        }
      }
    }
  }

  if (game.curply <= OPENING_LENGTH) {
    // piece development, not counted as bonus points, but as malus points from the other
    // skip the queen and king in the calculation
    for (iPos = 0; iPos < 3; iPos++)
      fBoardValue = fBoardValue + (m1_board.m_blackbegin[iPos] - m1_board.m_whitebegin[iPos]) * STATICPOS_FACTOR;
    for (iPos = 5; iPos < 8; iPos++)
      fBoardValue = fBoardValue + (m1_board.m_blackbegin[iPos] - m1_board.m_whitebegin[iPos]) * STATICPOS_FACTOR;
  }

  // calculate malus if queen has moved the first moves
  if (game.curply <= OPENING_LENGTH) {
    fBoardValue -= (m1_board.m_whitebegin[3] - m1_board.m_blackbegin[3]) * STATICPOS_FACTOR * (game.curply / 2);
  }

  // calculate malus if rook has moved before knight or bishop
  if ((!m1_board.m_whitebegin[0]) || (!m1_board.m_whitebegin[7])) {
    fBoardValue -= (m1_board.m_whitebegin[1] + m1_board.m_whitebegin[2]
            + m1_board.m_whitebegin[5] + m1_board.m_whitebegin[6]) * STATICPOS_FACTOR;
  }
  if ((!m1_board.m_blackbegin[0]) || (!m1_board.m_blackbegin[7])) {
    fBoardValue += (m1_board.m_blackbegin[1] + m1_board.m_blackbegin[2]
            + m1_board.m_blackbegin[5] + m1_board.m_blackbegin[6]) * STATICPOS_FACTOR;
  }
  // calculate malus for stupid moves like Rg1, Rh2....
  if (game.curply <= OPENING_LENGTH) {
    if ((m1_board.m_fields[26] == PIECE_ROOK) || (m1_board.m_fields[33] == PIECE_ROOK))
      fBoardValue -= STATICPOS_FACTOR;
    if ((m1_board.m_fields[110] == BLACK_ROOK) || (m1_board.m_fields[117] == BLACK_ROOK))
      fBoardValue += STATICPOS_FACTOR;
  }

  // add malus for double pawn
  fBoardValue += (countDoublePawn(myTRUE) - countDoublePawn(myFALSE)) * DOUBLE_PAWN_MALUS;

  if (isEndGame == myFALSE) {
    // add pawn structure evaluation
    fBoardValue += whitePawns();
    fBoardValue -= blackPawns();
  }

  // last part of evaluation, invert if needed
  if (!isWhite) {
    fBoardValue = -fBoardValue;
  }

  return fBoardValue / STATICPOS_DIVIDER;
}

/**
 * return the dynamic evaluation of the board
 */
t_score eval_dynamicpos(const t_boolean isWhite) {
  posid iPos;
  t_piece cPiece;
  unsigned int iWhiteAttack;
  unsigned int iBlackAttack;
  t_score fBoardValue = 0;

  // calculate piece evaluation and boardposition values
  // also calculate protection rate of own position
  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    cPiece = m1_board.m_fields[iPos];
    if ((cPiece != FIELD_BORDER) && (cPiece != FIELD_EMPTY)) {
      // calculate piece evaluation and boardposition values

      iWhiteAttack = countUnderAttack(iPos, myTRUE);
      iBlackAttack = countUnderAttack(iPos, myFALSE);
      if (!(cPiece & (LOWERCASEBIT))) {
        // white piece
        if (iWhiteAttack > iBlackAttack)
          // more attacked than protected
          fBoardValue -= m1_aPieceValues[(int) cPiece] * DYNAMICPOS_FACTOR;
        if (cPiece == PIECE_ROOK) {
          fBoardValue += onOpenLine(iPos, COLOUR_WHITE) * DYNAMICPOS_ROOKFACTOR;
          fBoardValue += offensiveRook(iPos, COLOUR_WHITE);
        } else if (cPiece == PIECE_BISHOP) {
          fBoardValue += offensiveBishop(iPos, COLOUR_WHITE);
        }
      } else {
        // black piece
        if (iBlackAttack > iWhiteAttack)
          // more attacked than protected
          fBoardValue += m1_aPieceValues[(int) cPiece] * DYNAMICPOS_FACTOR;
        if (cPiece == BLACK_ROOK) {
          fBoardValue -= onOpenLine(iPos, COLOUR_BLACK) * DYNAMICPOS_ROOKFACTOR;
          fBoardValue -= offensiveRook(iPos, COLOUR_BLACK);
        } else if (cPiece == BLACK_BISHOP) {
          fBoardValue -= offensiveBishop(iPos, COLOUR_BLACK);
        }
      }
    }
  }

  // last part of evaluation, invert if needed
  if (!isWhite) {
    fBoardValue = -fBoardValue;
  }

  if (isEndGame == myTRUE) {
    fBoardValue *= ENDGAME_MOBILITYFACTOR;
  }
  return fBoardValue / DYNAMICPOS_DIVIDER;
}

/**
 * count the number of double pawns for a side to move
 */
static int countDoublePawn(const t_boolean bEvalForWhite) {
  int iDouble = 0;
  int iCount;
  int iCol, iRow;
  t_piece cPieceToLookFor;

  if (bEvalForWhite) cPieceToLookFor = PIECE_PAWN;
  else cPieceToLookFor = BLACK_PAWN;

  for (iCol = 0; iCol <= 7; iCol++) {
    iCount = 0;
    for (iRow = 1; iRow < 7; iRow++) {
      if (m1_board.m_fields[iCol + 2 + (iRow + 2)*12] == cPieceToLookFor) iCount++;
    }
    if (iCount > 1) iDouble += (iCount - 1);
  }

  return iDouble;
}

// --------------------------------------------------------------------

static int whiteKingProtected() {
  int iProtected = 0;
  // make sure to keep 3 pawns in line at kingside
  if ((m1_board.m_whiteking == KINGWHITESTART) && (game.curply <= OPENING_LENGTH)) {
    if ((m1_board.m_fields[43] == PIECE_PAWN)
            && (m1_board.m_fields[44] == PIECE_PAWN)
            && (m1_board.m_fields[45] == PIECE_PAWN))
      iProtected += 1;
  }
  // avoid moving in the beginning g and h pawns
  if ((m1_board.m_whiteking == KINGWHITESTART) && (game.curply < OPENING_LENGTH)) {
    if ((m1_board.m_fields[44] != PIECE_PAWN)
            || (m1_board.m_fields[45] != PIECE_PAWN))
      iProtected -= 2;
  }
  if (m1_board.m_whiteking == 32) {
    if ((m1_board.m_fields[43] == PIECE_PAWN)
            && (m1_board.m_fields[44] == PIECE_PAWN)
            && (m1_board.m_fields[45] == PIECE_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[43] == PIECE_PAWN)
            && (m1_board.m_fields[44] == PIECE_PAWN)
            && (m1_board.m_fields[57] == PIECE_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[43] == PIECE_PAWN)
            && (m1_board.m_fields[44] == PIECE_PAWN)
            && (m1_board.m_fields[56] == PIECE_BISHOP)
            && (m1_board.m_fields[45] == PIECE_PAWN))
      iProtected += 3;
  }
  if (m1_board.m_whiteking == 28) {
    if ((m1_board.m_fields[40] == PIECE_PAWN)
            && (m1_board.m_fields[39] == PIECE_PAWN)
            && (m1_board.m_fields[38] == PIECE_PAWN))
      iProtected += 2;
    if ((m1_board.m_fields[40] == PIECE_PAWN)
            && (m1_board.m_fields[39] == PIECE_PAWN)
            && (m1_board.m_fields[50] == PIECE_PAWN))
      iProtected += 2;
  }
  if (m1_board.m_whiteking == 27) {
    if ((m1_board.m_fields[40] == PIECE_PAWN)
            && (m1_board.m_fields[39] == PIECE_PAWN)
            && (m1_board.m_fields[38] == PIECE_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[40] == PIECE_PAWN)
            && (m1_board.m_fields[39] == PIECE_PAWN)
            && (m1_board.m_fields[50] == PIECE_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[40] == PIECE_PAWN)
            && (m1_board.m_fields[39] == PIECE_PAWN)
            && (m1_board.m_fields[51] == PIECE_BISHOP)
            && (m1_board.m_fields[38] == PIECE_PAWN))
      iProtected += 3;
  }

  return iProtected;
}

// --------------------------------------------------------------------

static int blackKingProtected() {
  int iProtected = 0;
  if ((m1_board.m_blackking == KINGBLACKSTART) && (game.curply <= OPENING_LENGTH)) {
    if ((m1_board.m_fields[103] == BLACK_PAWN)
            && (m1_board.m_fields[104] == BLACK_PAWN)
            && (m1_board.m_fields[105] == BLACK_PAWN))
      iProtected += 1;
  }
  if ((m1_board.m_blackking == KINGBLACKSTART) && (game.curply <= OPENING_LENGTH)) {
    if ((m1_board.m_fields[104] != BLACK_PAWN)
            || (m1_board.m_fields[105] != BLACK_PAWN))
      iProtected -= 2;
  }

  if (m1_board.m_blackking == 116) {
    if ((m1_board.m_fields[103] == BLACK_PAWN)
            && (m1_board.m_fields[104] == BLACK_PAWN)
            && (m1_board.m_fields[105] == BLACK_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[103] == BLACK_PAWN)
            && (m1_board.m_fields[104] == BLACK_PAWN)
            && (m1_board.m_fields[93] == BLACK_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[103] == BLACK_PAWN)
            && (m1_board.m_fields[92] == BLACK_PAWN)
            && (m1_board.m_fields[104] == BLACK_BISHOP)
            && (m1_board.m_fields[105] == BLACK_PAWN))
      iProtected += 3;
  }
  if (m1_board.m_blackking == 112) {
    if ((m1_board.m_fields[100] == BLACK_PAWN)
            && (m1_board.m_fields[99] == BLACK_PAWN)
            && (m1_board.m_fields[98] == BLACK_PAWN))
      iProtected += 2;
    if ((m1_board.m_fields[100] == BLACK_PAWN)
            && (m1_board.m_fields[99] == BLACK_PAWN)
            && (m1_board.m_fields[86] == BLACK_PAWN))
      iProtected += 2;
  }
  if (m1_board.m_blackking == 111) {
    if ((m1_board.m_fields[100] == BLACK_PAWN)
            && (m1_board.m_fields[99] == BLACK_PAWN)
            && (m1_board.m_fields[98] == BLACK_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[100] == BLACK_PAWN)
            && (m1_board.m_fields[99] == BLACK_PAWN)
            && (m1_board.m_fields[86] == BLACK_PAWN))
      iProtected += 3;
    if ((m1_board.m_fields[100] == BLACK_PAWN)
            && (m1_board.m_fields[87] == BLACK_PAWN)
            && (m1_board.m_fields[99] == BLACK_BISHOP)
            && (m1_board.m_fields[98] == BLACK_PAWN))
      iProtected += 3;
  }

  return iProtected;
}

// --------------------------------------------------------------------

static int whitePawns(void) {
  int iResult = 0;
  int iPos;

  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    if (m1_board.m_fields[iPos] == PIECE_PAWN) {
      // are pawns being protected?
      if ((m1_board.m_fields[iPos - 11] == PIECE_PAWN)
              || (m1_board.m_fields[iPos - 13] == PIECE_PAWN))
        iResult += 2;
      // are pawns forming a wall?
      if ((m1_board.m_fields[iPos - 1] == PIECE_PAWN)
              || (m1_board.m_fields[iPos + 2] == PIECE_PAWN))
        iResult++;
    }
  }

  // check if d3 or e3 are blocking the pawn
  if ((m1_board.m_fields[41] == PIECE_PAWN)
          && (m1_board.m_fields[53] != FIELD_EMPTY)) iResult -= 60;

  if ((m1_board.m_fields[42] == PIECE_PAWN)
          && (m1_board.m_fields[54] != FIELD_EMPTY)) iResult -= 60;

  return iResult;
}

// --------------------------------------------------------------------

static int blackPawns(void) {
  int iResult = 0;
  int iPos;

  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    if (m1_board.m_fields[iPos] == BLACK_PAWN) {
      // are pawns being protected?
      if ((m1_board.m_fields[iPos + 11] == BLACK_PAWN)
              || (m1_board.m_fields[iPos + 13] == BLACK_PAWN))
        iResult += 2;
      // are pawns forming a wall?
      if ((m1_board.m_fields[iPos - 1] == BLACK_PAWN)
              || (m1_board.m_fields[iPos + 2] == BLACK_PAWN))
        iResult++;
    }
  }

  // check if d6 or e6 are blocking the pawn
  if ((m1_board.m_fields[101] == BLACK_PAWN)
          && (m1_board.m_fields[89] != FIELD_EMPTY)) iResult -= 60;
  if ((m1_board.m_fields[102] == BLACK_PAWN)
          && (m1_board.m_fields[90] != FIELD_EMPTY)) iResult -= 60;

  return iResult;
}

/**
 * check if rook is on open line
 */
static int onOpenLine(const posid iPos, const int iColour) {
  int iResult = 3;
  int i;

  if (iColour == COLOUR_WHITE) {
    for (i = iPos + 12; i < ENDPOS; i += 12) {
      if (m1_board.m_fields[i] == PIECE_PAWN) {
        iResult = 0;
        break;
      } else if (m1_board.m_fields[i] == BLACK_PAWN) {
        iResult = 1;
        break;
      }
    }
  } else {
    for (i = iPos - 12; i >= STARTPOS; i -= 12) {
      if (m1_board.m_fields[i] == BLACK_PAWN) {
        iResult = 0;
        break;
      } else if (m1_board.m_fields[i] == PIECE_PAWN) {
        iResult = 1;
        break;
      }
    }
  }

  return iResult;
}

/**
 * return 2 if we are seeing a first enemy piece in a certain direction
 * return 1 if we are seeing a border after 2 empty cases
 * return 0 in case of own piece or border
 */
static int onOpenDirection(const posid iPos, const int iColour, const int iDir) {
  int iResult = 0;
  int iIterator = 0;
  t_piece cPiece;

  posid iDest = iPos + iDir;
  cPiece = m1_board.m_fields[iDest];
  while (cPiece != FIELD_BORDER) {
    if (cPiece != FIELD_EMPTY) {
      if (iColour == COLOUR_WHITE) {
        if (cPiece == BLACK_KING) {
          iResult = 1;
        } else if (cPiece & LOWERCASEBIT) {
          // we are white and we are encountering a black piece
          iResult = 2;
        }
      } else {
        if (cPiece == PIECE_KING) {
          iResult = 1;
        } else if (!(cPiece & LOWERCASEBIT)) {
          // we are black and are encountering a white piece
          iResult = 2;
        }
      }
      break;
    }
    iDest += iDir;
    cPiece = m1_board.m_fields[iDest];
    iIterator++;
  }

  if ((iIterator > 1) && (cPiece == FIELD_BORDER))
    iResult = 1;

  return iResult;
}

/**
 *  see if Bishop is on same line as other higher pieces
 */
static int offensiveBishop(const posid iPos, t_colour iColour) {
  t_score fResult = 0.0;
  int iPieces;

  if (iColour == COLOUR_WHITE) {
    if (onSameDir(iPos, 13, BLACK_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;
    else if (onSameDir(iPos, 11, BLACK_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;
    else if (onSameDir(iPos, -13, BLACK_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;
    else if (onSameDir(iPos, -11, BLACK_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;

    if (onSameDir(iPos, 13, BLACK_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;
    else if (onSameDir(iPos, 11, BLACK_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;
    else if (onSameDir(iPos, -13, BLACK_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;
    else if (onSameDir(iPos, -11, BLACK_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;

    if (onSameDir(iPos, 13, BLACK_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
    else if (onSameDir(iPos, 11, BLACK_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
    else if (onSameDir(iPos, -13, BLACK_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
    else if (onSameDir(iPos, -11, BLACK_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
  } else {
    if (onSameDir(iPos, 13, PIECE_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;
    else if (onSameDir(iPos, 11, PIECE_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;
    else if (onSameDir(iPos, -13, PIECE_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;
    else if (onSameDir(iPos, -11, PIECE_KING, &iPieces))
      fResult += 15 / (t_score) iPieces;

    if (onSameDir(iPos, 13, PIECE_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;
    else if (onSameDir(iPos, 11, PIECE_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;
    else if (onSameDir(iPos, -13, PIECE_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;
    else if (onSameDir(iPos, -11, PIECE_QUEEN, &iPieces))
      fResult += 8 / (t_score) iPieces;

    if (onSameDir(iPos, 13, PIECE_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
    else if (onSameDir(iPos, 11, PIECE_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
    else if (onSameDir(iPos, -13, PIECE_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
    else if (onSameDir(iPos, -11, PIECE_ROOK, &iPieces))
      fResult += 2 / (t_score) iPieces;
  }

  return (int) fResult;
}

// --------------------------------------------------------------------

static int offensiveRook(const posid iPos, t_colour iColour) {
  t_score fResult = 0.0;
  int iPieces;

  if (iColour == COLOUR_WHITE) {
    if (onSameDir(iPos, 12, BLACK_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;
    else if (onSameDir(iPos, 1, BLACK_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;
    else if (onSameDir(iPos, -12, BLACK_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;
    else if (onSameDir(iPos, -1, BLACK_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;

    if (onSameDir(iPos, 12, BLACK_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;
    else if (onSameDir(iPos, 1, BLACK_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;
    else if (onSameDir(iPos, -12, BLACK_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;
    else if (onSameDir(iPos, -1, BLACK_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;

  } else {
    if (onSameDir(iPos, 12, PIECE_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;
    else if (onSameDir(iPos, 1, PIECE_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;
    else if (onSameDir(iPos, -12, PIECE_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;
    else if (onSameDir(iPos, -1, PIECE_KING, &iPieces))
      fResult += 10 / (t_score) iPieces;

    if (onSameDir(iPos, 12, PIECE_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;
    else if (onSameDir(iPos, 1, PIECE_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;
    else if (onSameDir(iPos, -12, PIECE_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;
    else if (onSameDir(iPos, -1, PIECE_QUEEN, &iPieces))
      fResult += 6 / (t_score) iPieces;

  }

  return (int) fResult;
}

// --------------------------------------------------------------------

static int onSameDir(const posid iPos, const int iDir, const t_piece cPieceToLookFor, int *iPieces) {
  t_piece cPiece;
  posid iDest = iPos + iDir;
  int iHit = myFALSE;

  *iPieces = 1;

  while (1) {
    cPiece = m1_board.m_fields[iDest];
    if (cPiece == FIELD_BORDER)
      break;
    if (cPiece != FIELD_EMPTY)
      (*iPieces)++;
    if (cPiece == cPieceToLookFor) {
      iHit = myTRUE;
      break;
    }
    iDest += iDir;
  }

  // comment out, avoid division by zero, calculation in calling function should be adapted though for
  // better evaluation
  // if (iHit) *iPieces--;

  return iHit;
}

/**
 * generate internal buffers to speed up evaluation functions
 */
void e1_constructor(void) {
  int i;

  for (i = 0; i < FIELDVALUE_SIZE; i++) m1_aPieceValues[i] = 0;

  m1_aPieceValues[(int) PIECE_PAWN] = 1;
  m1_aPieceValues[(int) BLACK_PAWN] = -1;
  m1_aPieceValues[(int) PIECE_KNIGHT] = 3;
  m1_aPieceValues[(int) BLACK_KNIGHT] = -3;
  m1_aPieceValues[(int) PIECE_BISHOP] = 3;
  m1_aPieceValues[(int) BLACK_BISHOP] = -3;
  m1_aPieceValues[(int) PIECE_ROOK] = 5;
  m1_aPieceValues[(int) BLACK_ROOK] = -5;
  m1_aPieceValues[(int) PIECE_QUEEN] = 10;
  m1_aPieceValues[(int) BLACK_QUEEN] = -10;
  //m1_aPieceValues[(int)PIECE_KING] = 0;
  //m1_aPieceValues[(int)BLACK_KING] = 0;
  //m1_aPieceValues[(int)FIELD_EMPTY] = 0;
}

// --------------------------------------------------------------------
// eof
