/***********************************************************************************
 * File: alg1_search.cpp
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#include "belofte.h"

#include "usercmd.h"
#include "board.h"
#include "search.h"

#include "alg1_moves.h"
#include "alg1_eval.h"
#include "alg1_book.h"
#include "alg1_board.h"
#include "alg1_search.h"

// --------------------------------------------------------------------

t_alg1_moves m1_movelist;

// --------------------------------------------------------------------

extern t_alg1_board m1_board;
extern t_alg1_board m1_boardmapper;
extern game_t game;
extern uimode_t uimode;
extern level_t level;
extern uistate_t uistate;
extern search_t search;

// --------------------------------------------------------------------

static int iCurPly; // local search buffer to find the current ply while iterating

static t_boolean bCountOnly;
static int iStoredMoves;

static char szBestMoveList[MAX_DATANAME_LENGTH];

// --------------------------------------------------------------------

static void alg1_storebest(const t_depth iDepth,
        const char * szSANMove, const t_score fScore, const hashkey_t position);
static void alg1_depthcomplete(const t_depth iDepth);
static void alg1_perft(const t_depth iSearchDepth);
static t_score alg1_search_negamax(const t_depth iSearchDepthRemaining,
        const t_depth iMaxQSDepth,
        const t_boolean bBreakAllowed,
        t_score fAlpha, t_score fBeta);
static t_score alg1_negamax_iterate(const t_depth iSearchDepthRemaining,
        const t_depth iInitialSearchDepth, const t_depth iMaxQSDepth,
        t_score fAlpha, t_score fBeta);
static t_score qs_atdepth(const t_depth iSearchDepthRemaining,
        const t_boolean bBreakAllowed, t_score fAlpha, t_score fBeta);

static void add_singlemove(const posid iFrom, const posid iTo);
static void add_singlemove_promotion(const posid iFrom, const posid iTo,
        const t_promotion cFlag, const int iWhoInCheck);

static void search_fill_whitemoves(void);
static void search_fill_blackmoves(void);

static void add_multiplemove_promotion_white(const posid iPos, const posid iDest);
static void add_multiplemove_promotion_black(const posid iPos, const posid iDest);
static void fill_movesforknightwhite(const posid iPos);
static void fill_movesforbishopwhite(const posid iPos);
static void fill_movesforrookwhite(const posid iPos);
static void fill_movesforknightblack(const posid iPos);
static void fill_movesforbishopblack(const posid iPos);
static void fill_movesforrookblack(const posid iPos);

static void fill_movesforpieceonewhite(const posid iPos, const posid iResult);
static void fill_movesforpieceoneblack(const posid iPos, const posid iResult);

static void fill_movesforpiecedir(const posid iPos, t_colour iCaptureAllowedColour,
        const int iDir, const int iRep);

static void fill_movesforkingwhite(const posid iPos);
static void fill_movesforkingblack(const posid iPos);

static void fill_movesforpawnwhite(const posid iPos, const int iDir);
static void fill_movesforpawnblack(const posid iPos, const int iDir);

static int isInCheckAfterMove(const posid iFrom, const posid iTo,
        const t_promotion cFlag);

static t_boolean pieceseesmore(const posid iPiecePos, const t_piece cPieceToLookFor1,
        const t_piece cPieceToLookFor2, const t_boolean bEvaluateForWhite,
        const int iJump, const int iCount);

/*----------------------------------------------------------------------+
 |      I/O functions                                                   |
 +----------------------------------------------------------------------*/

// --------------------------------------------------------------------

void alg1_search_printablemove(char *szMove, const int iSelectedMove) {
  // we found a move
  // store it intermediately so we can use the move for internal housekeeping tasks
  if (m1_movelist.m1_moves[iSelectedMove].bFlag) {
    // promotion flag is set, so write out = plus piece
    sprintf(szMove, "%c%c%c%c%c",
            M1_LIJN(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucFrom)) + 'a',
            M1_RIJ(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucFrom)) + '1',
            M1_LIJN(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucTo)) + 'a',
            M1_RIJ(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucTo)) + '1',
            m1_movelist.m1_moves[iSelectedMove].bFlag);
  } else {
    sprintf(szMove, "%c%c%c%c",
            M1_LIJN(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucFrom)) + 'a',
            M1_RIJ(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucFrom)) + '1',
            M1_LIJN(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucTo)) + 'a',
            M1_RIJ(MAP144TO64(m1_movelist.m1_moves[iSelectedMove].ucTo)) + '1');
  }
}

/** store best move found at certain level, in case we have to interrupt search,
 * we can take the best move found from previous depth search
 * this also displays pondering output
 * intermediate while level is not completed
 * TODO: store complete PV line instead of root move
 */
static void alg1_storebest(const t_depth iDepth,
        const char * szSANMove, const t_score fScore, const hashkey_t position) {

  if (search.search_cancel == myFALSE) {
    strcpy(search.bestmoves[iDepth].szMove, szSANMove);
    search.bestmoves[iDepth].m_hash = position;
    search.bestmoves[iDepth].fMoveScore = fScore;
    if (fScore > search.fSearchScore) {
      // we found a new bestscore even before full search is reached
      search.fSearchScore = fScore;
      if (iDepth != search.bestmovedepth) {
        // first at this depth
        search.bestmovedepth = iDepth;
        cmdresp_snd_thinking(iDepth, '+', fScore, szSANMove);
      } else {
        // improvement
        cmdresp_snd_thinking(iDepth, '&', fScore, szSANMove);
      }
    } else {
      // not new bestscore
      cmdresp_snd_thinking(iDepth, '-', fScore, szSANMove);
    }
  }
}

/** confirm best move found at certain level
 * final call when level search is completed
 */
static void alg1_depthcomplete(const t_depth iDepth) {
  search.fSearchScore = search.bestmoves[iDepth].fMoveScore;
  if ((search.bestmovedepth == iDepth) && (iDepth > 1)) {
    // we already moved to this depth
    cmdresp_snd_thinking(iDepth, ':', search.fSearchScore,
            search.bestmoves[iDepth].szMove);
  } else {
    search.bestmovedepth = iDepth;
    cmdresp_snd_thinking(iDepth, '.', search.fSearchScore,
            search.bestmoves[iDepth].szMove);
  }
}

/**
 * search until iMaxDepth
 * exit if search.cancel is set or iMaxDepth is reached
 */
void alg1_iterativedeepening(const t_depth iMaxDepth) {
  int iQSDepth = level.quiescence == myTRUE ? MAX_DEPTH_EXTENTION : 0;

  // set some global variables
  iCurPly = game.curply;
  bCountOnly = myFALSE;

  if (level.level_mode == LEVELMODE_PERFTEST) {
    uistate.iCurrentDepth = iMaxDepth;
    alg1_perft(uistate.iCurrentDepth);
    return;
  }

  search.fInitialMaterialScore = eval_material((t_boolean) (iCurPly & 1));

  if (level.level_mode == LEVELMODE_NOTIME) {
    // FD mode, only search full depth
    uistate.iCurrentDepth = iMaxDepth;
    alg1_search_negamax(uistate.iCurrentDepth,
            iQSDepth, myTRUE, -SCORE_WON, SCORE_WON);
    return;
  }

  // in case of iterative depth, first do search at depth 1 with no QS
  // to force the engine to do at least a valid move
  // no break possibility
  uistate.iCurrentDepth = 1;
  alg1_search_negamax(1, 1, myFALSE, -SCORE_WON, SCORE_WON);

  // test if mate is found, no need to search any further
  if ((search.bestmovedepth > 0) && (search.fSearchScore >= SCORE_ALMOST_WON))
    return;

  // in case we search only depth one, no iterative deepening required
  if (iMaxDepth < 2) {
    if (iQSDepth > 1) {
      alg1_search_negamax(1, iQSDepth, myTRUE, -SCORE_WON, SCORE_WON);
    }
    return;
  }

  // do not use QS search for depth search of one
  for (uistate.iCurrentDepth = 2;
          uistate.iCurrentDepth <= iMaxDepth;
          uistate.iCurrentDepth++) {

    // search to depth, allow to break out in case movenowe or end of time
    alg1_search_negamax(uistate.iCurrentDepth,
            iQSDepth, myTRUE, -SCORE_WON, SCORE_WON);

    if (search.bestmovedepth == uistate.iCurrentDepth) {
      // bestmove has been updated, test if it gets within bounds to
      // cancel search
      if (search.fSearchScore >= SCORE_ALMOST_WON)
        break;
    }

    if (search.search_cancel == myTRUE) {
      break;
    }
  }
}

/** search finding best move
 * @param iMaxDepth depth to search at
 * @param iMaxQSDepth qieuscence search extension maxdepth
 * @param bBreakAllowed allow to break out in case of search cancel
 * @return bestscore of search
 */
static t_score alg1_search_negamax(const t_depth iMaxDepth,
        const t_depth iMaxQSDepth,
        const t_boolean bBreakAllowed,
        t_score fAlpha, t_score fBeta) {
  int iMoves;
  int i;
  int iResult;
  t_score fResult;

  t_alg1_board m1_boardbuffer;
  t_alg1_moves m1_movesbuffer;

  t_score fBestScore = -SCORE_WON;
  char szMoveEvaluating[MAX_COORDMOVE_LENGTH];

  // curply is even for white
  iMoves = search_fill_movelist((t_colour) (iCurPly & 1));

  if (iMoves > 0) {
    moves_initorder(iMoves);

    // store board position and movelist, for later restore
    memcpy(&m1_boardbuffer, &m1_board, sizeof (m1_board));
    memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));

    for (i = 0; i < iMoves; i++) {

      alg1_search_printablemove(szMoveEvaluating, i);

      iCurPly++;

      b1_apply_move(iCurPly, m1_movelist.m1_moves[i].ucFrom,
              m1_movelist.m1_moves[i].ucTo,
              m1_movelist.m1_moves[i].bFlag);

      if (iMaxDepth > 1) {
        // search one level deeper
        fResult = -alg1_negamax_iterate(iMaxDepth - 1, iMaxDepth,
                iMaxQSDepth, -fBeta, -fAlpha);
      } else if (iMaxQSDepth > 0) {
        // search in quiescence mode
        fResult = -qs_atdepth(iMaxQSDepth, myFALSE, -fBeta, -fAlpha);
      } else {
        fResult = eval_pos((t_boolean) (iCurPly & 1));
      }

      iCurPly--;

      // keep the score of all moves for the higher level
      m1_movesbuffer.m1_moves[i].bValue = fResult;
      // keep the calculated hash value
      m1_movesbuffer.m1_moves[i].m_hash = m1_board.m_hash;

      // restore old position and movelist
      memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));
      memcpy(&m1_board, &m1_boardbuffer, sizeof (m1_board));

      if (i == 0) {
        // store first move in any case, here correct move ordering
        // becomes important
        alg1_storebest(iMaxDepth, szMoveEvaluating, fResult, m1_board.m_hash);
      }

      // cancel out before storing anything
      if ((iMaxDepth > 1)
              && (bBreakAllowed == myTRUE)
              && (search.search_cancel == myTRUE)) break;

      if (fResult > fBestScore) {
        fBestScore = fResult;
        if (i != 0)
          // already stored above, no update
          alg1_storebest(iMaxDepth, szMoveEvaluating, fResult, m1_board.m_hash);
      }

      if (fResult >= fBeta) {
        break;
      }

      if (fResult > fAlpha) {
        fAlpha = fResult;
      }

    } // for all moves

    if ((bBreakAllowed == myFALSE)
            || (search.search_cancel == myFALSE)) {

      // confirm stored move at depth 1 or when no cancel done or allowed
      alg1_depthcomplete(iMaxDepth);

      // print out if debug requested the scores of the best moves
      if (uimode.debug == myTRUE) {
        moves_resort();
        for (i = 0; i < m1_movelist.m1_count; i++) {
          alg1_search_printablemove(szMoveEvaluating, i);
          snprintf(szBestMoveList, sizeof (szBestMoveList),
                  "%2d: %s score %02.3f %llx",
                  i, szMoveEvaluating, m1_movelist.m1_moves[i].bValue,
                  m1_movelist.m1_moves[i].m_hash);
          cmdresp_snd_debug(szBestMoveList);
        }
      }
    }

  } else {
    // there are no moves found
    // curply is even for white
    // TODO: do we even reach here, should stalemate be covered in position
    // eval before reaching here?
    iResult = eval_stale_pos((t_boolean) (!(iCurPly & 1)));
    if (iResult != SCORE_UNDEFINED) fBestScore = iResult;
  }

  return fBestScore;
}

/** recursive search finding best move
 * @param iSearchDepthRemaining depth to search at
 * @param iInitialSearchDepth initial depth at which search is called
 * @param iMaxQSDepth qieuscence search extension maxdepth
 * @return bestscore of search
 */
static t_score alg1_negamax_iterate(const t_depth iSearchDepthRemaining,
        const t_depth iInitialSearchDepth, const t_depth iMaxQSDepth,
        t_score fAlpha, t_score fBeta) {
  int iMoves;
  int i;
  int iResult;
  t_score fResult;

  t_alg1_board m1_boardbuffer;
  t_alg1_moves m1_movesbuffer;

  t_score fBestScore = -SCORE_WON;
  char szMoveEvaluating[MAX_COORDMOVE_LENGTH];

  if (board_positionrepeats(m1_board.m_hash) > 0) return SCORE_DRAW;

  // curply is even for white
  iMoves = search_fill_movelist((t_colour) (iCurPly & 1));

  if (iMoves > 0) {
    moves_initorder(iMoves);

    // store board position and movelist, for later restore
    memcpy(&m1_boardbuffer, &m1_board, sizeof (m1_board));
    memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));

    for (i = 0; i < iMoves; i++) {

      alg1_search_printablemove(szMoveEvaluating, i);

      iCurPly++;

      b1_apply_move(iCurPly, m1_movelist.m1_moves[i].ucFrom,
              m1_movelist.m1_moves[i].ucTo,
              m1_movelist.m1_moves[i].bFlag);

      if (iSearchDepthRemaining > 1) {
        // search one level deeper
        fResult = -alg1_negamax_iterate(iSearchDepthRemaining - 1, iInitialSearchDepth,
                iMaxQSDepth, -fBeta, -fAlpha);
      } else if (iMaxQSDepth > 0) {
        // search in quiescence mode
        fResult = -qs_atdepth(iMaxQSDepth, myTRUE, -fBeta, -fAlpha);
      } else {
        fResult = eval_pos((t_boolean) (iCurPly & 1));
      }

      iCurPly--;

      // keep the score of all moves for the higher level
      m1_movesbuffer.m1_moves[i].bValue = fResult;
      // keep the calculated hash value
      m1_movesbuffer.m1_moves[i].m_hash = m1_board.m_hash;

      // restore old position and movelist
      memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));
      memcpy(&m1_board, &m1_boardbuffer, sizeof (m1_board));

      // cancel out before storing anything
      if ((iSearchDepthRemaining > 1)
              && (search.search_cancel == myTRUE)) break;

      if (fResult > fBestScore) {
        fBestScore = fResult;
      }

      if (fResult >= fBeta) {
        break;
      }

      if (fResult > fAlpha) {
        fAlpha = fResult;
      }

    } // for all moves

  } else {
    // there are no moves found
    // curply is even for white
    // TODO: do we even reach here, should stalemate be covered in position
    // eval before reaching here?
    iResult = eval_stale_pos((t_boolean) (!(iCurPly & 1)));
    if (iResult != SCORE_UNDEFINED) fBestScore = iResult;
  }

  // make sure the scores go to zero
  // or in case matenet, deduct mate in x from score
  if (fBestScore > DEPTH_PENALTY) {
    if (fBestScore > SCORE_ALMOST_WON)
      return fBestScore - SCORE_MATEINONE;
    return fBestScore - (t_score) DEPTH_PENALTY;
  } else if (fBestScore < -DEPTH_PENALTY) {
    if (fBestScore < -SCORE_ALMOST_WON)
      return fBestScore + SCORE_MATEINONE;

    return fBestScore + (t_score) DEPTH_PENALTY;
  }

  return fBestScore;
}

/** do move generation performance test
 */
static void alg1_perft(const t_depth iSearchDepth) {
  int iMoves;
  int i;

  t_alg1_board m1_boardbuffer;
  t_alg1_moves m1_movesbuffer;

  if (iSearchDepth == 0) {
    // resolve perftest 0
    uistate.perftnodes++;
    return;
  }
  // TODO: optimize, depth 1, do not make/unmake move but just count

  // curply is even for white
  iMoves = search_fill_movelist((t_colour) (iCurPly & 1));

  if (iMoves > 0) {

    // store board position and movelist, for later restore
    memcpy(&m1_boardbuffer, &m1_board, sizeof (m1_board));
    memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));

    for (i = 0; i < iMoves; i++) {

      iCurPly++;

      b1_apply_move(iCurPly, m1_movelist.m1_moves[i].ucFrom,
              m1_movelist.m1_moves[i].ucTo,
              m1_movelist.m1_moves[i].bFlag);

      alg1_perft(iSearchDepth - 1);

      iCurPly--;

      // restore old position and movelist
      memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));
      memcpy(&m1_board, &m1_boardbuffer, sizeof (m1_board));

    } // for all moves
  }
}

/**
 * quiescence search
 * @param iSearchMaxDepth is to avoid endless depth or to cut off QS early
 * @return bestscore of search
 * TODO: check evolution of fAlpha
 */
static t_score qs_atdepth(const t_depth iSearchMaxDepth,
        const t_boolean bBreakAllowed, t_score fAlpha, t_score fBeta) {
  int iMoves;
  int i;
  t_score fResult;

  t_alg1_board m1_boardbuffer;
  t_alg1_moves m1_movesbuffer;

  t_score fBestScore = eval_pos((t_boolean) (!(iCurPly & 1)));

  if (fBestScore >= fBeta) {
    return fBestScore;
  }

  // curply is even for white
  iMoves = qsfill_movelist((t_colour) (iCurPly & 1));

  if (iMoves > 0) {
    // store board position and movelist, for later restore
    memcpy(&m1_boardbuffer, &m1_board, sizeof (m1_board));
    memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));

    for (i = 0; i < iMoves; i++) {
      if ((search.search_cancel == myTRUE)
              && (bBreakAllowed == myTRUE)) break;

      iCurPly++;

      b1_apply_move(iCurPly, m1_movelist.m1_moves[i].ucFrom,
              m1_movelist.m1_moves[i].ucTo,
              m1_movelist.m1_moves[i].bFlag);

      if (iSearchMaxDepth > 0) {
        // search one level deeper, -max(alpha, best)
        // TODO: move test to function entry
        fResult = -qs_atdepth(iSearchMaxDepth - 1, bBreakAllowed, -fBeta, -fAlpha);
      } else {
        fResult = eval_pos_failmaterial((t_boolean) (iCurPly & 1),
                search.fInitialMaterialScore);
      }

      iCurPly--;

      // keep the score of all moves for the higher level
      m1_movesbuffer.m1_moves[i].bValue = fResult;

      // restore old position and movelist
      memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));
      memcpy(&m1_board, &m1_boardbuffer, sizeof (m1_board));

      // cancel out before storing anything
      if ((search.search_cancel == myTRUE)
              && (bBreakAllowed == myTRUE)) break;

      if (fResult > fBestScore) {
        fBestScore = fResult;
      }

      if (fResult >= fBeta) {
        break;
      }

      if (fResult > fAlpha) {
        fAlpha = fResult;
      }
    }
    // } else {
    // there are no quiescence moves found
  }

  return fBestScore;
}

/*----------------------------------------------------------------------+
|      board interface functions                                        |
+----------------------------------------------------------------------*/

/**
 * generate internal buffers to speed up further working of the program
 */
void alg1_constructor(void) {
  int iRow, iColumn;

  // reset board to invalid
  for (iRow = 0; iRow < 12; iRow++) {
    for (iColumn = 0; iColumn < 12; iColumn++) {
      m1_board.m_fields[iColumn + iRow * 12] = FIELD_BORDER;
    }
  }

  // reset board to invalid
  for (iRow = 0; iRow < 12; iRow++) {
    for (iColumn = 0; iColumn < 12; iColumn++) {
      m1_boardmapper.m_fields[iColumn + iRow * 12] = FIELD_BORDER;
    }
  }

  for (iRow = 0; iRow < 8; iRow++) {
    for (iColumn = 0; iColumn < 8; iColumn++) {
      m1_boardmapper.m_fields[iColumn + 2 + (iRow + 2)*12] = (t_piece) (iColumn + iRow * 8);
    }
  }

}

/*----------------------------------------------------------------------+
|      move generation                                                  |
+----------------------------------------------------------------------*/

/**
 * fill all possible moves in the movelist
 */
int search_fill_movelist(const t_colour nColour) {

  memset(&m1_movelist, 0, sizeof (m1_movelist));
  bCountOnly = myFALSE;

  // we will take each pieace and generate a move for it
  if (nColour == COLOUR_BLACK) {
    search_fill_blackmoves();
  } else {
    search_fill_whitemoves();
  }

  return m1_movelist.m1_count;
}

/**
 * count all possible moves in the movelist
 */
int count_movelist(const t_colour nColour) {
  iStoredMoves = 0;
  bCountOnly = myTRUE;

  // we will take each pieace and generate a move for it
  if (nColour == COLOUR_BLACK) {
    search_fill_blackmoves();
  } else {
    search_fill_whitemoves();
  }

  return iStoredMoves;
}

/** fill all possible white moves
 */
static void search_fill_whitemoves(void) {
  posid iPos;
  t_piece cPiece;
  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    cPiece = m1_board.m_fields[iPos];
    if (cPiece == PIECE_PAWN) {
      fill_movesforpawnwhite(iPos, 12);
    } else if (cPiece == PIECE_KNIGHT) {
      fill_movesforknightwhite(iPos);
    } else if (cPiece == PIECE_BISHOP) {
      fill_movesforbishopwhite(iPos);
    } else if (cPiece == PIECE_ROOK) {
      fill_movesforrookwhite(iPos);
    } else if (cPiece == PIECE_KING) {
      fill_movesforkingwhite(iPos);
    } else if (cPiece == PIECE_QUEEN) {
      fill_movesforbishopwhite(iPos);
      fill_movesforrookwhite(iPos);
    }
  }
}

/** fill all possible black moves
 */
static void search_fill_blackmoves(void) {
  posid iPos;
  t_piece cPiece;
  for (iPos = STARTPOS; iPos < ENDPOS; iPos++) {
    cPiece = m1_board.m_fields[iPos];
    if (cPiece == BLACK_PAWN) {
      fill_movesforpawnblack(iPos, -12);
    } else if (cPiece == BLACK_KNIGHT) {
      fill_movesforknightblack(iPos);
    } else if (cPiece == BLACK_BISHOP) {
      fill_movesforbishopblack(iPos);
    } else if (cPiece == BLACK_ROOK) {
      fill_movesforrookblack(iPos);
    } else if (cPiece == BLACK_KING) {
      fill_movesforkingblack(iPos);
    } else if (cPiece == BLACK_QUEEN) {
      fill_movesforbishopblack(iPos);
      fill_movesforrookblack(iPos);
    }
  }
}

/**
 * do a normal fill movelist and then only count non silent moves
 * place non silent moves at top
 */
int qsfill_movelist(const t_colour nColour) {
  int iMoves = search_fill_movelist(nColour);
  return moves_qsinitorder(iMoves);
}

/**
 * fill all possibe king moves
 */
static void fill_movesforkingwhite(const posid iPos) {
  fill_movesforpieceonewhite(iPos, iPos + 11);
  fill_movesforpieceonewhite(iPos, iPos + 12);
  fill_movesforpieceonewhite(iPos, iPos + 13);
  fill_movesforpieceonewhite(iPos, iPos + 1);
  fill_movesforpieceonewhite(iPos, iPos - 11);
  fill_movesforpieceonewhite(iPos, iPos - 12);
  fill_movesforpieceonewhite(iPos, iPos - 13);
  fill_movesforpieceonewhite(iPos, iPos - 1);

  if (isUnderAttack(m1_board.m_whiteking, myTRUE) == myFALSE) {
    if (m1_board.m_whitecastle[CASTLE_LONG]) {
      if ((m1_board.m_fields[27] == FIELD_EMPTY)
              && (m1_board.m_fields[28] == FIELD_EMPTY)
              && (m1_board.m_fields[29] == FIELD_EMPTY)
              && (m1_board.m_fields[26] == PIECE_ROOK)
              && (isUnderAttack(29, myTRUE) == myFALSE)) {
        add_singlemove(30, 28);
      }
    }
    if (m1_board.m_whitecastle[CASTLE_SHORT]) {
      if ((m1_board.m_fields[31] == FIELD_EMPTY)
              && (m1_board.m_fields[32] == FIELD_EMPTY)
              && (m1_board.m_fields[33] == PIECE_ROOK)
              && (isUnderAttack(31, myTRUE) == myFALSE)) {

        add_singlemove(30, 32);
      }
    }
  }
}

// --------------------------------------------------------------------

static void fill_movesforkingblack(const posid iPos) {
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 11);
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 12);
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 13);
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 1);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 11);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 12);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 13);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 1);

  if (isUnderAttack(m1_board.m_blackking, myFALSE) == myFALSE) {
    if (m1_board.m_blackcastle[CASTLE_LONG]) {
      if ((m1_board.m_fields[111] == FIELD_EMPTY)
              && (m1_board.m_fields[112] == FIELD_EMPTY)
              && (m1_board.m_fields[113] == FIELD_EMPTY)
              && (m1_board.m_fields[110] == BLACK_ROOK)
              && (isUnderAttack(113, myFALSE) == myFALSE)) {
        add_singlemove(114, 112);
      }
    }
    if (m1_board.m_blackcastle[CASTLE_SHORT]) {
      if ((m1_board.m_fields[115] == FIELD_EMPTY)
              && (m1_board.m_fields[116] == FIELD_EMPTY)
              && (m1_board.m_fields[117] == BLACK_ROOK)
              && (isUnderAttack(115, myFALSE) == myFALSE)) {

        add_singlemove(114, 116);
      }
    }
  }

}

static void fill_movesforrookwhite(const posid iPos) {

  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, 12, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, 1, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, -12, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, -1, 7);
}

static void fill_movesforbishopwhite(const posid iPos) {

  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, 13, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, 11, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, -13, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_BLACK, -11, 7);
}

// --------------------------------------------------------------------

static void fill_movesforrookblack(const posid iPos) {

  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, 12, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, 1, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, -12, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, -1, 7);
}

static void fill_movesforbishopblack(const posid iPos) {

  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, 13, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, 11, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, -13, 7);
  fill_movesforpiecedir(iPos, CAPTURE_COLOUR_WHITE, -11, 7);
}

// --------------------------------------------------------------------

static void fill_movesforknightwhite(const posid iPos) {
  // calculate all knights moves

  fill_movesforpieceonewhite(iPos, (const posid) iPos + 25);
  fill_movesforpieceonewhite(iPos, (const posid) iPos + 23);
  fill_movesforpieceonewhite(iPos, (const posid) iPos + 14);
  fill_movesforpieceonewhite(iPos, (const posid) iPos + 10);
  fill_movesforpieceonewhite(iPos, (const posid) iPos - 25);
  fill_movesforpieceonewhite(iPos, (const posid) iPos - 23);
  fill_movesforpieceonewhite(iPos, (const posid) iPos - 14);
  fill_movesforpieceonewhite(iPos, (const posid) iPos - 10);
}

// --------------------------------------------------------------------

static void fill_movesforknightblack(const posid iPos) {
  // calculate all knights moves

  fill_movesforpieceoneblack(iPos, (const posid) iPos + 25);
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 23);
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 14);
  fill_movesforpieceoneblack(iPos, (const posid) iPos + 10);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 25);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 23);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 14);
  fill_movesforpieceoneblack(iPos, (const posid) iPos - 10);
}

/**
 * add a single move
 * fill cflag in case of promotion
 * set cNonSilent in case of check or capture
 * TODO add notion of king protector, only when moving kingprotector the incheck test needs
 * to be performed
 * Also, no test required for opponent check through pawn or knight
 */
static void add_singlemove(const posid iFrom, const posid iTo) {
  int i;
  int iWhoInCheck = isInCheckAfterMove(iFrom, iTo, PROMOTION_NONE);

  if (bCountOnly) {
    if (iWhoInCheck != 1)
      iStoredMoves++;
  } else {
    i = m1_movelist.m1_count;

    if (iWhoInCheck != 1) {
      // the player to play is not in check
      m1_movelist.m1_moves[i].ucFrom = iFrom;
      m1_movelist.m1_moves[i].ucTo = iTo;
      m1_movelist.m1_moves[i].bFlag = PROMOTION_NONE;
      m1_movelist.m1_moves[i].bValue = -SCORE_ALMOST_WON;

      if (m1_board.m_fields[iTo] == FIELD_EMPTY) {
        m1_movelist.m1_moves[i].bNonSilent = myFALSE;
      } else {
        m1_movelist.m1_moves[i].bNonSilent = myTRUE;
      }

      m1_movelist.m1_count++;
    } else {

      iStoredMoves++;
    }

  }
}

static void add_singlemove_promotion(const posid iFrom, const posid iTo,
        const t_promotion cFlag, const int iWhoInCheck) {

  int i;
  if (bCountOnly) {
    if (iWhoInCheck != 1)
      iStoredMoves++;
  } else {

    i = m1_movelist.m1_count;

    if (iWhoInCheck != 1) {
      // the player to play is not in check
      m1_movelist.m1_moves[i].ucFrom = iFrom;
      m1_movelist.m1_moves[i].ucTo = iTo;
      m1_movelist.m1_moves[i].bFlag = cFlag;
      m1_movelist.m1_moves[i].bValue = -SCORE_ALMOST_WON;
      m1_movelist.m1_moves[i].bNonSilent = myTRUE;

      m1_movelist.m1_count++;
    } else {
      iStoredMoves++;
    }

  }
}

/**
 * add all moves in this direction for a piece
 */
static void fill_movesforpiecedir(const posid iPos, t_colour iCaptureAllowedColour,
        const int iDir, const int iRep) {
  int i;
  char cPieceResult;
  int isWhite;

  posid iResult = iPos;

  for (i = 0; i < iRep; i++) {
    // repeat all moves in a direction
    iResult += iDir;
    cPieceResult = m1_board.m_fields[iResult];
    if (cPieceResult == FIELD_EMPTY) {
      // empty case, no capture move, so move possible
      add_singlemove(iPos, iResult);
    } else if (cPieceResult == FIELD_BORDER) {
      // if we reach the border, we do not continue anymore
      break;
    } else {
      // we will look if we can capture
      isWhite = (!(cPieceResult & (LOWERCASEBIT)));
      if ((isWhite && (iCaptureAllowedColour == COLOUR_WHITE))
              || ((!isWhite) && (iCaptureAllowedColour == COLOUR_BLACK))) {
        // we know that the allowed colour for capture is the colour of the destination piece

        add_singlemove(iPos, iResult);
      }
      // we still break out as we cannot take 2 pieces in a row, and cannot jump over own pieces
      break;
    }
  }
}

/**
 * add single step move
 */
static void fill_movesforpieceonewhite(const posid iPos, const posid iResult) {
  char cPieceResult;

  cPieceResult = m1_board.m_fields[iResult];
  if (cPieceResult == FIELD_EMPTY) {
    // empty case, no capture move, so move possible
    add_singlemove(iPos, iResult);
  } else if (cPieceResult != FIELD_BORDER) {
    // we will look if we can capture
    if (cPieceResult & (LOWERCASEBIT)) {
      // is it a black piece?
      // we know that the allowed colour for capture is the colour of the destination piece
      add_singlemove(iPos, iResult);
    }
  }
}

/**
 * add single step move
 */
static void fill_movesforpieceoneblack(const posid iPos, const posid iResult) {
  char cPieceResult;

  cPieceResult = m1_board.m_fields[iResult];
  if (cPieceResult == FIELD_EMPTY) {
    // empty case, no capture move, so move possible
    add_singlemove(iPos, iResult);
  } else if (cPieceResult != FIELD_BORDER) {
    // we will look if we can capture
    // TODO: optimize by giving the bit as parameter
    if (!(cPieceResult & (LOWERCASEBIT))) {
      // we know that the allowed colour for capture is the colour of the destination piece
      add_singlemove(iPos, iResult);
    }
  }
}

/**
 * store all promotion moves
 */
static void add_multiplemove_promotion_white(const posid iPos, const posid iDest) {

  int iWhoInCheck = isInCheckAfterMove(iPos, iDest, PROMOTION_PAWN);

  add_singlemove_promotion(iPos, iDest, PROMOTION_QUEEN, iWhoInCheck);
  add_singlemove_promotion(iPos, iDest, PROMOTION_ROOK, iWhoInCheck);
  add_singlemove_promotion(iPos, iDest, PROMOTION_KNIGHT, iWhoInCheck);
  add_singlemove_promotion(iPos, iDest, PROMOTION_BISHOP, iWhoInCheck);
}

static void add_multiplemove_promotion_black(const posid iPos, const posid iDest) {

  int iWhoInCheck = isInCheckAfterMove(iPos, iDest, PROMOTION_BLACKPAWN);

  add_singlemove_promotion(iPos, iDest, PROMOTION_BLACKQUEEN, iWhoInCheck);
  add_singlemove_promotion(iPos, iDest, PROMOTION_BLACKROOK, iWhoInCheck);
  add_singlemove_promotion(iPos, iDest, PROMOTION_BLACKKNIGHT, iWhoInCheck);
  add_singlemove_promotion(iPos, iDest, PROMOTION_BLACKBISHOP, iWhoInCheck);
}

/**
 * fill all possible moves for a pawn
 */
static void fill_movesforpawnwhite(const posid iPos, const int iDir) {
  char cPieceResult;
  posid iResult = iPos + iDir;

  cPieceResult = m1_board.m_fields[iResult];
  if (cPieceResult == FIELD_EMPTY) {
    // empty case, no capture move, so move possible

    // add check for promotion moves
    if (ROW144TO64(iPos) == 6) {
      // always add promotion moves to the movelist
      add_multiplemove_promotion_white(iPos, iResult);
    } else {
      add_singlemove(iPos, iResult);
    }

    // check if we can do a double move
    if (ROW144TO64(iPos) == 1) {
      if (m1_board.m_fields[iResult + iDir] == FIELD_EMPTY) {
        add_singlemove(iPos, iResult + iDir);
      }
    }
  }

  cPieceResult = m1_board.m_fields[iResult - 1];
  if (cPieceResult != FIELD_BORDER) {
    if (cPieceResult != FIELD_EMPTY) {
      // we will look if we can capture
      // TODO: optimize by giving the bit as parameter
      if (cPieceResult & (LOWERCASEBIT)) {
        // we know that the allowed colour for capture is the colour of the destination piece

        // add check for promotion moves
        if (ROW144TO64(iPos) == 6) {
          add_multiplemove_promotion_white(iPos, iResult - 1);
        } else {
          add_singlemove(iPos, iResult - 1);
        }
      }
    } else if (iResult - 1 == m1_board.m_epfield) {
      // ep move
      add_singlemove(iPos, iResult - 1);
    }
  }

  cPieceResult = m1_board.m_fields[iResult + 1];
  if (cPieceResult != FIELD_BORDER) {
    if (cPieceResult != FIELD_EMPTY) {
      // we will look if we can capture
      // TODO: optimize by giving the bit as parameter
      if (cPieceResult & (LOWERCASEBIT)) {
        // we know that the allowed colour for capture is the colour of the destination piece

        // add check for promotion moves
        if (ROW144TO64(iPos) == 6) {
          add_multiplemove_promotion_white(iPos, iResult + 1);
        } else {
          add_singlemove(iPos, iResult + 1);
        }
      }
    } else if (iResult + 1 == m1_board.m_epfield) {
      // ep move
      add_singlemove(iPos, iResult + 1);
    }
  }
}

/**
 * fill all possible moves for a pawn
 */
static void fill_movesforpawnblack(const posid iPos, const int iDir) {
  char cPieceResult;
  posid iResult = iPos + iDir;

  cPieceResult = m1_board.m_fields[iResult];
  if (cPieceResult == FIELD_EMPTY) {
    // empty case, no capture move, so move possible

    // add check for promotion moves
    if (ROW144TO64(iPos) == 1) {
      // always add promotion moves to the movelist
      add_multiplemove_promotion_black(iPos, iResult);
    } else {
      add_singlemove(iPos, iResult);
    }

    // check if we can do a double move
    if (ROW144TO64(iPos) == 6) {
      if (m1_board.m_fields[iResult + iDir] == FIELD_EMPTY) {
        add_singlemove(iPos, iResult + iDir);
      }
    }
  }

  cPieceResult = m1_board.m_fields[iResult - 1];
  if (cPieceResult != FIELD_BORDER) {
    if (cPieceResult != FIELD_EMPTY) {
      // we will look if we can capture
      // TODO: optimize by giving the bit as parameter

      if (!(cPieceResult & (LOWERCASEBIT))) {
        // we know that the allowed colour for capture is the colour of the destination piece

        // add check for promotion moves
        if (ROW144TO64(iPos) == 1) {
          // always add promotion moves to the movelist
          add_multiplemove_promotion_black(iPos, iResult - 1);
        } else {
          add_singlemove(iPos, iResult - 1);
        }
      }
    } else if (iResult - 1 == m1_board.m_epfield) {
      // ep move
      add_singlemove(iPos, iResult - 1);
    }
  }

  cPieceResult = m1_board.m_fields[iResult + 1];
  if (cPieceResult != FIELD_BORDER) {
    if (cPieceResult != FIELD_EMPTY) {
      // we will look if we can capture
      // TODO: optimize by giving the bit as parameter
      if (!(cPieceResult & (LOWERCASEBIT))) {
        // we know that the allowed colour for capture is the colour of the destination piece

        // add check for promotion moves
        if (ROW144TO64(iPos) == 1) {
          // always add promotion moves to the movelist
          add_multiplemove_promotion_black(iPos, iResult + 1);
        } else {
          add_singlemove(iPos, iResult + 1);
        }
      }
    } else if (iResult + 1 == m1_board.m_epfield) {
      // ep move
      add_singlemove(iPos, iResult + 1);
    }
  }
}

/**
 * is the player in check after the move is executed  (return 1)
 * also do check if the current move causes a check  (return 2)
 */
static int isInCheckAfterMove(const posid iFrom, const posid iTo, const t_promotion cFlag) {
  int iCheck = 0;

  // do we need to consider invalid moves?
  t_alg1_board m1_boardbuffer;

  t_boolean isWhite = (m1_board.m_fields [iFrom] & (LOWERCASEBIT)) ? myFALSE : myTRUE;

  // store board position, for later restore
  memcpy(&m1_boardbuffer, &m1_board, sizeof (m1_board));

  if (isWhite) {
    b1_apply_movewhite(iFrom, iTo, cFlag);
  } else {
    b1_apply_moveblack(iFrom, iTo, cFlag);
  }

  iCheck = alg1_isInCheck(isWhite);

  if (!iCheck) {
    if (alg1_isInCheck((t_boolean) (!isWhite))) iCheck = 2;
  }

  // restore old position
  memcpy(&m1_board, &m1_boardbuffer, sizeof (m1_board));

  return (iCheck);
}

/**
 * is the position in check
 */
t_boolean alg1_isInCheck(const t_boolean bEvaluateForWhite) {
  posid iKingPos = bEvaluateForWhite ? m1_board.m_whiteking : m1_board.m_blackking;

  return isUnderAttack(iKingPos, bEvaluateForWhite);
}

/**
 * test if after executing the move, the king is in check
 * the king to test will be of the same colour as the colour moving the piece
 * we will first apply the move, then see if the king jumps to any of the opposing colours
 */
#define pieceseesone(x,y,z) if(m1_board.m_fields[(z)]==((y)?TO_BLACK(x):TO_WHITE(x))) { isAttack++; }
#define pieceseesdirm(x,y,z,a,b,c) if(pieceseesmore((x),(y),(z),(a),(b),(c))) { isAttack++; }
#define pieceopenline(x,y,z,a) if(pieceseesopen((x),(y),(z),(a))) { isAttack++; }

#define pieceseesone_r(x,y,z) if(m1_board.m_fields[(z)]==((y)?TO_BLACK(x):TO_WHITE(x))) { return myTRUE; }
#define pieceseesdirm_r(x,y,z,a,b,c) if(pieceseesmore((x),(y),(z),(a),(b),(c))) { return myTRUE; }
#define pieceopenline_r(x,y,z,a) if(pieceseesopen((x),(y),(z),(a))) { return myTRUE; }

t_boolean isUnderAttack(const posid iPiecePos, const t_boolean bEvaluateForWhite) {
  // calculate all knights moves
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 25);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 23);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 14);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 10);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 25);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 23);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 14);
  pieceseesone_r(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 10);
  // calculate all rook moves
  pieceseesdirm_r(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, 12, 7);
  pieceseesdirm_r(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, 1, 7);
  pieceseesdirm_r(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, -12, 7);
  pieceseesdirm_r(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, -1, 7);
  // calculate all bishop moves
  pieceseesdirm_r(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, 13, 7);
  pieceseesdirm_r(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, 11, 7);
  pieceseesdirm_r(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, -13, 7);
  pieceseesdirm_r(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, -11, 7);
  // calculate all king moves
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos + 12);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos + 1);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos - 12);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos - 1);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos + 13);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos + 11);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos - 13);
  pieceseesone_r(PIECE_KING, bEvaluateForWhite, iPiecePos - 11);
  // handle pawn moves
  if (bEvaluateForWhite) {
    pieceseesone_r(PIECE_PAWN, bEvaluateForWhite, iPiecePos + 11);
    pieceseesone_r(PIECE_PAWN, bEvaluateForWhite, iPiecePos + 13);
  } else {
    pieceseesone_r(PIECE_PAWN, bEvaluateForWhite, iPiecePos - 11);
    pieceseesone_r(PIECE_PAWN, bEvaluateForWhite, iPiecePos - 13);
  }
  return myFALSE;
}

unsigned int countUnderAttack(const posid iPiecePos, const t_boolean bEvaluateForWhite) {
  unsigned int isAttack = 0;

  // calculate all knights moves
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 25);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 23);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 14);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 10);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 25);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 23);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 14);
  pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 10);
  // calculate all rook moves
  pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, 12, 7);
  pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, 1, 7);
  pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, -12, 7);
  pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_QUEEN, bEvaluateForWhite, -1, 7);
  // calculate all bishop moves
  pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, 13, 7);
  pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, 11, 7);
  pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, -13, 7);
  pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_QUEEN, bEvaluateForWhite, -11, 7);
  // calculate all king moves
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos + 12);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos + 1);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos - 12);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos - 1);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos + 13);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos + 11);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos - 13);
  pieceseesone(PIECE_KING, bEvaluateForWhite, iPiecePos - 11);
  // handle pawn moves
  if (bEvaluateForWhite) {
    pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos + 11);
    pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos + 13);
  } else {
    pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos - 11);
    pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos - 13);
  }

  return (isAttack);
}

/**
 * see if the piece gets attached by a minor piece, needed for evaluation
 */
unsigned int isUnderMinorAttack(const posid iPiecePos, const t_piece cPiece, const t_boolean bEvaluateForWhite) {
  int isAttack = 0;

  if ((cPiece == PIECE_ROOK) || (cPiece == PIECE_QUEEN)) {
    // calculate all knights moves
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 25);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 23);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 14);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos + 10);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 25);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 23);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 14);
    pieceseesone(PIECE_KNIGHT, bEvaluateForWhite, iPiecePos - 10);
  }

  if (cPiece == PIECE_QUEEN) {
    // calculate all rook moves
    pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_ROOK, bEvaluateForWhite, 12, 7);
    pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_ROOK, bEvaluateForWhite, 1, 7);
    pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_ROOK, bEvaluateForWhite, -12, 7);
    pieceseesdirm(iPiecePos, PIECE_ROOK, PIECE_ROOK, bEvaluateForWhite, -1, 7);
  }

  if ((cPiece == PIECE_QUEEN) || (cPiece == PIECE_ROOK)) {
    // calculate all bishop moves
    pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_BISHOP, bEvaluateForWhite, 13, 7);
    pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_BISHOP, bEvaluateForWhite, 11, 7);
    pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_BISHOP, bEvaluateForWhite, -13, 7);
    pieceseesdirm(iPiecePos, PIECE_BISHOP, PIECE_BISHOP, bEvaluateForWhite, -11, 7);
  }

  if (cPiece != PIECE_PAWN) {
    // handle pawn moves
    if (bEvaluateForWhite) {
      pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos + 11);
      pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos + 13);
    } else {
      pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos - 11);
      pieceseesone(PIECE_PAWN, bEvaluateForWhite, iPiecePos - 13);
    }
  }

  return (isAttack);
}

// --------------------------------------------------------------------

static t_boolean pieceseesmore(const posid iPiecePos, const t_piece cPieceToLookFor1, const t_piece cPieceToLookFor2,
        const t_boolean bEvaluateForWhite, const int iJump, const int iCount) {
  int i;
  t_piece cPiece1;
  t_piece cPiece2;
  t_piece cPieceResult;

  int iResult = iPiecePos;

  if (bEvaluateForWhite) {
    cPiece1 = (t_piece) (cPieceToLookFor1 + LOWERCASEBIT);
    cPiece2 = (t_piece) (cPieceToLookFor2 + LOWERCASEBIT);
  } else {
    cPiece1 = cPieceToLookFor1;
    cPiece2 = cPieceToLookFor2;
  }

  for (i = 0; i < iCount; i++) {
    // repeat all moves in a direction
    iResult += iJump;
    cPieceResult = m1_board.m_fields[iResult];
    if (cPieceResult == FIELD_EMPTY) {
      // empty case, no capture move, so move possible
    } else if (cPieceResult == cPiece1) {
      return myTRUE;
    } else if (cPieceResult == cPiece2) {
      return myTRUE;
    } else {
      // if we reach the border or another piece, we do not continue anymore
      break;
    }
  }

  return myFALSE;
}

// eof
