/*
 * File:   alg2_search.h
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 *
 * Created on 6 mars 2019, 00:19
 */

#ifndef __ALG2_BOARD_H__
#define __ALG2_BOARD_H__

#include <vector>

#include "board.h"

typedef signed char rank_t;
typedef signed char col_t;

// forward declaration
class alg2_piece;
class alg2_move;

typedef std::vector<alg2_move> alg2_move_vec_t;

/**
 * class alg2_boardpos
 */
class alg2_boardpos {
public:

  alg2_boardpos(rank_t r = -1, col_t c = -1) : rank(r), col(c) {
  }

  alg2_boardpos(const alg2_boardpos& orig);

  virtual ~alg2_boardpos() {
  }

  void setPos(const rank_t r, const col_t c) {
    rank = r;
    col = c;
  }

  rank_t getRank() const {
    return rank;
  }

  col_t getCol() const {
    return col;
  }

  bool isValid() const;

private:
  rank_t rank;
  col_t col;
};

/**
 * class alg2_move
 */
class alg2_move {
public:

  alg2_move(const alg2_boardpos& f, const alg2_boardpos& t) :
  from(f), to(t), isCapture(false), isCheck(false), isPromotion(false) {
  }
  alg2_move(const alg2_move& orig);
  virtual ~alg2_move();

  alg2_boardpos from;
  alg2_boardpos to;
private:
  bool isCapture;
  bool isCheck;
  bool isPromotion;
};

/**
 * class alg2_movelist, list of moves for piece or for board
 */
class alg2_movelist {
public:

  alg2_movelist();
  alg2_movelist(const alg2_movelist& orig);
  virtual ~alg2_movelist();

  void addMove(const alg2_move& move);
  void addMoves(const alg2_move_vec_t& move);
  alg2_move getMove(const int iSelectedMove);
  int size() const;

private:
  alg2_move_vec_t moves;
};

/**
 * class alg2_board, lazy implementation
 * most data is only filled upon call of initialisation function
 * as board is passed along in different function calls
 * TODO: overload operator=
 */
class alg2_board {
public:

  alg2_board() {
  }
  alg2_board(const alg2_board& orig);
  virtual ~alg2_board();

  // initializers
  void copyboardfromui();
  int generatemoves();

  // functions
  bool hasMoves();
  int getMovesFound();
  alg2_move getSelectedMove(const int iMove);
  void addMoves(const alg2_move_vec_t& m);

  // access members
  t_piece getPieceAt(const rank_t rank, const col_t col) const;
  t_piece getPieceAt(const alg2_boardpos& pos) const;
private:
  alg2_movelist moves;
  alg2_piece *boardPiece[8][8];
};

/**
 * piece class
 */
class alg2_piece {
public:

  // TODO: field-border as default value?

  alg2_piece(const t_piece id = FIELD_BORDER) : pieceid(id) {
  }
  alg2_piece(const alg2_piece& orig);
  virtual ~alg2_piece();

  virtual int generateMoves(const alg2_board& board, const rank_t rank, const col_t col);
  virtual void invertColour();

  t_piece getPiece() {
    return pieceid;
  }

  bool tryMove(const alg2_board& board, alg2_move_vec_t& moves,
          const alg2_boardpos& orig,
          const rank_t incRank, const col_t incCol);
  void tryDirection(const alg2_board& board, alg2_move_vec_t& moves,
          const alg2_boardpos& orig,
          const rank_t incRank, const col_t incCol);

protected:
  t_piece pieceid;
  alg2_movelist moves;
};

class alg2_pieceEmpty : public alg2_piece {
public:

  alg2_pieceEmpty(t_piece id = FIELD_EMPTY) : alg2_piece(id) {
  }

  ~alg2_pieceEmpty() override {
  }
};

class alg2_piecePawn : public alg2_piece {
public:

  alg2_piecePawn(t_piece id = PIECE_PAWN) : alg2_piece(id) {
  }

  ~alg2_piecePawn() override {
  }

  int generateMoves(const alg2_board& board, const rank_t rank, const col_t col) override;
  void invertColour() override;
};

class alg2_piecePawnBlack : public alg2_piecePawn {
public:

  alg2_piecePawnBlack(t_piece id = BLACK_PAWN) : alg2_piecePawn(id) {
  }

  ~alg2_piecePawnBlack() override {
  }
};

class alg2_pieceRook : public alg2_piece {
public:

  alg2_pieceRook(t_piece id = PIECE_ROOK) : alg2_piece(id) {
  }

  int generateMoves(const alg2_board& board, const rank_t rank, const col_t col) override;
};

class alg2_pieceRookBlack : public alg2_pieceRook {
public:

  alg2_pieceRookBlack(t_piece id = BLACK_ROOK) : alg2_pieceRook(id) {
  }

  ~alg2_pieceRookBlack() override {
  }
};

class alg2_pieceBishop : public alg2_piece {
public:

  alg2_pieceBishop(t_piece id = PIECE_BISHOP) : alg2_piece(id) {
  }

  int generateMoves(const alg2_board& board, const rank_t rank, const col_t col) override;
};

class alg2_pieceBishopBlack : public alg2_pieceBishop {
public:

  alg2_pieceBishopBlack(t_piece id = BLACK_BISHOP) : alg2_pieceBishop(id) {
  }

  ~alg2_pieceBishopBlack() override {
  }
};

class alg2_pieceKnight : public alg2_piece {
public:

  alg2_pieceKnight(t_piece id = PIECE_KNIGHT) : alg2_piece(id) {
  }

  int generateMoves(const alg2_board& board, const rank_t rank, const col_t col) override;
};

class alg2_pieceKnightBlack : public alg2_pieceKnight {
public:

  alg2_pieceKnightBlack(t_piece id = BLACK_KNIGHT) : alg2_pieceKnight(id) {
  }

  ~alg2_pieceKnightBlack() override {
  }
};

class alg2_pieceQueen : public alg2_piece {
public:

  alg2_pieceQueen(t_piece id = PIECE_QUEEN) : alg2_piece(id) {
  }

  int generateMoves(const alg2_board& board, const rank_t rank, const col_t col) override;
};

class alg2_pieceQueenBlack : public alg2_pieceQueen {
public:

  alg2_pieceQueenBlack(t_piece id = BLACK_KNIGHT) : alg2_pieceQueen(id) {
  }

  ~alg2_pieceQueenBlack() override {
  }
};

class alg2_pieceKing : public alg2_piece {
public:

  alg2_pieceKing(t_piece id = PIECE_KING) : alg2_piece(id) {
  }

  int generateMoves(const alg2_board& board, const rank_t rank, const col_t col) override;
};

class alg2_pieceKingBlack : public alg2_pieceKing {
public:

  alg2_pieceKingBlack(t_piece id = BLACK_KING) : alg2_pieceKing(id) {
  }

  ~alg2_pieceKingBlack() override {
  }
};

#endif /* __ALG2_BOARD_H__ */
// eof
