/*---------------------------------------------------------------------+
|                                                                      |
|            belofte.cpp - A Promising Chess Program                   |
|                                                                      |
+----------------------------------------------------------------------+
|   @author Yves De Billoëz                                            |
|   @since 01-oct-2004                                                 |
+----------------------------------------------------------------------+
|   Copyright (C) 2004-2020 Yves De Billoëz                            |
|   This program is distributed under the GNU GPL Version 2, June 1991 |
|   General Public License. See license.html for more information.     |
+----------------------------------------------------------------------*/

/**
 * The main program, doing all coordnation stuff in between the rest
 * @see usercmd.cpp and board.cpp
 */

#include "belofte.h"

#include "usercmd.h"
#include "YThread.h"
#include "board.h"
#include "search.h"
#include "evaluation.h"

void * EngineThread(YThreadFunParam_t pInit);
void * PonderThread(YThreadFunParam_t pInit);
void * TimeObserverThread(YThreadFunParam_t pInit);

/*----------------------------------------------------------------------+
 |      data structures                                                 |
 +----------------------------------------------------------------------*/

uistate_t uistate;
uimode_t uimode;
game_t game;
level_t level;
search_t search;
board_t m_board;

YThread_t tidEngine = 0;
YThreadMutex_t tidEngineMutex = YThread_MUTEXINIT;
YThread_t tidTime = 0;
YThreadMutex_t tidTimeMutex = YThread_MUTEXINIT;

/**
 * Main entry point
 * <b>contains main loop
 * @param argc The number of parameters
 * @param argv The strings pointing to each parameter
 */

int main(int argc, char *argv[]) {
  int rcLoaded = 0;

  usercmd_constructor(); // sets SIGINT
#if defined(FEATURE_XBOARD_INTERFACE)
  if ((argc > 1) &&
          ((strcmp(argv[1], "xboard") == 0)
          || (strcmp(argv[1], "-xboard") == 0)
          || (strcmp(argv[1], "--xboard") == 0)
          || (strcmp(argv[1], "/xboard") == 0))) {
    // do not print out anything if the application gets booted with an xboard parameter as athe
    // first parameter
    uimode.xboard_mode = myTRUE;
    uimode.engine_mode = myTRUE;
  } else {
    // TODO: handle other parameters in case first is not xboard mode
    uimode.xboard_mode = myFALSE;
    uimode.engine_mode = myFALSE;
    uistate.showprompt = myTRUE;
  }
#else
  uimode.xboard_mode = myFALSE;
  uimode.engine_mode = myFALSE;
  uimode.uci_mode = myTRUE;
#endif

  uimode.debug = myFALSE;
  uimode.fpDebugOutput = NULL;

  level.search_algoritme = DEFAULT_ALG;
#if defined(FEATURE_QS)
  level.quiescence = myTRUE;
#else
  level.quiescence = myFALSE;
#endif
  srand((unsigned) time(0));

  board_constructor();
  search_constructor();
  eval_constructor();

#if defined(FEATURE_XBOARD_INTERFACE)
  board_delayedconstructor();
  search_delayedconstructor();
  eval_delayedconstructor();
  usercmd_delayedconstructor();

  cmd_new("", "nobd"); // TODO, this should come after reading command line

  if (argc > 3) {
    // do not print out anything if the application gets booted with an xboard parameter as athe
    // first parameter
    // TODO: allow to omit first parameter
    if ((strcmp(argv[2], "-i") == 0)
            || (strcmp(argv[2], "/i") == 0)) {
      rcLoaded += usercmd_execute("@", argv[3]);
    }
  } else {
    rcLoaded += usercmd_execute("-i", "/etc/belofte/belofte.rc");
    rcLoaded += usercmd_execute("-i", "~/belofte.rc");
    rcLoaded += usercmd_execute("-i", "belofte.rc");
  }
#else
  rcLoaded += usercmd_execute("-i", "belofte.rc");
  cmd_debug("", "on");
#endif

  if (!rcLoaded) {
    // if rc file is loaded, about command and new command are omitted
    if (!uimode.engine_mode) {
      cmd_about("", "");
#if defined(FEATURE_XBOARD_INTERFACE)
      board_print();
#endif
    }
    cmdresp_snd_prompt();
  }

  StartTimeThread();

  for (;;) {
    usercmd_readcommand();
    YThread_Snooze(200);
  }

  return 0;
}

void StartEngineThread(void) {
  YThread_MutexLock(&tidEngineMutex);
  if (tidEngine == 0) {
    tidEngine = YThread_Create(EngineThread, 0, 0);
  }
  uistate.engine_running = myTRUE;
  YThread_MutexUnLock(&tidEngineMutex);
}

void StopEngineThread(void) {
  YThreadRc_t rcode = 0;

  YThread_MutexLock(&tidEngineMutex);
  uistate.engine_running = myFALSE;
  if (tidEngine != 0) {
    YThread_Destroy(tidEngine, &rcode);
    tidEngine = 0;
  }
  YThread_MutexUnLock(&tidEngineMutex);
}

void * EngineThread(UNUSED YThreadFunParam_t pInit) {
  int iSearchResult;
  char szMove[MAX_SANMOVE_LENGTH];
  char szResult[MAX_LINE_LENGTH];

  while (1) {
    if ((uistate.engine_running == myTRUE)
            && (search.engine_searching == myFALSE)
            && (uistate.computer_engines[game.curply % 2])) {
      iSearchResult = search_root(szMove, sizeof (szMove), szResult, sizeof (szResult));
      uistate.engine_running = myFALSE;
      cmdresp_snd_move(szMove, iSearchResult, szResult);
      // in between each computer move, we print the board
      if (!uimode.engine_mode) {
        board_print();
        cmdresp_snd_prompt();
      }
      if (!uistate.computer_engines[game.curply % 2]) {
        // if next player is not computer engine, stop thread
        YThread_Terminate(&tidEngine);
      } else {
        // do small pause in between moves and continue
        YThread_Snooze(500);
        uistate.engine_running = myTRUE;
      }
    } else {
      // there is no computer engine that plays, so return to user
      YThread_Snooze(500);
    }
  }
  YThread_MutexUnLock(&tidTimeMutex);
}

void * PonderThread(UNUSED YThreadFunParam_t pInit) {
  while (1) {
    if ((uistate.engine_running)
            && (!(uistate.computer_engines[game.curply % 2]))
            && (uistate.computer_engines[(game.curply + 1) % 2])
            && (uistate.ponder_mode)) {
      // the engine is not thinking, ponder mode is on
      // and not both engines are set to think
      uistate.pondering = myTRUE;
      puts(" 0 0 0 0 (pondering not working yet)");
    }

    YThread_Snooze(5000);
  }
}

/** analyze each half second if cancel is required
 */
void * TimeObserverThread(UNUSED YThreadFunParam_t pInit) {
  time_t end_t;
  t_boolean moveNow = myFALSE;

  while (1) {
    if (search.engine_searching == myTRUE) {
      time(&end_t);
      uistate.thinkingtime = difftime(end_t, uistate.start_t);
      if (level.level_mode != LEVELMODE_NOTIME) {
        // not fixed depth
        if (level.level_mode == LEVELMODE_TIMEPERMOVE) {
          // fixed time per move
          if ((int) uistate.thinkingtime >= level.max_movetime) {
            moveNow = myTRUE;
          }
        } else {
          // fixed time per game - LEVELMODE_TIMEKO
          // incremental clocks and/or time per number of moves- LEVELMODE_TIMEANDMOVES
          if ((int) uistate.thinkingtime >= level.estimated_movetime) {
            moveNow = myTRUE;
          }
        }
        if (moveNow == myTRUE) {
          search.search_cancel = myTRUE;
          moveNow = myFALSE;
        }
      }
    }
    YThread_Snooze(100);
  }
}

void StartTimeThread(void) {
  YThread_MutexLock(&tidTimeMutex);
  if (tidTime == 0) {
    tidTime = YThread_Create(TimeObserverThread, 0, 0);
  }
  YThread_MutexUnLock(&tidTimeMutex);
}

void StopTimeThread(void) {
  YThreadRc_t rcode = 0;

  YThread_MutexLock(&tidTimeMutex);
  if (tidTime != 0) {
    YThread_Destroy(tidTime, &rcode);
    tidTime = 0;
  }
  YThread_MutexUnLock(&tidTimeMutex);
}
// eof
