/***********************************************************************************
 * File: belofte.h
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 * This is the main include file, needs to be included before any other include.
 ***********************************************************************************/

#if !defined(__BELOFTE_H__)
#define __BELOFTE_H__

// attention, when adding conditional defines,  please modify cmd_about
//#define FEATURE_UCI_INTERFACE 1
//#define FEATURE_XBOARD_ICS 1
//#define FEATURE_FENINVERT 1
//#define FEATURE_CHESS_VARIANT 1
#define FEATURE_QS 1
//#define FEATURE_USEBOOK 1 // define in case we want to manage opening books

#if defined(__DEBUG__) && !defined(_DEBUG)
#define _DEBUG
#endif

#if !defined(FEATURE_UCI_INTERFACE) && !defined(FEATURE_XBOARD_INTERFACE)
// define at least one of the possible interfaces
#define FEATURE_XBOARD_INTERFACE 1
#endif

#if defined(FEATURE_XBOARD_INTERFACE) && defined(FEATURE_UCI_INTERFACE)
#error "Cannot define both xboard and UCI"
#endif

#if defined(FEATURE_UCI_INTERFACE) && defined(FEATURE_XBOARD_ICS)
#error "Cannot define both UCI and ICS"
#endif

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// for srand(time(0)) call
#include <time.h>
#include <limits.h>

#if !defined(_MSC_VER)
#include <unistd.h>
#include <pthread.h>
#endif

// required for select
#include <sys/types.h>

#include <signal.h>

#if !defined(_MSC_VER)
#include <dirent.h>
#endif

#if defined(_MSC_VER) || defined(_WIN32)

#pragma warning( disable : 4100 )
#pragma warning( disable : 4996 )
#define UNUSED
#include <io.h>
#include <Winsock2.h>

#endif

// --------------------------------------------------------------------

#if defined(__i386) && !defined(__LITTLE_ENDIAN__)
#define __LITTLE_ENDIAN__ 1
#endif

#if defined(_MSC_VER) && !defined(__LITTLE_ENDIAN__) && !defined(__BIG_ENDIAN__)
#define __LITTLE_ENDIAN__ 1
#endif

#if defined(__x86_64__)
#define __LITTLE_ENDIAN__ 1
#endif

#if !defined(__BIG_ENDIAN__) && !defined(__LITTLE_ENDIAN__)
#error Do not know the endianess of this architecture
#endif

#if !defined(_MSC_VER)
#if !defined(_POSIX_C_SOURCE)
#define _POSIX_C_SOURCE 1
#endif
#if !defined(_POSIX_SOURCE)
#define _POSIX_SOURCE 1
#endif
#if !defined(__USE_POSIX)
extern int fileno(FILE *__stream);
#endif
#endif

#if !defined(_MSC_VER) && !defined(_WIN32)
#define UNUSED __attribute__((unused))
#endif

#if __cplusplus < 201103L
#define override
#endif

// --------------------------------------------------------------------

#define MYNAME      "Belofte"
#define MYVERSION   "0.9.14"
#define MYFULLNAME  MYNAME " " MYVERSION " (computer)"
#define MYAUTHOR    "Yves De Billoëz"

#define BOOKNAME    "belofte-" MYVERSION ".book"
#define PGNNAME     "belofte-" MYVERSION ".pgn"
#define LOGNAME     "belofte-" MYVERSION ".log"

#define MAX_DATANAME_LENGTH 1024
#define MAX_LINE_LENGTH 256
#define MAX_FEN_LENGTH 128
#define MAX_DEBUGFILE_LENGTH 128

#define MAX_GAME_LENGTH 512
#define MAX_MOVES_PER_POSITION 200
#define MAX_SEARCH_DEPTH 64
#define MAX_PIECES_ON_BOARD 32

// 0-0-0+, e4xd3ep#, e7xf8Q+
#define MAX_SANMOVE_LENGTH    9
// e1g1+, e7e8Q#
#define MAX_COORDMOVE_LENGTH 7

#define DEFAULT_MAXDEPTH 40
#define DEFAULT_TIMEPERMOVE 10
#define DEFAULT_ALG 1

#define LOWERCASEBIT ('a' - 'A')
#define UPPERCASEBIT (255 - LOWERCASEBIT)

// --------------------------------------------------------------------

#define SCORE_WON 9999
#define SCORE_MATEINONE 1
#define SCORE_ALMOST_WON 8888
#define SCORE_ABORT 6969
#define SCORE_UNDEFINED 7777
#define SCORE_RESIGN 14
#define SCORE_DRAW 0

// --------------------------------------------------------------------

enum tBoolean {
  myFALSE = 0, myTRUE = 1
};

enum tColour {
  COLOUR_WHITE = 0, COLOUR_BLACK = 1
};

enum tLevelMode {
  LEVELMODE_NOTIME = 0,
  LEVELMODE_TIMEANDMOVES,
  LEVELMODE_TIMEKO,
  LEVELMODE_TIMEPERMOVE,
  LEVELMODE_NPS,
  LEVELMODE_NODES,
  LEVELMODE_PERFTEST
};

typedef enum tLevelMode levelmode_t;

// TODO: change to flag_t, ...
typedef unsigned char t_flag;
typedef enum tColour t_colour;
typedef int t_ply;
typedef unsigned char byte;
typedef unsigned short int posid;
typedef enum tBoolean t_boolean;
typedef float t_score;
typedef int t_depth;

/** userinterface mode
 */
struct st_uimode {
  t_boolean ics_mode; // set if on chess server
  t_boolean post_mode; // send output while thinking
  int protover; // interface towards xboard set
  t_boolean engine_mode; // no prompt, xboard compatible interface if set
  t_boolean xboard_mode; // xboard command received
  t_boolean uci_mode; // uci command received
  t_boolean processing_mode; // set if no display is to be given on cmd_new
  t_boolean debug; // we are in debug mode
  FILE * fpDebugOutput; // handle for debug file
  char szDebugFile[MAX_DEBUGFILE_LENGTH]; // name of debug file
};

typedef struct st_uimode uimode_t;

/** userinterface state of engine, is engine working or not
 */
struct st_uistate {
  t_boolean engine_running; // engine running if set
  t_boolean computer_engines[2]; // = { 0, 1 }; // default to computer engine for black
  t_boolean pondering; // actually pondering at the moment
  t_boolean ponder_mode; // do ponder during client thinking
  t_boolean showprompt; // is the prompt to be shown to user
  int searchnodes; // total number of nodes searched
  int searchmaxdepth; // the maximum depth reached during search
  int perftnodes; // total number of leaf nodes searched, used for perft
  t_depth iCurrentDepth;
  double thinkingtime; // total seconds of thinking time
  time_t start_t; // time when searching started
};

typedef struct st_uistate uistate_t;

/** level information
 */
struct st_level {
  int search_algoritme;
  // 1 - original belofte 0.2.7 till 0.9.6
  // 2 - belofte 0.9.15 and up
  t_boolean quiescence;
  t_boolean random_mode; // randomize game if set

  int maxdepth;
  levelmode_t level_mode; // define what kind of level is taken

  int game_timeleft; // number of seconds left, for game
  int moves_perperiod; // number of moves per persiod, or zero if KO
  int moves_leftinperiod; // number of moves left for this period, or zero if KO
  int additional_secs; // incremental clock, increments section

  int estimated_gametime; // in case of ficherclock, gametime will increase for each move
  int estimated_movetime; // initial time calculated for move, adjusted during search
  int max_movetime; // maximum time allowed for move, after this, cancel search

  int opponent_time; // keep track of opponent time in order to know if he is in timetrouble
};

typedef struct st_level level_t;

// --------------------------------------------------------------------

enum {
  A1 = 0, B1, C1, D1, E1, F1, G1, H1,
  A2, B2, C2, D2, E2, F2, G2, H2,
  A3, B3, C3, D3, E3, F3, G3, H3,
  A4, B4, C4, D4, E4, F4, G4, H4,
  A5, B5, C5, D5, E5, F5, G5, H5,
  A6, B6, C6, D6, E6, F6, G6, H6,
  A7, B7, C7, D7, E7, F7, G7, H7,
  A8, B8, C8, D8, E8, F8, G8, H8
};

// --------------------------------------------------------------------

#define MOVERESULT_OK    0
#define MOVERESULT_DRAW_LACKMAT 1
#define MOVERESULT_DRAW_THREEFOLD  2
#define MOVERESULT_DRAW_50  3
#define MOVERESULT_NIL   -1
#define MOVERESULT_ERROR -99

// --------------------------------------------------------------------

#if !defined(MIN)
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#if !defined(MAX)
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif

// --------------------------------------------------------------------

void StartEngineThread(void);
void StopEngineThread(void);
void StartTimeThread(void);
void StopTimeThread(void);
void StartPonderThread(void);
void StopPonderThread(void);

#endif // def __BELOFTE_H__

// eof
