/***********************************************************************************
 * File: board.cpp
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#include "belofte.h"

#include "board.h"
#include "usercmd.h"
#include "search.h"

#include "YUtils.h"

#include "alg1_moves.h"
#include "alg1_board.h"
#include "alg1_board.h"
#include "alg1_search.h"

// --------------------------------------------------------------------

extern board_t m_board;
extern t_alg1_moves m1_movelist;
extern game_t game;
extern int busy_book_creating;
extern t_alg1_board m1_boardmapper;

// --------------------------------------------------------------------

hashkey_t hashmultiplexer[HASH_PIECELIST_SIZE][8][8];
int hashPieceindex[127];
static hashkey_t getRandom(void);

// --------------------------------------------------------------------

static void board_setpiece(const posid iIndex, const t_piece cPiece);
static void board_setinvpiece(const posid iIndex, const t_piece cPiece);
static t_piece board_getpiece(const posid iIndex);

static void board_sanambigious(char *szDest, const t_piece cPiece, const posid iFrom, const posid iTo);
static void board_sanmovedecorate(char *szDest, const t_piece cPieceMoved);

// board layout is left bottum index = 0, right bottom = 7, left top = 56, right top = 63
// board element is m_field[column][row]

// --------------------------------------------------------------------

/**
 * initial constructor, should return asap
 */
void board_constructor(void) {
  b1_constructor();
}

/**
 * initialize zobrist keys
 * TODO: random sequence
 */
void board_delayedconstructor(void) {
  int i;
  int iRow, iColumn;

  // piece keys
  for (i = 0; i < HASH_PIECELIST_SIZE; i++) {
    for (iColumn = 0; iColumn < 8; iColumn++)
      for (iRow = 0; iRow < 8; iRow++)
        hashmultiplexer[i][iColumn][iRow] = getRandom();
  }

  // special moves
  for (i = 0; i < HP_SPECIAL1_L; i++)
    hashmultiplexer[HP_SPECIAL1][0][i] = getRandom();

  // conversion table for piece letters to piece indexes
  hashPieceindex[PIECE_PAWN] = HP_WP;
  hashPieceindex[PIECE_BISHOP] = HP_WB;
  hashPieceindex[PIECE_KNIGHT] = HP_WN;
  hashPieceindex[PIECE_ROOK] = HP_WR;
  hashPieceindex[PIECE_QUEEN] = HP_WQ;
  hashPieceindex[PIECE_KING] = HP_WK;
  hashPieceindex[BLACK_PAWN] = HP_BP;
  hashPieceindex[BLACK_BISHOP] = HP_BB;
  hashPieceindex[BLACK_KNIGHT] = HP_BN;
  hashPieceindex[BLACK_ROOK] = HP_BR;
  hashPieceindex[BLACK_QUEEN] = HP_BQ;
  hashPieceindex[BLACK_KING] = HP_BK;

  b1_delayedconstructor();
}

/**
 * Calculate hash based on board position
 * Byte 0: bits 1-7 capture count - 2 (kings)
 *         bit 0 player to move
 * We are using 64 - 3 bits for hash, to ease debugging for now
 * TODO: increase to 64 bits
 * @return hash of current board
 */
hashkey_t board_calchash(void) {
  int iRow, iColumn;
  int iCountPieces = -2;
  hashkey_t hash = HASH_INIT;
  for (iColumn = 0; iColumn < 8; iColumn++)
    for (iRow = 0; iRow < 8; iRow++) {
      if (m_board.m_fields[iColumn][iRow] != FIELD_EMPTY) {
        hash = hash ^ hashmultiplexer[hashPieceindex[m_board.m_fields[iColumn][iRow]]]
                [iColumn][iRow];
        iCountPieces++;
      }
    }

  if (m_board.m_piecemoved[COLOUR_WHITE][1] == myFALSE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_WHITECASTLE_S];
  if (m_board.m_piecemoved[COLOUR_WHITE][0] == myFALSE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_WHITECASTLE_L];
  if (m_board.m_piecemoved[COLOUR_BLACK][1] == myFALSE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_BLACKCASTLE_S];
  if (m_board.m_piecemoved[COLOUR_BLACK][0] == myFALSE)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_BLACKCASTLE_L];
  if (m_board.m_epfield != 0)
    hash = hash ^ hashmultiplexer[HP_SPECIAL1][0][HP_EPCASE];

  // add higher 8 bits
  if (m_board.m_sidetomove == COLOUR_BLACK)
    hash = hash | 0x0100000000000000ULL;

  hash = hash | (((hashkey_t) (30 - iCountPieces)) << 57);

  return hash;
}

/** return hashkey, as rand is giving by definition
 * result in between 0 and 32768, do not take the risk of zero bits
 * and shift possibly getting bits falling off
 * specification is only giving 15 significant bits, as we do not need
 * upper bits, take this into account
 */
static hashkey_t getRandom(void) {
  hashkey_t r = (rand() & 0x8FFFU); // ensure same result independent of platform
  r = (r << 15) + (rand() & 0x8FFFU);
  r = (r << 15) + (rand() & 0x8FFFU);
  r = (r << 15) + (rand() & 0x8FFFU);
  r = r & 0x00FFFFFFFFFFFFFFULL; // strip off higher 8 bits
  return r;
}
// --------------------------------------------------------------------

void board_print(void) {
  int iRow, iColumn;
  int c;
  char szBuffer[MAX_DATANAME_LENGTH];
  char szBufferLine[MAX_DATANAME_LENGTH];

  if (!busy_book_creating) {
    // do not print the board position during book creation, speedup
    // for creating book faster
    if (game.curply < MAX_GAME_LENGTH - 1) {
      snprintf(szBuffer, sizeof (szBuffer),
              "\n FEN: %s %d %d\n",
              game.m_boardhistory[game.curply].szFEN,
              (int) game.m_boardhistory[game.curply].m_rule50,
              (int) game.m_boardhistory[game.curply].m_movenumber);
      cmdresp_snd(szBuffer);
    }

    for (iRow = 0; iRow < 8; iRow++) {
      cmdresp_snd("+---+---+---+---+---+---+---+---+");
      szBuffer[0] = '\0';
      for (iColumn = 0; iColumn < 8; iColumn++) {
        c = m_board.m_fields[iColumn][7 - iRow];
        sprintf(szBufferLine, "| %c ", c ? (char) c : ' ');
        strcat(szBuffer, szBufferLine);
      }
      sprintf(szBufferLine, "|");
      strcat(szBuffer, szBufferLine);
      cmdresp_snd(szBuffer);
    }
    cmdresp_snd("+---+---+---+---+---+---+---+---+");
  }
}

/**
 * apply standard move
 * move entering is a move in the format from-case, to case, attribute
 * will return a code if there is something special about the move applied
 */
int board_applymove(const char *szCoordMove) {
  posid iFrom;
  posid iTo;
  t_piece cPieceMoved;
  t_piece cPieceTo;
  t_piece cFlag = FIELD_EMPTY;

  // keep past board for undo
  if (game.curply < MAX_GAME_LENGTH - 1) {
    memcpy(&(game.m_boardhistory[game.curply].m_board), &m_board, sizeof (m_board));
    strncpy(game.m_boardhistory[game.curply].szIntMove, szCoordMove, MAX_COORDMOVE_LENGTH);
    board_sanmove(game.m_boardhistory[game.curply].szSANMove, szCoordMove, MAX_SANMOVE_LENGTH);
  }

  if (strlen(szCoordMove) > 3) {
    // very shorthand way to copy move into local working variables
    iFrom = (szCoordMove[0] - 'a') + (szCoordMove[1] - '1') * 8;
    iTo = (szCoordMove[2] - 'a') + (szCoordMove[3] - '1') * 8;

    // get a flag for promotion, ep, check, ....
    if (strlen(szCoordMove) > 4) cFlag = (t_piece) szCoordMove[4];

    cPieceMoved = board_getpiece(iFrom);

    // add ep move, if not double move and switch columns
    if ((cPieceMoved == PIECE_PAWN)
            && ((iTo - iFrom) != 24)
            && (board_getpiece(iTo) == FIELD_EMPTY)
            && (M1_LIJN(iFrom) != M1_LIJN(iTo))) {
      // capture, so clear position
      board_setpiece((posid) (iTo - 8), FIELD_EMPTY);
    }
    if ((cPieceMoved == BLACK_PAWN)
            && ((iFrom - iTo) != 24)
            && (board_getpiece(iTo) == FIELD_EMPTY)
            && (M1_LIJN(iFrom) != M1_LIJN(iTo))) {
      // capture, so clear position
      board_setpiece((posid) (iTo + 8), FIELD_EMPTY);
    }

    cPieceTo = board_getpiece(iTo);
    if ((cPieceTo != FIELD_EMPTY)
            || (cPieceMoved == PIECE_PAWN)
            || (cPieceMoved == BLACK_PAWN)) {
      // capture, pawn move, increase counter
      m_board.m_rule50 = 0;
    } else {
      // increase 50 moves rule counter
      m_board.m_rule50++;
    }

    board_setpiece(iTo, cPieceMoved);
    board_setpiece(iFrom, FIELD_EMPTY);

    if ((cPieceMoved == PIECE_KING) || (cPieceMoved == BLACK_KING)) {
      // in case of a king move, evaluate if there is no castling to be done
      if (abs(iFrom - iTo) == 2) {
        // castling move, determine side
        if (iTo == 2) {
          board_setpiece(0, FIELD_EMPTY);
          board_setpiece(3, PIECE_ROOK);
        } else if (iTo == 6) {
          board_setpiece(7, FIELD_EMPTY);
          board_setpiece(5, PIECE_ROOK);
        } else if (iTo == 58) {
          board_setpiece(56, FIELD_EMPTY);
          board_setpiece(59, BLACK_ROOK);
        } else {
          board_setpiece(63, FIELD_EMPTY);
          board_setpiece(61, BLACK_ROOK);
        }
      }
      // clear castling flags
      if (cPieceMoved == PIECE_KING) {
        m_board.m_piecemoved[COLOUR_WHITE][0] = myTRUE;
        m_board.m_piecemoved[COLOUR_WHITE][1] = myTRUE;
      } else {
        m_board.m_piecemoved[COLOUR_BLACK][0] = myTRUE;
        m_board.m_piecemoved[COLOUR_BLACK][1] = myTRUE;
      }
    }

    // clear castling flags
    if (cPieceMoved == PIECE_ROOK) {
      if (iFrom == 0) m_board.m_piecemoved[COLOUR_WHITE][0] = myTRUE;
      if (iFrom == 7) m_board.m_piecemoved[COLOUR_WHITE][1] = myTRUE;
    }

    if (cPieceMoved == BLACK_ROOK) {
      if (iFrom == 56) m_board.m_piecemoved[COLOUR_BLACK][0] = myTRUE;
      if (iFrom == 63) m_board.m_piecemoved[COLOUR_BLACK][1] = myTRUE;
    }

    // also check promotion
    if (cFlag != FIELD_EMPTY) {
      board_setpiece(iTo, cFlag);
    }

    // and set the board ply
    game.curply++;
    m_board.m_ply = game.curply;
    m_board.m_sidetomove = (t_colour) (m_board.m_ply % 2);

    search_apply_move(szCoordMove);

    if (game.curply < MAX_GAME_LENGTH - 1) {
      board_fullfenstring(game.m_boardhistory[game.curply].szFEN, szCoordMove,
              &(game.m_boardhistory[game.curply].m_rule50),
              &(game.m_boardhistory[game.curply].m_movenumber));
      game.m_boardhistory[game.curply].m_hash = board_calchash();
      board_sanmovedecorate(game.m_boardhistory[game.curply - 1].szSANMove,
              cPieceMoved);
    }
  }

  return board_getmetadata(szCoordMove);
}

/**
 * apply move in san format
 * for this we first have to make a list of all possible moves and then map
 * the san move to this list
 */
int board_sanapplymove(const char *szMove, char *szCoordMove) {
  int i;
  int iMoves;
  t_alg1_moves m1_movesbuffer;

  szCoordMove[0] = '\0';
  memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));
  iMoves = search_fill_movelist(game.curply % 2 ? COLOUR_BLACK : COLOUR_WHITE);
  for (i = 0; i < iMoves; i++) {
    alg1_search_printablemove(szCoordMove, i);
    board_applymove(szCoordMove);
    if (strcmp(game.m_boardhistory[game.curply - 1].szSANMove, szMove) == 0) {
      // there is a match, so this move can be applied
      board_unapplymove();
      break;
    }
    board_unapplymove();
    szCoordMove[0] = '\0';
  }
  // copy back
  memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));

  // if there was a match, the szIntMove has been retained
  if (strlen(szCoordMove) > 0) {
    board_applymove(szCoordMove);
    return MOVERESULT_OK;
  }

  return MOVERESULT_NIL; // no matching move, nothing done
}

/**
 * unapply standard move
 * return MOVERESULT_ERROR if unapply did not succeed
 */
int board_unapplymove(void) {
  // keep past board for undo
  if ((game.curply < MAX_GAME_LENGTH - 1) && (game.curply > 0)) {
    game.curply--;
    m_board.m_ply = game.curply;
    m_board.m_sidetomove = (t_colour) (m_board.m_ply % 2);
    memcpy(&m_board, &(game.m_boardhistory[game.curply].m_board), sizeof (m_board));

    // bring search engine up to date
    b1_init();

    return MOVERESULT_OK;
  }

  return MOVERESULT_ERROR;
}

/**
 * get some statistics about the board, to test threefold position, ...
 * return MOVERESULT_OK: no board statistics of any meaning
 * return MOVERESULT_DRAW_LACKMAT: draw by lack of material
 * return MOVERESULT_DRAW_THREEFOLD: draw by 3 times same position
 * return other value: return from move generation
 */
int board_getmetadata(const char *szFlag) {

  if (strlen(szFlag) == 2) {
    // we have some flag returning from the search engine,
    // return it back after conversion from string
    return -(szFlag[1] - '0');
  }

  if (board_drawinsufficientmat()) return MOVERESULT_DRAW_LACKMAT;
  if (board_positionrepeats(m_board.m_hash) > 2) return MOVERESULT_DRAW_THREEFOLD;
  if (board_draw50moves(50)) return MOVERESULT_DRAW_50;

  return MOVERESULT_OK;
}

/**
 * Test if board Position is a theoretic draw or not
 * TODO: Attention, make sure this is winboard compatible
 * @return 1 in case of draw
 */
int board_drawinsufficientmat() {
  posid i;
  int iPieces = 0;

  for (i = 0; i < 64; i++)
    if (board_getpiece(i) != FIELD_EMPTY) iPieces++;

  if (iPieces < 3)
    return 1; // draw by lack of material

  return 0;
}

/**
 * Find if position occurred before
 * @param position
 * @return number of repeats
 */
int board_positionrepeats(hashkey_t position) {
  int i;
  int iRepeats = 0;
  if (game.firstply < game.curply) {
    for (i = game.firstply; i <= game.curply; i++) {
      if (game.m_boardhistory[i].m_hash == position) iRepeats++;
    }
  }
  return iRepeats;
}

bool board_draw50moves(const int nMovesAlert) {
  return (m_board.m_rule50 > (nMovesAlert * 2) - 1);
}

/*
bool board_testmaterialdiff(hashkey_t position) {
  int i;
  hashkey_t check = position & 0xFE00000000000000ULL;
  int iCount = 0;
  if (game.firstply < game.curply) {
    for (i = game.curply; i <= game.firstply; i--) {
      if ((game.m_boardhistory[i].m_hash & 0xFE00000000000000ULL) == check) {
        // upper bits are equal, so same number of pieces
        if (++iCount>48) return true;
      } else {
        return false;
      }
    }
  }
  return false;
}
 */

/** Set initial position
 */
int board_setupdefaultfen(void) {
  return board_setupfen("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
}

/**
 * setboard function, will initialise by default to this
 * cmd_setboard("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
 * @return unsigned int the index where the EPD or FEN string stops
 */
int board_setupfen(const char *szConstFen) {
  int i, j;
  int iFenLength, iPosLength, iIndex;
  int nMoveNumber, n50Moves;
  char *szPly;
  char *szFen;
  char szFenAlloc[MAX_FEN_LENGTH];
  char szShortFen[MAX_FEN_LENGTH];

  strncpy(szFenAlloc, szConstFen, MAX_FEN_LENGTH);
  szFenAlloc[MAX_FEN_LENGTH - 1] = '\0';
  szFen = szFenAlloc;

  // empty the board first
  for (i = 0; i < 8; i++)
    for (j = 0; j < 8; j++)
      m_board.m_fields[i][j] = FIELD_EMPTY;

  iFenLength = (int) strlen(szFenAlloc);
  iPosLength = iFenLength;

  // skip blanks at start of position
  while (szFen[0] == ' ') szFen++;

  // calculate where the setting for castling are, they are separated by a space
  for (i = 0; i < iFenLength; i++)
    if (szFen[i] == ' ') {
      iPosLength = i;
      break;
    }

  iIndex = 0;
  for (i = 0; i < iPosLength; i++) {
    switch (szFen[i]) {
      case PIECE_PAWN: case PIECE_ROOK: case PIECE_KNIGHT: case PIECE_BISHOP: case PIECE_QUEEN: case PIECE_KING:
      case BLACK_PAWN: case BLACK_ROOK: case BLACK_KNIGHT: case BLACK_BISHOP: case BLACK_QUEEN: case BLACK_KING:
        board_setinvpiece((posid) iIndex, (t_piece) szFen[i]);
        iIndex++;
        break;
      case '/': // go beginning to next row
        iIndex = (iIndex + 7) & 0xF8;
        break;
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8': // increment by number of empty cases
        iIndex += (szFen[i] - '0');
        // if skip to new row, wait until a slash forces to do so
        if (iIndex == (iIndex & 0xF8)) iIndex--;
    }
  }

  // we have done all pieces, now store side to move information
  m_board.m_sidetomove = szFen[iPosLength + 1] == 'w' ? COLOUR_WHITE : COLOUR_BLACK;

  // read castling information, first reset flags
  for (i = 0; i < 2; i++) {
    m_board.m_piecemoved[COLOUR_WHITE][i] = myTRUE;
    m_board.m_piecemoved[COLOUR_BLACK][i] = myTRUE;
  }
  iIndex = iPosLength + 3;
  while (szFen[iIndex] != ' ') {
    // we actually only flag as the rook having moved, it is enough to avoid castling
    if (szFen[iIndex] == 'K') m_board.m_piecemoved[COLOUR_WHITE][1] = myFALSE;
    if (szFen[iIndex] == 'Q') m_board.m_piecemoved[COLOUR_WHITE][0] = myFALSE;
    if (szFen[iIndex] == 'k') m_board.m_piecemoved[COLOUR_BLACK][1] = myFALSE;
    if (szFen[iIndex] == 'q') m_board.m_piecemoved[COLOUR_BLACK][0] = myFALSE;
    iIndex++;
  }

  // read ep information
  iIndex++; // we are now on the ep reading
  if (szFen[iIndex] != '-') {
    m_board.m_epfield = (posid) (szFen[iIndex] - 'a' + ((szFen[iIndex + 1] - '1') * 8));
    iIndex++;
  } else {
    m_board.m_epfield = 0;
  }

  iIndex++;
  strncpy(szShortFen, szFen, iIndex); // store the initial position fen string
  szShortFen[iIndex] = '\0'; // cutting out the space

  iIndex++; // we are now on the 50 rules reading
  szPly = szFen + iIndex;

  if ((szPly[0] >= '0') && (szPly[0] <= '9')) {
    // read 50 move rule information
    // read current move information
    sscanf(szPly, "%d %d", &n50Moves, &nMoveNumber);
    // in case of out of bounds, correct
    if ((nMoveNumber < 0) || (nMoveNumber >= MAX_GAME_LENGTH / 2)) nMoveNumber = 1;
    if ((n50Moves < 0) || (n50Moves > 150)) n50Moves = 0;
  } else {
    // 50 move information is not numeric, we could be in a epd setup
    n50Moves = 0;
    nMoveNumber = 1;
  }

  game.curply = (nMoveNumber * 2) - 2;
  if (m_board.m_sidetomove == COLOUR_BLACK) game.curply++;

  m_board.m_rule50 = n50Moves;
  game.m_boardhistory[game.curply].m_rule50 = n50Moves;

  game.m_boardhistory[game.curply].m_movenumber = nMoveNumber;

  m_board.m_ply = game.curply;
  game.firstply = game.curply;

  // now store FEN in history at good position
  strncpy(game.m_boardhistory[game.curply].szFEN, szShortFen, MAX_FEN_LENGTH);

  m_board.m_hash = board_calchash();
  game.m_boardhistory[game.curply].m_hash = m_board.m_hash;

  // store for prefilling the history, after the first move, this board will be overwritten
  // anyway but other operations might rely on the board being initialised
  memcpy(&(game.m_boardhistory[game.curply].m_board), &m_board, sizeof (m_board));

  // copy the board into the search engine
  search_init();

  return iIndex;
}

/**
 * copy the fen string into the dest
 * copy the ply info
 * move is passed to indicate e.p. field
 */
void board_fullfenstring(char *szDest, const char *szMove,
        t_ply *n50Moves, t_ply *nMoveNumber) {

  board_fenstring(szDest, szMove);
  // 50 move rule and ply
  *n50Moves = m_board.m_rule50;
  *nMoveNumber = (game.curply / 2) + 1;
}

/**
 * copy the short fen string into the dest, short FEN string
 * is a string without number of non pawn moves and without ply indication
 * move is passed to indicate e.p. field
 */
void board_fenstring(char *szDest, const char *szMove) {
  int iRow, iColumn;
  int iEmpty;
  t_piece cPiece;

  szDest[0] = '\0';

  // board values
  for (iRow = 0; iRow < 8; iRow++) {
    iEmpty = 0;
    for (iColumn = 0; iColumn < 8; iColumn++) {
      cPiece = m_board.m_fields[iColumn][7 - iRow];
      if (cPiece == FIELD_EMPTY) {
        iEmpty++;
      } else {
        if (iEmpty) add_char(szDest, (const char) ('0' + iEmpty));
        add_char(szDest, (const char) cPiece);
        iEmpty = 0;
      }
    }
    if (iEmpty) add_char(szDest, (const char) ('0' + iEmpty));
    if (iRow < 7) add_char(szDest, '/');
  }

  // side to move
  add_char(szDest, ' ');
  add_char(szDest, (const char) (game.curply % 2 ? 'b' : 'w'));

  // castling flags
  add_char(szDest, ' ');
  if (!m_board.m_piecemoved[COLOUR_WHITE][1]) add_char(szDest, 'K');
  if (!m_board.m_piecemoved[COLOUR_WHITE][0]) add_char(szDest, 'Q');
  if (!m_board.m_piecemoved[COLOUR_BLACK][1]) add_char(szDest, 'k');
  if (!m_board.m_piecemoved[COLOUR_BLACK][0]) add_char(szDest, 'q');

  // if none of the flags added, we add a dash
  if (szDest[strlen(szDest) - 1] == ' ') add_char(szDest, '-');

  add_char(szDest, ' ');

  // ep field
  if (strlen(szMove) > 3) {
    posid iFrom = (posid) ((szMove[0] - 'a') + (szMove[1] - '1') * 8);
    posid iTo = (posid) ((szMove[2] - 'a') + (szMove[3] - '1') * 8);

    if (board_getpiece(iTo) == PIECE_PAWN) {
      if ((iTo - iFrom) == 16) {
        add_char(szDest, szMove[0]);
        add_char(szDest, szMove[1] + 1);
      } else {
        add_char(szDest, '-');
      }
    } else if (board_getpiece(iTo) == BLACK_PAWN) {
      if ((iFrom - iTo) == 16) {
        add_char(szDest, szMove[0]);
        add_char(szDest, szMove[3] + 1);
      } else {
        add_char(szDest, '-');
      }
    } else {
      add_char(szDest, '-');
    }
  } else {
    // no valid move given, so no ep move possible
    add_char(szDest, '-');
  }
}

/**
 * normalize sanmove in current position from szMove as listed
 */
void board_sanmove(char *szDest, const char *szMove, const size_t iDestLen) {
  posid iFrom;
  posid iTo;
  t_piece cPiece;

  szDest[0] = '\0';

  if (strlen(szMove) > 3) {
    // very shorthand way to copy move into local working variables
    iFrom = (posid) ((szMove[0] - 'a') + (szMove[1] - '1') * 8);
    iTo = (posid) ((szMove[2] - 'a') + (szMove[3] - '1') * 8);

    cPiece = board_getpiece(iFrom);

    if ((cPiece == PIECE_KING) || (cPiece == BLACK_KING)) {
      // look for castling
      if (abs(iFrom - iTo) == 2) {
        // only now we have a castling move, so load it
        if (iFrom > iTo) {
          // long castle
          strncpy(szDest, "O-O-O", iDestLen);
        } else {
          // short castle
          strncpy(szDest, "O-O", iDestLen);
        }
        return;
      }
    }

    if ((cPiece == PIECE_PAWN) || (cPiece == BLACK_PAWN)) {
      if ((iFrom % 8) != (iTo % 8)) {
        // pawn captures are actually noted with the from field and a 'x' sign
        add_char(szDest, szMove[0]);
        add_char(szDest, 'x');
      }
    } else {
      add_char(szDest, toupper(cPiece));
      if (board_getpiece(iTo) != FIELD_EMPTY)
        add_char(szDest, 'x');
    }

    // destination field
    add_char(szDest, szMove[2]);
    add_char(szDest, szMove[3]);

    if ((toupper(cPiece) == PIECE_QUEEN)
            || (toupper(cPiece) == PIECE_ROOK)
            || (toupper(cPiece) == PIECE_BISHOP)
            || (toupper(cPiece) == PIECE_KNIGHT)) {
      // check if another piece could have made this move
      // modify destination move by inserting from square
      board_sanambigious(szDest, cPiece, iFrom, iTo);
    }

    // add promotion flags
    if ((cPiece == PIECE_PAWN) && (szMove[3] == '8')) {
      add_char(szDest, '=');
      add_char(szDest, szMove[4]);
    }

    if ((cPiece == BLACK_PAWN) && (szMove[3] == '1')) {
      add_char(szDest, '=');
      add_char(szDest, (szMove[4] & UPPERCASEBIT));
    }
  }
}

/**
 * check if there are 2 different pieces that can go to the destination
 * if so, change the move to make it more descriptive
 * normally this is done by looking at all possible moves and see
 * if there is the same piece that can go to the same destination
 * TODO: beware: exception is when both rown and column need to be specified
 */
static void board_sanambigious(char *szDest, const t_piece cPiece,
        const posid iFrom, const posid iTo) {
  posid iPos;
  int iMatches = 0;
  int iMoves;
  int i;
  t_alg1_moves m1_movesbuffer;
  char szMove[MAX_DATANAME_LENGTH];

  for (iPos = 0; iPos < 64; iPos++) {
    if (iPos != iFrom) {
      if (board_getpiece(iPos) == cPiece) {
        // there is a second piece, check if it can play to the given position
        iMatches = 1;
      }
    }
  }

  if (iMatches == 0) return;

  // we know that there is at least a second piece of the same kind
  // create list of all possible moves for later calculation

  // work with temp buffer to protect against overflow
  strcpy(szMove, szDest);

  memcpy(&m1_movesbuffer, &m1_movelist, sizeof (m1_movelist));
  iMoves = search_fill_movelist(game.curply % 2 ? COLOUR_BLACK : COLOUR_WHITE);

  // we loop through all, because there could be multiple pieces moving to the same
  // theoretically anyway, because normally there is only conflict after promotion
  for (iPos = 0; iPos < 64; iPos++) {
    if (iPos != iFrom) {
      if (board_getpiece(iPos) == cPiece) {
        // there is a second piece, check if it can play to the given position
        for (i = 0; i < iMoves; i++) {
          if ((MAP144TO64(m1_movelist.m1_moves[i].ucFrom) == iPos)
                  && (MAP144TO64(m1_movelist.m1_moves[i].ucTo) == iTo)) {
            // a move from this position is possible to the destination
            if (M1_LIJN(MAP144TO64(m1_movelist.m1_moves[i].ucFrom)) == M1_LIJN(iPos)) {
              // add col, it will be enough to make it different
              ins_char(szMove, M1_LIJN(iFrom) + 'a', (int) strlen(szMove)-(szMove[1] == 'x' ? 3 : 2));
            } else if (M1_RIJ(MAP144TO64(m1_movelist.m1_moves[i].ucFrom)) == M1_RIJ(iPos)) {
              // add row, it will be enough to make it different
              ins_char(szMove, M1_RIJ(iFrom) + '1', (int) strlen(szMove)-(szMove[1] == 'x' ? 3 : 2));
            }
          }
        }
      }
    }
  }

  // copy back
  memcpy(&m1_movelist, &m1_movesbuffer, sizeof (m1_movelist));

  strcpy(szDest, szMove);
}

/**
 * Decorate szSanMove with + ot # sign
 * @param szDest
 * @param szMove
 */
static void board_sanmovedecorate(char *szDest, const t_piece cPieceMoved) {
  t_boolean bBlackToMove = myFALSE;

  if ((cPieceMoved == PIECE_PAWN)
          || (cPieceMoved == PIECE_ROOK)
          || (cPieceMoved == PIECE_KNIGHT)
          || (cPieceMoved == PIECE_BISHOP)
          || (cPieceMoved == PIECE_QUEEN)
          || (cPieceMoved == PIECE_KING))
    bBlackToMove = myTRUE;

  if (alg1_isInCheck(bBlackToMove == myTRUE ? myFALSE : myTRUE)) {
    if (count_movelist(bBlackToMove == myTRUE ? COLOUR_BLACK : COLOUR_WHITE) < 1) {
      add_char(szDest, '#');
    } else {
      add_char(szDest, '+');
    }
  }
}

// --------------------------------------------------------------------

static void board_setpiece(const posid iIndex, const t_piece cPiece) {
  m_board.m_fields[M1_LIJN(iIndex)][M1_RIJ(iIndex)] = cPiece;
}

/**
 * because the fen string goes row 8 to row 1, we invert the positioning here
 * only used for fen reading
 */
static void board_setinvpiece(const posid iIndex, const t_piece cPiece) {
  m_board.m_fields[M1_LIJN(iIndex)][7 - M1_RIJ(iIndex)] = cPiece;
}

// --------------------------------------------------------------------

static t_piece board_getpiece(const posid iIndex) {
  return m_board.m_fields[M1_LIJN(iIndex)][M1_RIJ(iIndex)];
}

// --------------------------------------------------------------------
// eof
