/*
 * epd_testsuite.cpp
 * Project: part of belofte - A Promising Chess Program
 * attention, code is no re-entrant because of the global statics
 * on Visual C++, some open source is required in order to compile.  See below
 * for details.
 *
 * Created by Yves De Billoëz on 1/05/2006.
 * Copyright 2006 Yves De Billoëz. All rights reserved.
 */

#include "belofte.h"

#include "YUtils.h"
#include "usercmd.h"
#include "board.h"
#include "search.h"
#include "YLineParser.h"

#include "epd_testsuite.h"

/** global entrypoint */
static int epd_parseline(const char *szLine, const int iStartPos);
static int epd_evalresult(const char *szResult);

/** local functions */
static void epd_parsecmdchain(const char *szLine);
static void epd_parsecmd(const char *szCmd);
static void epd_addbm(const char *szCmd);
static void epd_addpm(const char *szCmd);
static void epd_addam(const char *szCmd);
static void epd_addid(const char *szCmd);

/**
 */
#define BM_LISTLEN 10
#define ELEMENTLEN 20
#define LINE_MAXLEN 127

#define BM_MAXTOKENS 20

extern game_t game;
extern uimode_t uimode;

/**
 */
static char bm[BM_LISTLEN][ELEMENTLEN]; // priority high
static int bm_index = -1;
static char am[BM_LISTLEN][ELEMENTLEN];
static int am_index = -1;
static char pm[BM_LISTLEN][ELEMENTLEN]; // priority low
static int pm_index = -1;
static char id[LINE_MAXLEN];

void epd_parsefile(const char *szFile) {

  FILE *fpOpen;
  char szBuffer[MAX_LINE_LENGTH];
  char szLine[MAX_LINE_LENGTH];
  int iLines = 0;
  int iGoods = 0;

  fpOpen = fopen(szFile, "rt");
  if (fpOpen != NULL) {
    while (file_readline(szLine, fpOpen, sizeof (szLine)) == 0) {
      if ((strlen(szLine) > 5) && (szLine[0] != ';')) {
        iLines++;
        if (epd_parseposition(szLine) > 0) {
          iGoods++;
        }
      }
    }
    fclose(fpOpen);
    snprintf(szBuffer, sizeof (szBuffer), "Score = %d/%d\n", iGoods, iLines);
    cmdresp_snd(szBuffer);
  } else {
    snprintf(szBuffer, sizeof (szBuffer), "Could not open %s!", szFile);
    cmdresp_snd(szBuffer);
  }
}

/**
 * Parse a position
 * @param szLine
 * @return non zero if found
 */
int epd_parseposition(const char *szLine) {
  int nPos;
  int iSearchResult;
  char szBuffer[MAX_LINE_LENGTH];
  char szMove[MAX_SANMOVE_LENGTH];
  char szResult[MAX_LINE_LENGTH];
  char szSANMove[30];
  int iEpdResult = 0;

  nPos = board_setupfen(szLine);
  epd_parseline(szLine, nPos);
  // TODO: we should indicate that a move is always expected, no resign
  // or draw proposal
  iSearchResult = search_root(szMove, sizeof (szMove), szResult, sizeof (szResult));
  if (iSearchResult >= SEARCH_FLAG) {
    // play move in order to retrieve SAN move
    board_applymove(szMove);
    strcpy(szSANMove, game.m_boardhistory[game.curply - 1].szSANMove);
    // there was no problem during search, so we eval
    iEpdResult = epd_evalresult(szSANMove);
    // in case no matching SAN move found, retry with coordinate move
    if (iEpdResult == 0) iEpdResult = epd_evalresult(szMove);
    if (iEpdResult > 0) {
      snprintf(szBuffer, sizeof (szBuffer), "%s Solved: %s", id, szSANMove);
      cmdresp_snd(szBuffer);
    } else if (iEpdResult < 0) {
      snprintf(szBuffer, sizeof (szBuffer), "%s Failed: %s played", id, szSANMove);
      cmdresp_snd(szBuffer);
    } else {
      snprintf(szBuffer, sizeof (szBuffer), "%s Not solved: %s played instead - %s", id, szSANMove, szLine);
      cmdresp_snd(szBuffer);
      if (uimode.post_mode)
        board_print();
    }
  }
  return iEpdResult;
}

/**
 * parse a single epd line
 * TODO: add support for ! and ? glyphs, for bm comma separated enumerators
 * TODO: add support for denormalized strings
 */
int epd_parseline(const char *szLine, const int iStartPos) {
  int iIndex;
  int iResult = 0;

  if (strlen(szLine) > 0) {
    if (szLine[0] != ';') {
      iIndex = iStartPos;
      epd_parsecmdchain(szLine + iIndex);
    }
  }

  return iResult;
}

/**
 * verify that the expected result is present after the line has been parsed
 * @return 0 if nothing found
 *         1 if predected move found or no avoid move played
 *         2 if best move found
 *         -1 if avoid move played
 */
int epd_evalresult(const char *szResult) {
  int i;
  int iResult = 0;

  if (pm_index != -1) {
    i = pm_index;
    while (i>-1) {
      if (strcmp(pm[i--], szResult) == 0) iResult = 1;
    }
  } else if (bm_index != -1) {
    i = bm_index;
    while (i>-1) {
      if (strcmp(bm[i--], szResult) == 0) iResult = 2;
    }
  } else if (am_index != -1) {
    i = am_index;
    while (i>-1) {
      if (strcmp(am[i--], szResult) == 0) return -1;
    }
    return 1;
  }

  return iResult;
}

/**
 * parse all epd commands that come on the line of the epd string
 */
static void epd_parsecmdchain(const char *szLine) {
  int i;
  unsigned int j;
  char *aMainParts[BM_MAXTOKENS];
  int iNumTokens = YParse_Line(szLine, ";", aMainParts, BM_MAXTOKENS, LINE_MAXLEN);

  // init some variables to their init value
  bm_index = -1;
  pm_index = -1;
  am_index = -1;
  id[0] = '\0';

  for (i = 0; i < iNumTokens; i++) {
    for (j = 0; j < strlen(aMainParts[i]); j++)
      if (aMainParts[i][j] != ' ') {
        // the tokens are actually filled, so we can handle them
        epd_parsecmd((aMainParts[i]) + j);
        break;
      }
  }

  YParse_Line_Clean(aMainParts, iNumTokens);
}

/**
 * parse a single epd command, store results in a static file buffer
 */
void epd_parsecmd(const char *szCmd) {
  if (strncmp(szCmd, "bm", 2) == 0) {
    epd_addbm(szCmd);
  } else if (strncmp(szCmd, "pm", 2) == 0) {
    epd_addpm(szCmd);
  } else if (strncmp(szCmd, "am", 2) == 0) {
    epd_addam(szCmd);
  } else if (strncmp(szCmd, "id", 2) == 0) {
    epd_addid(szCmd);
  }
}

/**
 * add bm to index
 */
static void epd_addbm(const char *szCmd) {
  size_t i;
  char *aSubParts[BM_MAXTOKENS];

  size_t iNumTokens = YParse_Line(szCmd + 2, " !?"
          , aSubParts, BM_MAXTOKENS, LINE_MAXLEN);

  for (i = 0; i < iNumTokens; i++) {
    strncpy(bm[++bm_index], aSubParts[i], ELEMENTLEN);
  }

  YParse_Line_Clean(aSubParts, iNumTokens);
}

/**
 * add am to index
 */
static void epd_addam(const char *szCmd) {
  strncpy(am[++am_index], szCmd + 3, ELEMENTLEN);
}

/**
 * add pm to index
 */
static void epd_addpm(const char *szCmd) {
  strncpy(pm[++pm_index], szCmd + 3, ELEMENTLEN);
}

/**
 * add id
 */
static void epd_addid(const char *szCmd) {
  strncpy(id, szCmd + 3, ELEMENTLEN);
}
