/**
 * search.cpp
 * Project: part of belofte - A Promising Chess Program
 *
 * Interface to different search algorithms
 * @see alg1_search.cpp and alg2_search.cpp
 */

#include "belofte.h"

#include "board.h"
#include "search.h"
#include "usercmd.h"

#include "alg1_search.h"
#include "alg1_board.h"
#include "alg2_search.h"

// --------------------------------------------------------------------

alg2_search *searchObj;

// --------------------------------------------------------------------

extern game_t game;
extern uimode_t uimode;
extern level_t level;
extern uistate_t uistate;
extern search_t search;

// --------------------------------------------------------------------

/**
 * function called once for engine construction, after change of
 * algorithm, needs to be called again
 * execution should not take a lot of time
 */
void search_constructor(void) {
  if (level.search_algoritme == 1) {
    alg1_constructor();
  } else {
    searchObj = new alg2_search();
  }
  search.engine_searching = myFALSE;
}

/**
 * constructor that can take a bit more time
 */
void search_delayedconstructor(void) {
  search.engine_searching = myFALSE;
}

/**
 * function called when changing algorithm or when exiting (TODO)
 */
void search_destructor(void) {
  if (level.search_algoritme != 1) {
    if (searchObj != NULL) {
      delete searchObj;
      searchObj = NULL;
    }
  }
}

/**
 * function called once for new game, here the board position is copied from the board into the engine
 * calling this function will cause ep, rock and repetition and 50 moves flags to be reset
 */
void search_init(void) {
  if (level.search_algoritme == 1) {
    b1_init();
  } else {
    searchObj->initialboard();
  }
}

/**
 * Do performance test
 * @return number of nodes searched
 */
int search_perft(void) {
  uistate.perftnodes = 0;
  usercmd_delayedconstructor(); // in case we are not in xboard mode and the protover
  // command is not called, make sure it gets called upon the first move

  search.search_cancel = myFALSE;
  search.bestmovedepth = 0;

  search.engine_searching = myTRUE;

  if (level.search_algoritme == 1) {
    alg1_iterativedeepening(level.maxdepth);
  } else {
    searchObj->search();
  }
  search.engine_searching = myFALSE;
  return uistate.perftnodes;
}

/**
 * do perform search
 * @param szMove szMove move to be filled in
 * @param iLenMove maximum lenght of string
 * @param szResult can be special code (-1, -2, ...)
 * @param iLenResult maximum lenght of string
 * @return SEARCH_ERROR in case of some problem, SEARCH_OK in case of ok
 *         SEARCH_FLAG in case of end position reached
 */
int search_root(char *szMove, const size_t iLenMove,
        char *szResult, const size_t iLenResult) {
  t_score fScore;
  char szBuffer[MAX_DATANAME_LENGTH];

  uistate.searchnodes = 0;
  uistate.thinkingtime = 0.0;
  uistate.searchmaxdepth = 0;
  uistate.iCurrentDepth = 0;
  time(&uistate.start_t);

  if (iLenMove < 8) return SEARCH_ERROR; // some error, return right away
  if (iLenResult < MAX_LINE_LENGTH) return SEARCH_ERROR;

  szMove[0] = '\0';
  szResult[0] = '\0';

  usercmd_delayedconstructor(); // in case we are not in xboard mode and the protover
  // command is not called, make sure it gets called upon the first move

#if defined(FEATURE_USEBOOK)
  if (book_move(game.m_boardhistory[game.curply].szFEN, szMove)) {
    // here perform opening book move and return
    cmdresp_snd_info("book move");
    return SEARCH_OK;
  }
#endif

  if (game.curply < MAX_GAME_LENGTH - 1) {
    snprintf(szBuffer, sizeof (szBuffer),
            "Starting search for %s, %s %d %d",
            game.curply & 1 ? "black" : "white",
            game.m_boardhistory[game.curply].szFEN,
            game.m_boardhistory[game.curply].m_rule50,
            game.m_boardhistory[game.curply].m_movenumber);
  } else {
    snprintf(szBuffer, sizeof (szBuffer),
            "Starting search for %s, ply %d (warning, exceeds bounds!)",
            game.curply & 1 ? "black" : "white",
            (int) game.curply);
  }
  cmdresp_snd_info(szBuffer);
  cmdresp_snd_debug(szBuffer);

  search_inittimeformove(game.curply > 20 ? 20 : 32); // TODO: pass real pieces on board
  search_timeinformation(szBuffer, sizeof (szBuffer));
  cmdresp_snd_info(szBuffer);
  cmdresp_snd_debug(szBuffer);

  search.search_cancel = myFALSE;
  search.bestmovedepth = 0;

  search.engine_searching = myTRUE;

  if (level.search_algoritme == 1) {
    alg1_iterativedeepening(level.maxdepth);
  } else {
    searchObj->search();
  }
  search.engine_searching = myFALSE;

  strncpy(szMove, search.bestmoves[search.bestmovedepth].szMove, iLenMove);
  fScore = search.fSearchScore;

  if (fScore < -SCORE_RESIGN) {
    if (game.curply & 1) {
      // black resigns
      strncpy(szResult, SEARCHRESULT_BLACKRESIGNS, iLenMove);
    } else {
      // white resigns
      strncpy(szResult, SEARCHRESULT_WHITERESIGNS, iLenMove);
    }
    return SEARCH_FLAG;
  } else if (fScore >= SCORE_WON) {
    if (game.curply & 1) {
      // black wins
      strncpy(szResult, SEARCHRESULT_BLACKMATES, iLenMove);
    } else {
      // white wins
      strncpy(szResult, SEARCHRESULT_WHITEMATES, iLenMove);
    }
    return SEARCH_FLAG;
  }
  return SEARCH_OK;
}

/**
 * try to increase time per move if score oscillates in between search depths
 */
int search_adjusttimeformove(void) {
  if (level.level_mode == LEVELMODE_TIMEKO) {
    // fixed time per game
    // adjust if possible
  } else if (level.level_mode == LEVELMODE_TIMEANDMOVES) {
    // incremental clocks or tournament clocks
    // adjust if possible
  }
  return 0; // no adjustment done
}

/**
 * calculate time available for current move and set it in the structure
 */
void search_inittimeformove(const int iPiecesOnBoard) {
  if (level.level_mode == LEVELMODE_TIMEPERMOVE) {
    // fixed time per move
    // level.max_movetime has been set already
    level.estimated_movetime = level.max_movetime;
    level.estimated_gametime = 0; // not defined
  } else if (level.level_mode == LEVELMODE_TIMEKO) {
    // fixed time per game
    level.estimated_gametime = level.game_timeleft;
    level.estimated_movetime = level.game_timeleft / (iPiecesOnBoard + 10);
    level.max_movetime = level.game_timeleft / 5;
  } else if (level.level_mode == LEVELMODE_TIMEANDMOVES) {
    // incremental clocks or time per period or combination
    if (level.additional_secs && level.moves_leftinperiod == 0) {
      // ficher clock, but with no repeat after n moves
      level.estimated_gametime = level.game_timeleft
              + (10 + iPiecesOnBoard) * level.additional_secs;
      level.max_movetime = level.estimated_gametime / iPiecesOnBoard;
      if (level.max_movetime > level.game_timeleft / 2) {
        level.max_movetime = level.game_timeleft / 2;
      }
      level.estimated_movetime = ((level.game_timeleft / (10 + iPiecesOnBoard))
              + level.max_movetime) / 2;
    } else if (level.moves_leftinperiod == 0) {
      // same as KO calculate, can we get here
      level.estimated_gametime = level.game_timeleft;
      level.estimated_movetime = level.game_timeleft / (iPiecesOnBoard + 10);
      level.max_movetime = level.game_timeleft / 5;
    } else if (level.moves_leftinperiod < 3) {
      // spread calculation of time used in last moves equally
      level.estimated_gametime = level.game_timeleft;
      level.max_movetime = level.game_timeleft / level.moves_leftinperiod;
      level.estimated_movetime = level.max_movetime;
    } else {
      // take into account that we can speed up just before time control
      level.estimated_gametime = level.game_timeleft
              + level.moves_leftinperiod * level.additional_secs
              + (10 + iPiecesOnBoard) * level.additional_secs;
      level.max_movetime = level.estimated_gametime / level.moves_leftinperiod;
      if (level.max_movetime > level.game_timeleft / (level.moves_leftinperiod - 2)) {
        level.max_movetime = level.game_timeleft / (level.moves_leftinperiod - 2);
      }
      level.estimated_movetime = ((level.game_timeleft / level.moves_leftinperiod)
              + level.max_movetime) / 2;
    }
  } else {
    level.max_movetime = INT_MAX; // infinite, dummy number
    level.estimated_movetime = level.max_movetime;
    level.game_timeleft = 0; // not defined
    level.estimated_gametime = 0; // not defined
  }
}

void search_timeinformation(char *szInfo, const size_t nInfoLen) {
  char szBuffer[MAX_DATANAME_LENGTH];
  if (level.level_mode == LEVELMODE_TIMEPERMOVE) {
    sprintf(szBuffer, "Time per move: %d seconds",
            level.max_movetime);
  } else if (level.level_mode == LEVELMODE_TIMEKO) {
    sprintf(szBuffer, "Time left for game (KO) %d:%d, for move %d (max %d)",
            level.game_timeleft / 60, level.game_timeleft % 60,
            level.estimated_movetime, level.max_movetime);
  } else if (level.level_mode == LEVELMODE_TIMEANDMOVES) {
    if (level.moves_leftinperiod && level.additional_secs) {
      sprintf(szBuffer, "Time left for moves %d:%d, for move %d (max %d), "
              "Ficher increment %d, next control %d",
              level.estimated_gametime / 60, level.estimated_gametime % 60,
              level.estimated_movetime, level.max_movetime,
              level.additional_secs, level.moves_leftinperiod);
    } else if (level.moves_leftinperiod) {
      sprintf(szBuffer, "Time left for moves %d:%d, for move %d (max %d), "
              "Next control %d",
              level.estimated_gametime / 60, level.estimated_gametime % 60,
              level.estimated_movetime, level.max_movetime,
              level.moves_leftinperiod);
    } else if (level.additional_secs) {
      sprintf(szBuffer, "Time left for game %d:%d, for move %d (max %d), "
              "Ficher increment %d",
              level.estimated_gametime / 60, level.estimated_gametime % 60,
              level.estimated_movetime, level.max_movetime,
              level.additional_secs);
    } else {
      // TODO: can we be here
      sprintf(szBuffer, "??? left for game %d:%d, for move %d (max %d)",
              level.game_timeleft / 60, level.game_timeleft % 60,
              level.estimated_movetime, level.max_movetime);
    }
  } else if (level.level_mode == LEVELMODE_NOTIME) {
    sprintf(szBuffer, "Fixed depth: %d",
            level.maxdepth);
  } else if (level.level_mode == LEVELMODE_PERFTEST) {
    sprintf(szBuffer, "Performance Test at depth %d",
            level.maxdepth);
  }
  strncpy(szInfo, szBuffer, nInfoLen);
}

/** apply move on board after search
 */
int search_apply_move(const char *szMove) {
  posid iFrom;
  posid iTo;
  t_promotion cFlag = PROMOTION_NONE;

  if (strlen(szMove) == 5)
    cFlag = (t_promotion) szMove [4];

  iFrom = (2 + szMove[0] - 'a') + (2 + szMove[1] - '1')*12;
  iTo = (2 + szMove[2] - 'a') + (2 + szMove[3] - '1')*12;

  if (game.curply & 1) {
    // curply is odd for white, because it is updated before this call
    b1_apply_movewhite(iFrom, iTo, cFlag);
  } else {
    b1_apply_moveblack(iFrom, iTo, cFlag);
  }

  return 0;
}

// eof
