/***********************************************************************************
 * File: usercmd.cpp
 * Project: part of belofte - A Promising Chess Program
 * Author: yves
 ***********************************************************************************/

#include "belofte.h"

#include "usercmd.h"
#include "board.h"
#include "YUtils.h"
#include "search.h"
#include "evaluation.h"

#include "alg1_moves.h"
#include "alg1_book.h"
#include "alg1_eval.h"
#include "epd_testsuite.h"
#include "YLineParser.h"

// --------------------------------------------------------------------
// data structures needed for commands and search

extern uistate_t uistate;
extern uimode_t uimode;
extern level_t level;
extern game_t game;
extern search_t search;

/*----------------------------------------------------------------------+
|      message                                                         |
+----------------------------------------------------------------------*/

static char startup_message[] =
#if defined(FEATURE_UCI_INTERFACE)
        "This is '" MYNAME "-uci " MYVERSION "' (A Promising Chess Program)\n"
#else
        "This is '" MYNAME " " MYVERSION "' (A Promising Chess Program)\n"
#endif
        "Copyright (C) 1981-2020 Yves De Billoëz\n"
        "Please refer to the license terms (license.html) before using.\n"
        "\n"
        "Type 'help' for a list of commands.";

/*----------------------------------------------------------------------+
|      commands                                                        |
+----------------------------------------------------------------------*/

#if defined(FEATURE_XBOARD_INTERFACE)
static void cmdresp_snd_protover(const int nProtover);
#endif
static void cmdresp_snd_error(const char *szCommand, const char *szReason);
static void cmdresp_snd_errormove(const char *szMove);
static void cmdresp_not_supported(const char *szCommand);

static int usercmd_readarg(char *szLine, char *szArg, const size_t nSizeArg);
static void usercmd_signalcatch(int s);

static void cmd_bd(UNUSED const char *szCommand, UNUSED const char *szArg);
static void cmd_execute(const char *szCommand, const char *szArg);
static void cmd_help(UNUSED const char *szCommand, const char *szArg);
static void cmd_ls(UNUSED const char *szCommand, const char *szArg);
static void cmd_nps(UNUSED const char * szCommand, UNUSED const char *szArg);
static void cmd_otim(UNUSED const char *szCommand, const char *szArg);
static void cmd_save(UNUSED const char *szCommand, UNUSED const char *szArg);
static void cmd_setboard(const char *szCommand, const char *szArg);
static void cmd_st(UNUSED const char *szCommand, const char *szArg);
static void cmd_time(UNUSED const char *szCommand, const char *szArg);
static void cmd_usermove(UNUSED const char *szCommand, const char *szArg);
static void cmd_white(UNUSED const char *szCommand, UNUSED const char *szArg);

/**
 * configure the engine and all local parameters
 * this is called when a total reset is required such as launching or
 * relaunching the application whilst in memory
 */
void usercmd_constructor() {
  // make sure we catch this signal
  signal(SIGINT, usercmd_signalcatch);

  uimode.post_mode = myTRUE;
  uimode.protover = 1; // assume we only support level 1 commands unless told differently
  uimode.ics_mode = myFALSE;
  uimode.processing_mode = myFALSE;

  level.level_mode = LEVELMODE_TIMEPERMOVE;
  level.max_movetime = DEFAULT_TIMEPERMOVE;
  level.maxdepth = DEFAULT_MAXDEPTH;

  setbuf(stdout, NULL);
  setbuf(stdin, NULL);
}

/**
 * second constructor, taking more time and issued during protover command
 */
void usercmd_delayedconstructor(void) {
  static int iInit = myFALSE;

  if (!iInit) {
    // make sure it is only executed once
    iInit = myTRUE;
    book_load(BOOKNAME);
  }
}

// --------------------------------------------------------------------

static void usercmd_signalcatch(int s) {
  signal(s, usercmd_signalcatch);
}

/**
 * gnuchess compatible implementation
 * argument can be nosave - file is not saved before (cmd_save)
 *                 nobd - not followed by automatic cmd_bd command
 */
void cmd_new(UNUSED const char *szCommand, const char *szArg) {
  if (strcmp(szArg, "nosave") && game.curply > 0)
    cmd_save("", "");
  cmd_force("", "");

  level.random_mode = myFALSE;
  level.level_mode = LEVELMODE_TIMEPERMOVE;
  level.max_movetime = DEFAULT_TIMEPERMOVE;
  level.maxdepth = DEFAULT_MAXDEPTH;
  level.search_algoritme = DEFAULT_ALG;
  StartTimeThread();

  uistate.ponder_mode = myFALSE;
  uistate.computer_engines[0] = myFALSE;
  uistate.computer_engines[1] = myTRUE;

  game.curply = 0;
  game.firstply = 0;
  strncpy(game.szResult, "*", sizeof (game.szResult));
  strncpy(game.szOpponentName, "?", sizeof (game.szOpponentName));
  strncpy(game.szSite, "127.0.0.1", sizeof (game.szSite));

  board_setupdefaultfen();
  if ((!uimode.engine_mode) && (!uimode.processing_mode))
    if (strcmp(szArg, "nobd"))
      board_print();
}

#if defined(FEATURE_UCI_INTERFACE)

static void cmd_ucinewgame(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmd_new("", "nobd");
}

#define POSITION_MAXTOKENS 6 + MAX_GAME_LENGTH

/**
 * UCI handle position command
 *  position [fen <fenstring>|startpos] [moves <move*>]
 */
static void cmd_position(UNUSED const char *szCommand, UNUSED const char *szArg) {
  int i, j;
  char szBuffer[MAX_DATANAME_LENGTH];
  char szBuffer2[MAX_DATANAME_LENGTH];
  char *aMainParts[POSITION_MAXTOKENS];
  int iNumTokens;

  cmd_force("", "");

  strcpy(szBuffer, szArg);
  iNumTokens = (int) YParse_Line(szBuffer, " ", aMainParts, POSITION_MAXTOKENS, MAX_DATANAME_LENGTH);

  // set startup position
  if (strcmp(aMainParts[0], "startpos") == 0) {
    board_setupdefaultfen();
  } else if (strcmp(aMainParts[0], "fen") == 0) {
    strcpy(szBuffer2, aMainParts[1]);
    for (i = 2; i <= 6; i++) {
      strcat(szBuffer2, " ");
      strcat(szBuffer2, aMainParts[i]);
    }
    board_setupfen(szBuffer2);
  }

  // now apply all moves
  for (i = 0; i < (iNumTokens - 1); i++) {
    if (strcmp(aMainParts[i], "moves") == 0) {
      for (j = i + 1; j < iNumTokens; j++) {
        cmd_usermove("", aMainParts[j]);
      }
      break;
    }
  }

  YParse_Line_Clean(aMainParts, (size_t) iNumTokens);
}
#endif

/**
 * print the board to the user
 */
static void cmd_bd(UNUSED const char *szCommand, UNUSED const char *szArg) {
  board_print();
}

#if defined(FEATURE_XBOARD_INTERFACE)

/**
 * set xboard mode, xboard means compatible
 */
static void cmd_xboard(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uimode.xboard_mode = myTRUE;
  uimode.engine_mode = myTRUE;
  cmdresp_snd("");
}
#endif

/**
 * toggle random mode
 */
static void cmd_random(UNUSED const char *szCommand, UNUSED const char *szArg) {
  level.random_mode = (t_boolean) (1 - (int) level.random_mode);
}

#if defined(FEATURE_CHESS_VARIANT)

/**
 * allow to set game variation, is not supported
 */
static void cmd_variant(const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}
#endif

/**
 * exit engine
 * TODO: also understands bye commands
 */
void cmd_quit(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmd_debug("", "off");
  exit(0);
}

/**
 * player plays white
 */
static void cmd_white(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uistate.computer_engines[0] = myFALSE;
  uistate.computer_engines[1] = myTRUE;
}

/**
 * player plays black
 */
static void cmd_black(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uistate.computer_engines[0] = myTRUE;
  uistate.computer_engines[1] = myFALSE;
}

/**
 * switch to force mode, turn off ponder
 * search thread needs to be aborted
 */
void cmd_force(UNUSED const char *szCommand, UNUSED const char *szArg) {
  // TODO: about search
  uistate.pondering = myFALSE;
  uistate.computer_engines[0] = myFALSE;
  uistate.computer_engines[1] = myFALSE;
}

#define CMD_GO_MAXTOKENS 10

/**
 * computer plays colour to move
 * UCI: supports sub-command: depth, movetime, infinite, wtime, btime
 */
void cmd_go(UNUSED const char *szCommand, UNUSED const char *szArg) {
#if defined(FEATURE_UCI_INTERFACE)
  int iNumElements;
  int iTokenPos, iTokenIncPos, iTokenMovesPos; // attention, index starts from 1
  int iTime;
  char *aMainParts[CMD_GO_MAXTOKENS];
  char szBuffer[MAX_DATANAME_LENGTH];
  char szBuffer2[MAX_DATANAME_LENGTH];

  if (uimode.uci_mode == myTRUE) {
    strcpy(szBuffer, szArg);
    iNumElements = (int) YParse_Line(szBuffer, " ", aMainParts, POSITION_MAXTOKENS, MAX_DATANAME_LENGTH);
    if ((iTokenPos = findinstringarray(aMainParts, iNumElements, "depth"))) {
      sprintf(szBuffer2, "cmd_fd(\"\",\"%s\")", aMainParts[iTokenPos]);
      cmdresp_snd_debug(szBuffer2);
      cmd_fd("", aMainParts[iTokenPos]);
    } else if ((iTokenPos = findinstringarray(aMainParts, iNumElements, "movetime"))) {
      // millisec to sec conversion
      sprintf(szBuffer, "%d", atoi(aMainParts[iTokenPos]) / 1000);
      sprintf(szBuffer2, "cmd_st(\"\",\"%s\")", szBuffer);
      cmdresp_snd_debug(szBuffer2);
      cmd_st("", szBuffer);
    } else if ((iTokenPos = findinstringarray(aMainParts, iNumElements, "infinite"))) {
      // infinite converted to 7 days
      sprintf(szBuffer2, "cmd_st(\"\",\"604800\")");
      cmdresp_snd_debug(szBuffer2);
      cmd_st("", "604800");
    } else {
      // we are supposedly in conventional or incremental clocks
      if (game.curply % 2) {
        // white to move
        iTokenPos = findinstringarray(aMainParts, iNumElements, "wtime");
        iTokenIncPos = findinstringarray(aMainParts, iNumElements, "winc");
        iTokenMovesPos = findinstringarray(aMainParts, iNumElements, "movestogo");
        iTime = atoi(aMainParts[iTokenPos]) / 1000;
        sprintf(szBuffer, "%d %d:%d %d",
                iTokenMovesPos ? atoi(aMainParts[iTokenMovesPos - 1]) : 0,
                iTime / 60, iTime % 60,
                atoi(aMainParts[iTokenIncPos]) / 1000);
        sprintf(szBuffer2, "cmd_level(\"\",\"%s\")", szBuffer);
        cmdresp_snd_debug(szBuffer2);
        cmd_level("", szBuffer);
        if ((iTokenPos = findinstringarray(aMainParts, iNumElements, "btime"))) {
          // millisecs to centisecs conversion
          sprintf(szBuffer, "%d", atoi(aMainParts[iTokenPos]) / 10);
          sprintf(szBuffer2, "cmd_otim(\"\",\"%s\")", szBuffer);
          cmdresp_snd_debug(szBuffer2);
          cmd_otim("", szBuffer);
        }
      } else {
        // black to move
        iTokenPos = findinstringarray(aMainParts, iNumElements, "btime");
        iTokenIncPos = findinstringarray(aMainParts, iNumElements, "binc");
        iTokenMovesPos = findinstringarray(aMainParts, iNumElements, "movestogo");
        iTime = atoi(aMainParts[iTokenPos]) / 1000;
        sprintf(szBuffer, "%d %d:%d %d",
                iTokenMovesPos ? atoi(aMainParts[iTokenMovesPos - 1]) : 0,
                iTime / 60, iTime % 60,
                atoi(aMainParts[iTokenIncPos]) / 1000);
        sprintf(szBuffer2, "cmd_level(\"\",\"%s\")", szBuffer);
        cmdresp_snd_debug(szBuffer2);
        cmd_level("", szBuffer);
        if ((iTokenPos = findinstringarray(aMainParts, iNumElements, "wtime"))) {
          // millisecs to centisecs conversion
          sprintf(szBuffer, "%d", atoi(aMainParts[iTokenPos]) / 10);
          sprintf(szBuffer2, "cmd_otim(\"\",\"%s\")", szBuffer);
          cmdresp_snd_debug(szBuffer2);
          cmd_otim("", szBuffer);
        }
      }
    }
    YParse_Line_Clean(aMainParts, (size_t) iNumElements);
  }
#endif
  if (game.curply % 2) cmd_white("", "");
  else cmd_black("", "");
  StartEngineThread();
}

/**
 * play oponent colour, start pondering, and make sure to start
 * thinking as soon as the last move is received
 * (Freechess sends playother and then the last move)
 */
static void cmd_playother(UNUSED const char *szCommand, UNUSED const char *szArg) {
  if (game.curply % 2) cmd_white("", "");
  else cmd_black("", "");
  uistate.ponder_mode = myTRUE;
  StartEngineThread();
}

/**
 * configure board using fen string
 */
static void cmd_setboard(UNUSED const char *szCommand, const char *szArg) {
  cmd_force("", "");
  board_setupfen(szArg);
}

/**
 * move now, registered as question mark in xboard or stop in uci
 */
static void cmd_movenow(UNUSED const char *szCommand, UNUSED const char *szArg) {
  search.search_cancel = myTRUE;
  cmd_force("", "");
}

#if defined(FEATURE_XBOARD_INTERFACE)

static void cmd_protover(UNUSED const char *szCommand, const char *szArg) {
  int d;
  sscanf(szArg, "%d", &d);

  if (d != uimode.protover) {
    uimode.protover = d;
    cmdresp_snd_protover(uimode.protover);
  } else {
    cmdresp_snd_debug("This protover already active");
  }
}
#endif

/**
 * Implement level such as level 0 5 0 for Blitz KO
 *    or 40 20:5 30 for 20 minutes 5 seconds for 40 moves + 30s/move
 * If last parameter is zero, level mode is LEVELMODE_TIMEKO
 * TODO: ICS might pass 0 on first move which actually means 0:10
 */
void cmd_level(UNUSED const char *szCommand, const char *szArg) {
  int period_minutes, period_seconds;

  if (strchr(szArg, ':')) {
    sscanf(szArg, "%d %d:%2d %d",
            &level.moves_perperiod, &period_minutes,
            &period_seconds, &level.additional_secs);

    level.game_timeleft = (period_minutes * 60) + period_seconds;
  } else {
    sscanf(szArg, "%d %d %d",
            &level.moves_perperiod,
            &period_minutes, &level.additional_secs);

    level.game_timeleft = period_minutes * 60;
  }
  level.moves_leftinperiod = level.moves_perperiod;

  level.level_mode = LEVELMODE_TIMEKO;
  if (level.moves_perperiod != 0) level.level_mode = LEVELMODE_TIMEANDMOVES;
  if (level.additional_secs != 0) level.level_mode = LEVELMODE_TIMEANDMOVES;
  search_inittimeformove(32);
  StartTimeThread();
}

void cmd_sd(UNUSED const char *szCommand, const char *szArg) {
  sscanf(szArg, "%d", &level.maxdepth);
  if (level.maxdepth < 0) level.maxdepth = 0;
  if (level.maxdepth > MAX_SEARCH_DEPTH) level.maxdepth = MAX_SEARCH_DEPTH;

  if (level.level_mode == LEVELMODE_NOTIME) {
    // if we set fixed depth before, we reset to defaults
    level.level_mode = LEVELMODE_TIMEPERMOVE;
    level.max_movetime = DEFAULT_TIMEPERMOVE;
    StartTimeThread();
  }
  search_inittimeformove(32);
}

void cmd_fd(UNUSED const char *szCommand, const char *szArg) {
  sscanf(szArg, "%d", &level.maxdepth);
  if (level.maxdepth < 0) level.maxdepth = 0;
  if (level.maxdepth > MAX_SEARCH_DEPTH) level.maxdepth = MAX_SEARCH_DEPTH;

  level.level_mode = LEVELMODE_NOTIME;
  StopTimeThread();
  search_inittimeformove(32);
}

/**
 * Do performace test of move generator
 * @param szArg Depth to generate moves
 * TODO: check if restarting time thread is required
 */
void cmd_perft(const char *szCommand, const char *szArg) {
  int nNodes;
  levelmode_t nOldLevelMode = level.level_mode;
  char szBuffer[MAX_DATANAME_LENGTH];

  cmd_force("", "");
  cmd_fd(szCommand, szArg);
  level.level_mode = LEVELMODE_PERFTEST;
  StopTimeThread();
  nNodes = search_perft();
  snprintf(szBuffer, sizeof (szBuffer), "Performance Test"
          " at depth %s - searched %d nodes",
          szArg, nNodes);
  cmdresp_snd_info(szBuffer);
  StartTimeThread();
  level.level_mode = nOldLevelMode;
}

static void cmd_st(UNUSED const char *szCommand, UNUSED const char *szArg) {
  level.level_mode = LEVELMODE_TIMEPERMOVE;
  sscanf(szArg, "%d", &level.max_movetime);
  StartTimeThread();
}

static void cmd_time(UNUSED const char *szCommand, UNUSED const char *szArg) {
  char szBuffer[MAX_DATANAME_LENGTH];
  int iTime;
  sscanf(szArg, "%d", &iTime);
  // convert from centiseconds to seconds
  iTime = iTime / 100;

  if (level.level_mode == LEVELMODE_TIMEPERMOVE) {
    level.max_movetime = iTime;
    snprintf(szBuffer, sizeof (szBuffer), "got time %s (for move)", szArg);
  } else {
    level.game_timeleft = iTime;
    snprintf(szBuffer, sizeof (szBuffer), "got time %s (for game)", szArg);
  }
  cmdresp_snd_info(szBuffer);
}

static void cmd_otim(UNUSED const char *szCommand, UNUSED const char *szArg) {
  char szBuffer[MAX_DATANAME_LENGTH];
  int iTime;
  sscanf(szArg, "%d", &iTime);
  // convert from centiseconds to seconds
  iTime = iTime / 100;

  snprintf(szBuffer, sizeof (szBuffer), "got otim %s", szArg);
  cmdresp_snd_info(szBuffer);

  level.opponent_time = iTime;
}

/**
 * give a hint about the best move, if no hint available, ignore
 */
static void cmd_hint(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

#if defined(FEATURE_XBOARD_INTERFACE)

static void cmd_draw(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}
#endif

/** playing against a computer, TODO: remove randomness on best move
 */
static void cmd_computer(UNUSED const char *szCommand, UNUSED const char *szArg) {
  level.random_mode = myFALSE;
  strncat(game.szOpponentName, " (computer)", sizeof (game.szOpponentName));
}

static void cmd_exit(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

#if defined(FEATURE_XBOARD_INTERFACE)

static void cmd_pause(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

static void cmd_resume(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}
#endif

static void cmd_result(UNUSED const char *szCommand, UNUSED const char *szArg) {
  // cut out comments from standard results, keep only the score data normalized
  if (strncmp("1-0", szArg, 3) == 0) {
    strncpy(game.szResult, "1-0", sizeof (game.szResult));
  } else if (strncmp("0-1", szArg, 3) == 0) {
    strncpy(game.szResult, "0-1", sizeof (game.szResult));
  } else if (strncmp("1/2-1/2", szArg, 7) == 0) {
    strncpy(game.szResult, "1/2-1/2", sizeof (game.szResult));
  } else {
    // in other cases, just take the string
    strncpy(game.szResult, szArg, sizeof (game.szResult));
  }

  cmd_save("", "");
  cmd_force("", "");
}

#if defined(FEATURE_USEBOOK)

/**
 * create a book from a PGN file
 */
static void cmd_createbook(UNUSED const char *szCommand, UNUSED const char *szArg) {
  book_create(szArg, BOOKNAME);
}

/**
 * load a different book from a book file
 */
static void cmd_loadbook(UNUSED const char *szCommand, const char *szArg) {
  book_load(szArg);
}
#endif

#if defined(FEATURE_XBOARD_ICS)

/**
 * recieve message from xboard that we are running on ICS server
 */
static void cmd_ics(UNUSED const char *szCommand, const char *szArg) {
  if (strcmp(szArg, "-") != 0) {
    uimode.ics_mode = myTRUE;
    strncpy(game.szSite, szArg, sizeof (game.szSite));
  }
}
#endif

/**
 * tell something about myself
 */
void cmd_about(UNUSED const char *szCommand, UNUSED const char *szArg) {
  puts(startup_message);
#if defined(FEATURE_XBOARD_INTERFACE)
  if (uimode.protover < 4) puts("Type 'protover 4' to have more modern interface.");
#endif
}

static void cmd_edit(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

#if defined(FEATURE_USEBOOK)

static void cmd_bk(UNUSED const char *szCommand, UNUSED const char *szArg) {
  book_listmoves(game.m_boardhistory[game.curply].szFEN, myTRUE);
}
#endif

/**
 * undo one move
 * TODO: set result to *, set time left to unknown
 */
static void cmd_undo(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmd_force("", "");
  board_unapplymove();
  board_print();
}

/**
 * undo the move and play it again
 * TODO: set result to *, set time left to unknown
 */
static void cmd_again(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmd_force("", "");
  board_unapplymove();
  cmd_go("", "");
}

/**
 * remove 2 moves
 * TODO: set result to *, set time left to unknown
 */
static void cmd_remove(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmd_force("", "");
  board_unapplymove();
  board_unapplymove();
  board_print();

  // TODO: now make sure to start thinking again
}

/**
 * Output movelist to output file
 * @param fOutPut
 */
void game_movelist(FILE * fOutPut) {
  int i;
  int iSpace = myTRUE;

  if (game.firstply != 0) {
    if ((game.firstply % 2) != 0) {
      fprintf(fOutPut, "%d. ... ", (game.firstply + 1) / 2);
    }
  }
  for (i = game.firstply; i < MIN(game.curply, MAX_GAME_LENGTH - 1); i++) {
    if ((i % 2) == 0) {
      fprintf(fOutPut, "%d. %s", (i / 2) + 1, game.m_boardhistory[i].szSANMove);
      iSpace = myFALSE;
    } else {
      fprintf(fOutPut, " %s ", game.m_boardhistory[i].szSANMove);
      iSpace = myTRUE;
    }
    if (((i) % 10) == 9) {
      fprintf(fOutPut, "\n");
      iSpace = myTRUE;
    }
  }
  if (!iSpace) fprintf(fOutPut, " ");
}

/**
 * save current game to a file
 */
static void game_savefile(const char *szFileName, const t_boolean bAppend) {
  FILE *fpSave;

  struct tm *p_tm;
  time_t l_tm;
  char szTime[80];

  if (game.curply > 0) {
    l_tm = time(NULL); // get system time
    p_tm = localtime(&l_tm); // return local time in the form of tm structure
    strftime(szTime, sizeof (szTime), "%Y.%m.%d", p_tm);

    fpSave = fopen(szFileName, bAppend == myTRUE ? "a+" : "w+");
    if (fpSave != NULL) {
      // Seven standard elements for PGN
      fprintf(fpSave, "[Event \"%s autosave\"]\n", MYNAME);
      fprintf(fpSave, "[Site \"%s\"]\n", game.szSite);
      fprintf(fpSave, "[Date \"%s\"]\n", szTime);
      fprintf(fpSave, "[Round \"?\"]\n");
      fprintf(fpSave, "[White \"%s\"]\n", uistate.computer_engines[0] ? MYFULLNAME : game.szOpponentName);
      fprintf(fpSave, "[Black \"%s\"]\n", uistate.computer_engines[1] ? MYFULLNAME : game.szOpponentName);
      fprintf(fpSave, "[Result \"%s\"]\n", game.szResult);
      fprintf(fpSave, "\n");
      game_movelist(fpSave);
      fprintf(fpSave, "%s\n\n", game.szResult);
      fclose(fpSave);
    }
  }
}

/**
 * save the game to a file
 */
static void cmd_save(UNUSED const char * szCommand, UNUSED const char *szArg) {
  if (strlen(szArg) == 0) {
    game_savefile(PGNNAME, myTRUE);
  } else {
    game_savefile(szArg, myFALSE);
  }
}

/** TODO: add some patern matching */
static void cmd_ls(UNUSED const char *szCommand, const char *szArg) {
  DIR *hDir;
  struct dirent *hFile;

  hDir = opendir(".");
  if (hDir) {
    while ((hFile = readdir(hDir)) != NULL) {
      if (!strcmp(hFile->d_name, ".")) continue;
      if (!strcmp(hFile->d_name, "..")) continue;
      if (hFile->d_name[0] == '.') continue;
      if ((strlen(szArg) == 0) || strstr(hFile->d_name, szArg))
        printf("%s\n", hFile->d_name);
    }

    closedir(hDir);
  }
}

/**
 * load an EPD file and parse it
 * @param szArg filename containing EPD positions
 */
static void cmd_epd(UNUSED const char *szCommand, const char *szArg) {
  // as EPD file can cause variations on who to move, deactive all
  cmd_force("", "");
  epd_parsefile(szArg);
}

/**
 * evaluate an EPD position passed as argument
 * @param szArg EPD position
 */
static void cmd_epdpos(UNUSED const char *szCommand, const char *szArg) {
  cmd_force("", "");
  board_setupfen(szArg);
  epd_parseposition(szArg);
}

/**
 * load a file and execute all moves
 */
static void cmd_load(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

#if defined(FEATURE_UCI_INTERFACE)

/**
 * go into uci mode
 */
static void cmd_uci(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uimode.uci_mode = myTRUE;
  uimode.engine_mode = myTRUE;
  uimode.protover = 8;

  cmdresp_snd("id name " MYFULLNAME);
  cmdresp_snd("id author " MYAUTHOR);
  //cmdresp_snd("setoption name UCI_AnalyseMode value true");
  cmdresp_snd("uciok");
}
#endif

/**
 * get some debugging info
 */
static void cmd_info(UNUSED const char *szCommand, UNUSED const char *szArg) {
  char szBuffer[MAX_DATANAME_LENGTH];
  char szBuffer2[MAX_DATANAME_LENGTH / 2];

  search_timeinformation(szBuffer2, sizeof (szBuffer2));
  snprintf(szBuffer, sizeof (szBuffer),
          " Xboard mode: %d UCI mode: %d Engine mode: %d ICS mode: %d Protover: %d\n"
          " Debug: %d, Out: %s\n"
          " Engine running %d Engine White %d Engine Black %d \n"
          " Random mode: %d, ponder mode: %d, pondering %d, algorithm %d\n"
          " %s\n"
          " Current move: %d - %s to move\n",
          (int) uimode.xboard_mode, (int) uimode.ics_mode,
          (int) uimode.uci_mode, (int) uimode.engine_mode, uimode.protover,
          uimode.debug,
          uimode.fpDebugOutput == NULL ? "--none--" : uimode.szDebugFile,
          (int) uistate.engine_running,
          (int) uistate.computer_engines[0], (int) uistate.computer_engines[1],
          (int) level.random_mode,
          (int) uistate.ponder_mode, (int) uistate.pondering, (int) level.search_algoritme,
          szBuffer2,
          (game.curply / 2) + 1,
          game.curply & 1 ? "black" : "white"
          );

  cmdresp_snd(szBuffer);

  strcpy(szBuffer, " Compiled with following flags: ");

#if defined(_DEBUG)
  strcat(szBuffer, "_DEBUG ");
#endif

#if defined(FEATURE_FENINVERT)
  strcat(szBuffer, "FEATURE_FENINVERT ");
#endif

#if defined(FEATURE_UCI_INTERFACE)
  strcat(szBuffer, "FEATURE_UCI_INTERFACE ");
#endif

#if defined(FEATURE_XBOARD_INTERFACE)
  strcat(szBuffer, "FEATURE_XBOARD_INTERFACE ");
#endif

#if defined(FEATURE_XBOARD_ICS)
  strcat(szBuffer, "FEATURE_XBOARD_ICS ");
#endif

#if defined(FEATURE_CHESS_VARIANT)
  strcat(szBuffer, "FEATURE_CHESS_VARIANT ");
#endif

#if defined(FEATURE_QS)
  strcat(szBuffer, "FEATURE_QS ");
#endif

#if defined(FEATURE_USEBOOK)
  strcat(szBuffer, "FEATURE_USEBOOK ");
#endif

  puts(szBuffer);
}

/**
 * start pondering
 */
static void cmd_hard(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uistate.ponder_mode = myTRUE;
}

/**
 * stop pondering
 */
static void cmd_easy(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uistate.ponder_mode = myFALSE;
}

void cmd_post(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uimode.post_mode = myTRUE;
}

void cmd_nopost(UNUSED const char *szCommand, UNUSED const char *szArg) {
  uimode.post_mode = myFALSE;
}

static void cmd_ping(UNUSED const char *szCommand, const char *szArg) {
  char szBuffer[MAX_DATANAME_LENGTH];

  snprintf(szBuffer, sizeof (szBuffer), "pong %s", szArg);
  cmdresp_snd(szBuffer);
}

#if defined(FEATURE_UCI_INTERFACE)

/**
 * UCI query to see if engine is ready, constructors are called
 * TODO: check how to avoid that constructors have been called before
 */
static void cmd_isready(UNUSED const char *szCommand, const char *szArg) {
  board_delayedconstructor();
  search_delayedconstructor();
  eval_delayedconstructor();
  usercmd_delayedconstructor();
  cmdresp_snd("readyok");
}
#endif

#if defined(FEATURE_XBOARD_INTERFACE)

static void cmd_analyze(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}
#endif

/**
 * print game to console
 */
static void cmd_game(UNUSED const char *szCommand, UNUSED const char *szArg) {
  game_movelist(stdout);
  cmdresp_snd("");
}

static void cmd_name(UNUSED const char *szCommand, const char *szArg) {
#if defined(FEATURE_XBOARD_ICS)
  char szBuffer[MAX_DATANAME_LENGTH];
#endif

  strncpy(game.szOpponentName, szArg, sizeof (game.szOpponentName)); // used when saving game

#if defined(FEATURE_XBOARD_ICS)
  if (uimode.ics_mode) {
    snprintf(szBuffer, sizeof (szBuffer),
            "tellopponent Welcome to my domain %s!  I am a computer.", szArg);
    cmdresp_snd(szBuffer);
    snprintf(szBuffer, sizeof (szBuffer),
            "tellopponent My version is %s, finger %s for more details", MYVERSION, MYNAME);
    cmdresp_snd(szBuffer);
  }
#endif
}

static void cmd_nps(const char * szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

static void cmd_rating(UNUSED const char * szCommand, const char *szArg) {
  char szBuffer[MAX_DATANAME_LENGTH];

  snprintf(szBuffer, sizeof (szBuffer), "got rating %s", szArg);
  cmdresp_snd_info(szBuffer);
}

/**
 * return position evaluation
 */
static void cmd_eval(UNUSED const char *szCommand, UNUSED const char *szArg) {
  t_score fResult;
  t_boolean isWhite = (t_boolean) (!(game.curply & 1));

  fResult = eval_pos(isWhite);
  if (isWhite) {
    cmdresp_snd_eval(fResult, "Position evaluation (white playing)");
  } else {
    cmdresp_snd_eval(fResult, "Position evaluation (black playing)");
  }

  fResult = eval_material(isWhite);
  cmdresp_snd_eval(fResult, "Material evaluation");

  fResult = eval_kingpos(isWhite);
  cmdresp_snd_eval(fResult, "Kingposition evaluation");

  fResult = eval_mobility(isWhite);
  cmdresp_snd_eval(fResult, "Mobility evaluation");

  fResult = eval_staticpos(isWhite);
  cmdresp_snd_eval(fResult, "Board static evaluation");

  fResult = eval_dynamicpos(isWhite);
  cmdresp_snd_eval(fResult, "Board dynamic evaluation");
}

/**
 * TODO: set flags for all accepted features
 */
static void cmd_accepted(UNUSED const char *szCommand, UNUSED const char *szArg) {
}

/**
 * TODO: handle the fact that there is an incompatibility
 */
static void cmd_rejected(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmdresp_not_supported(szCommand);
}

/**
 * enter move in algebraic notation
 */
static void cmd_usermove(UNUSED const char *szCommand, const char *szArg) {
  if (board_applymove(szArg) == MOVERESULT_ERROR)
    cmdresp_snd_errormove(szArg);
  StartEngineThread();
}

/**
 * play both colours
 */
static void cmd_both(UNUSED const char *szCommand, UNUSED const char *szArg) {
  cmd_force("", "");
  uistate.computer_engines[0] = myTRUE;
  uistate.computer_engines[1] = myTRUE;
  StartEngineThread();
}

/**
 * select the engine to be used
 */
static void cmd_alg(UNUSED const char *szCommand, const char *szArg) {
  search_destructor();
  sscanf(szArg, "%d", &level.search_algoritme);
  search_constructor();
}

/**
 * execute
 */
static void cmd_execute(const char *szCommand, const char *szArg) {
  char szBuffer[MAX_DATANAME_LENGTH];
  if ((szCommand[0] == '@') && (strlen(szArg) == 0)) {
    strncpy(szBuffer, szCommand + 1, sizeof (szBuffer));
    usercmd_execute("@", szBuffer);
  } else {
    usercmd_execute(szCommand, szArg);
  }
}

#if defined(FEATURE_XBOARD_INTERFACE)
/** enter move in coordinate notation
static void cmd_coordmove(const char * szCommand, const char *szArg) {
  // as of XBoard 4.2.7 force the use of usermode
  if (uimode.protover == 2) {
    cmdresp_not_supported(szCommand);
  } else {
    if (board_applymove(szArg) == MOVERESULT_ERROR)
      cmdresp_snd_errormove(szCommand);
  }
}
 */
#endif

// --------------------------------------------------------------------

static void cmdresp_not_supported(const char *szCommand) {
  char szBuffer[MAX_DATANAME_LENGTH];
  if (!uimode.xboard_mode) {
    cmdresp_snd_error(szCommand, "unknown command");
  } else {
    snprintf(szBuffer, sizeof (szBuffer), "unknown command %s", szCommand);
    stat_logadd(szBuffer, 2);
  }
}

/**
 * reply to xboard, main and only routine for this purpose
 */
void cmdresp_snd(const char *szCommand) {
  // make sure not to add the newline twice
  if ((strlen(szCommand) > 0) && (szCommand[strlen(szCommand) - 1] == '\n')) {
    puts(szCommand);
  } else {
    printf("%s\n", szCommand);
  }
  if (uimode.fpDebugOutput != NULL) {
    fprintf(uimode.fpDebugOutput, "%s\n", szCommand);
    fflush(uimode.fpDebugOutput);
  }
  stat_logadd(szCommand, 1);
}

void cmdresp_snd_prompt(void) {
  if (search.engine_searching == myTRUE) {
    if (uistate.showprompt == myFALSE) {
      uistate.showprompt = myTRUE;
    }
  } else {
    if (uistate.showprompt == myTRUE) {
      uistate.showprompt = myFALSE;
      if (!uimode.engine_mode) {
        fputs(MYNAME "> ", stdout);
      }
    }
  }
}

/**
 * send the move generated by the engine to xboard
 * TODO: if belofte has received accepted san, output is sent in SAN format
 */
void cmdresp_snd_move(const char *szMove, const int iSearchResult, const char *szResult) {
  int iResult;
  char szBuffer[MAX_DATANAME_LENGTH];

  iResult = board_applymove(szMove);

  if (iSearchResult >= SEARCH_FLAG) {
    if (uimode.uci_mode == myTRUE) {
      snprintf(szBuffer, sizeof (szBuffer), "bestmove %s", szMove);
    } else {
      snprintf(szBuffer, sizeof (szBuffer), "move %s", szMove);
    }
    cmdresp_snd(szBuffer);
  }

  if (iResult != MOVERESULT_OK) {
    if (iResult == MOVERESULT_DRAW_LACKMAT) {
      // one move, draw by lack of material
      cmdresp_snd("1/2-1/2 {Draw by lack of material}");
      strncpy(game.szResult, "1/2-1/2", sizeof (game.szResult));
    } else if (iResult == MOVERESULT_DRAW_THREEFOLD) {
      // one move, draw by threefold repetition
      cmdresp_snd("1/2-1/2 {Draw by threefold repetition}");
      strncpy(game.szResult, "1/2-1/2", sizeof (game.szResult));
    } else if (iResult == MOVERESULT_DRAW_50) {
      // one move, draw by 50 move rule
      cmdresp_snd("1/2-1/2 {Draw by 50 move rule}");
      strncpy(game.szResult, "1/2-1/2", sizeof (game.szResult));
    }
    // make sure that after end of game, searching ends
    cmd_force("", "");
    cmd_save("", "");
  }

  if (iSearchResult == SEARCH_FLAG) {
    if (strcmp(szResult, SEARCHRESULT_STALEMATE) == 0) {
      // one move, draw by stalemate
      cmdresp_snd("1/2-1/2 {Stalemate}");
      strncpy(game.szResult, "1/2-1/2", sizeof (game.szResult));
    } else if (strcmp(szResult, SEARCHRESULT_BLACKMATES) == 0) {
      // one move, black wins
      cmdresp_snd("0-1 {Black mates}");
      strncpy(game.szResult, "0-1", sizeof (game.szResult));
    } else if (strcmp(szResult, SEARCHRESULT_WHITEMATES) == 0) {
      // one move, white wins
      cmdresp_snd("1-0 {White mates}");
      strncpy(game.szResult, "1-0", sizeof (game.szResult));
    } else if (strcmp(szResult, SEARCHRESULT_BLACKRESIGNS) == 0) {
      // one move, black resigns
      cmdresp_snd("1-0 {Black resigns}");
      strncpy(game.szResult, "1-0", sizeof (game.szResult));
    } else if (strcmp(szResult, SEARCHRESULT_WHITERESIGNS) == 0) {
      // one move, white resigns
      cmdresp_snd("0-1 {White resigns}");
      strncpy(game.szResult, "0-1", sizeof (game.szResult));
    }
    // make sure that after end of game, searching ends
    cmd_force("", "");
    cmd_save("", "");
  }
}

/**
 * return illegal move back to xboard
 */
static void cmdresp_snd_errormove(const char *szMove) {
  char szBuffer[MAX_DATANAME_LENGTH];
  snprintf(szBuffer, sizeof (szBuffer), "Illegal move: %s", szMove);
  cmdresp_snd(szBuffer);
}

/**
 * return an error back to xboard
 */
static void cmdresp_snd_error(const char *szCommand, const char *szReason) {
  char szBuffer[MAX_DATANAME_LENGTH];
  snprintf(szBuffer, sizeof (szBuffer), "Error (%s): %s", szReason, szCommand);
  cmdresp_snd(szBuffer);
}

/**
 * return an info back to xboard
 */
void cmdresp_snd_info(const char *szMessage) {
  char szBuffer[MAX_DATANAME_LENGTH];

  snprintf(szBuffer, sizeof (szBuffer), " %s", szMessage);

#if defined(FEATURE_UCI_INTERFACE)
  if (uimode.post_mode && (uimode.uci_mode == myFALSE)) {
#else
  if (uimode.post_mode) {
#endif
    cmdresp_snd(szBuffer);
  } else {
    stat_logadd(szBuffer, 2);
  }
}

/**
 * return thinking statistics back
 */
void cmdresp_snd_thinking(const int nPly, const char cPlyFlag, const t_score fEval,
        const char *szThinking) {
  char szBuffer[MAX_DATANAME_LENGTH];

#if defined(FEATURE_UCI_INTERFACE)
  if (uimode.uci_mode == myTRUE) {
    // uci mode has different output
    if (nPly != 0) {
      snprintf(szBuffer, sizeof (szBuffer),
              "info depth %d seldepth %d", nPly, uistate.iCurrentDepth);
      cmdresp_snd(szBuffer);
      snprintf(szBuffer, sizeof (szBuffer), "info time %d score cp %d nodes %d pv %s",
              (int) uistate.thinkingtime,
              (int) (fEval * 100.0),
              uistate.searchnodes,
              szThinking);
    } else {
      snprintf(szBuffer, sizeof (szBuffer), "info string %s",
              szThinking);
    }
    cmdresp_snd(szBuffer);
  } else {
#endif
    snprintf(szBuffer, sizeof (szBuffer), " %2d%c %4d %4d %7d %s",
            nPly, cPlyFlag,
            (int) (fEval * 100.0),
            (int) uistate.thinkingtime,
            uistate.searchnodes,
            szThinking);
    if (uimode.post_mode) {
      cmdresp_snd(szBuffer);
    } else {
      stat_logadd(szBuffer, 2);
    }
#if defined(FEATURE_UCI_INTERFACE)
  }
#endif
}

/**
 * return evalutation statistics back
 */
void cmdresp_snd_eval(const t_score fEval, const char *szThinking) {
  char szBuffer[MAX_DATANAME_LENGTH];

  snprintf(szBuffer, sizeof (szBuffer), " %2.3f %s",
          fEval,
          szThinking);

  if (uimode.post_mode) {
    cmdresp_snd(szBuffer);
  } else {
    stat_logadd(szBuffer, 2);
  }
}

#if defined(FEATURE_XBOARD_INTERFACE)

/**
 *
 */
static void cmdresp_snd_protover(UNUSED const int nProtover) {
  cmdresp_snd("feature done=0");

  // TODO: replace by code that will interprete the command structure and disable
  // features dynamically
  cmdresp_snd("feature ping=1 setboard=1 playother=1 san=0 usermove=1 time=1");
  cmdresp_snd("feature draw=0 sigint=1 sigterm=1 reuse=0 debug=1");
#if defined(FEATURE_XBOARD_ICS)
  cmdresp_snd("feature ics=1");
#else
  cmdresp_snd("feature ics=0");
#endif
  cmdresp_snd("feature name=1 pause=0 analyze=0 colors=0 nps=0");
  cmdresp_snd("feature myname=\"" MYNAME " " MYVERSION "\"");
  cmdresp_snd("feature variants=\"normal\"");

  // time intensive constructor
  usercmd_delayedconstructor();

  cmdresp_snd("feature done=1");
}
#endif

/**
 * list of st_command pointers
 * interfacelevel is  1: winboard 1 protocol
 *                    2: winboard 2 protocol
 *                    4: belofte extension on winboard protocol
 *                    8: uci protocol
 */
static command_t list_commands[] = {
  { "help", cmd_help, "show this list of commands", 15, CMD_PARAM_STRING | CMD_OPT},
  { "?", cmd_movenow, "force the computer to move now", 7, CMD_NOPARAM},
  { "about", cmd_about, "tell something about myself", 7, CMD_NOPARAM},
  { "accepted", cmd_accepted, "a requested feature is implemented", 6, CMD_PARAM_STRING},
  { "again", cmd_again, "research again from this position", 4, CMD_NOPARAM},
  { "alg", cmd_alg, "select algorithm", 4, CMD_PARAM_NUMBER},
#if defined(FEATURE_XBOARD_INTERFACE)
  { "analyze", cmd_analyze, "enter analyze mode", 1, CMD_PARAM_STRING | CMD_OPT | CMD_NOTSUPPORTED},
#endif
  { "bd", cmd_bd, "display board", 15, CMD_PARAM_STRING | CMD_OPT},
#if defined(FEATURE_USEBOOK)
  { "bk", cmd_bk, "book", 6, CMD_PARAM_NUMBER | CMD_OPT},
#endif
  { "black", cmd_black, "set player to play black", 1, CMD_NOPARAM},
  { "both", cmd_both, "set computer to play both", 7, CMD_NOPARAM},
  { "computer", cmd_computer, "playing against a computer", 6, CMD_NOPARAM},
#if defined(FEATURE_USEBOOK)
  { "createbook", cmd_createbook, "create the opening book from PGN", 4, CMD_PARAM_STRING},
#endif
  { "debug", cmd_debug, "enter debug mode [fenhistory|hashhistory|on|off|filename.ext]",
    14, CMD_PARAM_STRING | CMD_OPT},
#if defined(FEATURE_XBOARD_INTERFACE)
  { "draw", cmd_draw, "offer draw", 1, CMD_NOPARAM | CMD_NOTSUPPORTED},
#endif
  { "easy", cmd_easy, "do not think when not on move", 7, CMD_NOPARAM | CMD_NOTSUPPORTED},
  { "edit", cmd_edit, "go to edit mode", 1, CMD_PARAM_STRING | CMD_OPT | CMD_NOTSUPPORTED},
  { "epd", cmd_epd, "load and evaluate an file in EPD format from disk", 4, CMD_PARAM_STRING},
  { "epdpos", cmd_epdpos, "evaluate epd position", 4, CMD_PARAM_STRING},
  { "eval", cmd_eval, "give evaluation of position", 4, CMD_NOPARAM},
  { "exec", cmd_execute, "execute command file", 7, CMD_PARAM_STRING},
  { "exit", cmd_exit, "leave analysis mode", 1, CMD_NOPARAM | CMD_NOTSUPPORTED},
  { "fd", cmd_fd, "set fixed search depth mode", 7, CMD_PARAM_NUMBER},
  { "force", cmd_force, "computer plays neither side", 7, CMD_NOPARAM},
  { "game", cmd_game, "print current game", 4, CMD_NOPARAM},
  { "go", cmd_go, "computer starts playing", 15, CMD_PARAM_STRING | CMD_OPT},
  { "hard", cmd_hard, "computer thinks all the time", 7, CMD_NOPARAM | CMD_NOTSUPPORTED},
  { "hint", cmd_hint, "computer gives a hint", 1, CMD_NOPARAM | CMD_NOTSUPPORTED},
#if defined(FEATURE_XBOARD_ICS)
  { "ics", cmd_ics, "set hostname if playing on ics", 6, CMD_PARAM_STRING},
#endif
#if defined(FEATURE_UCI_INTERFACE)
  { "isready", cmd_isready, "let engine finalize initialisation", 8, CMD_PARAM_STRING},
#endif
  { "info", cmd_info, "get more information", 7, CMD_PARAM_STRING},
  { "level", cmd_level, "select level to play at", 7, CMD_PARAM_NUMBER},
  { "load", cmd_load, "load game in pgn form from disk", 4, CMD_PARAM_STRING | CMD_NOTSUPPORTED},
#if defined(FEATURE_USEBOOK)
  { "loadbook", cmd_loadbook, "load book from disk", 4, CMD_PARAM_STRING},
#endif
  { "ls", cmd_ls, "list directory content", 4, CMD_PARAM_STRING | CMD_OPT},
  { "name", cmd_name, "give your name", 7, CMD_PARAM_STRING},
  { "new", cmd_new, "new game", 7, CMD_NOPARAM},
  { "nopost", cmd_nopost, "stop outputting search details", 7, CMD_NOPARAM},
  { "nps", cmd_nps, "nodes per second", 6, CMD_PARAM_NUMBER | CMD_NOTSUPPORTED},
  { "otim", cmd_otim, "set opponent time", 7, CMD_PARAM_NUMBER},
#if defined(FEATURE_XBOARD_INTERFACE)
  { "pause", cmd_pause, "suspend", 2, CMD_NOPARAM | CMD_NOTSUPPORTED},
#endif
  { "perft", cmd_perft, "do a performance test", 7, CMD_PARAM_NUMBER},
  { "ping", cmd_ping, "check if still alive", 6, CMD_NOPARAM},
  { "playother", cmd_playother, "set computer to play the other colour", 2, CMD_NOPARAM},
#if defined(FEATURE_UCI_INTERFACE)
  { "position", cmd_position, "set position [fen <fenstring>|startpos] [moves <move*>]",
    8, CMD_PARAM_STRING | CMD_OPT},
#endif
  { "post", cmd_post, "output search details", 7, CMD_PARAM_SWITCH},
#if defined(FEATURE_XBOARD_INTERFACE)
  { "protover", cmd_protover, "set command interface level", 7, CMD_PARAM_NUMBER},
#endif
  { "quit", cmd_quit, "leave chess program", 15, CMD_NOPARAM},
  { "random", cmd_random, "causes the computer to play randomly", 7, CMD_PARAM_SWITCH | CMD_OPT},
  { "rating", cmd_rating, "set computer and player rating", 1, CMD_PARAM_NOCHECK | CMD_NOTSUPPORTED},
  { "rejected", cmd_rejected, "a feature is not supported", 6, CMD_PARAM_STRING | CMD_NOTSUPPORTED},
  { "remove", cmd_remove, "undo last 2 moves", 7, CMD_PARAM_NUMBER | CMD_OPT},
  { "result", cmd_result, "set result string", 7, CMD_PARAM_STRING},
#if defined(FEATURE_XBOARD_INTERFACE)
  { "resume", cmd_resume, "unpause", 6, CMD_PARAM_STRING | CMD_OPT | CMD_NOTSUPPORTED},
#endif
  { "save", cmd_save, "save game in PGN format", 4, CMD_NOPARAM},
  { "sd", cmd_sd, "set maximum depth for search", 7, CMD_PARAM_NUMBER},
  { "st", cmd_st, "set level to time per move", 7, CMD_PARAM_NUMBER},
#if defined(FEATURE_UCI_INTERFACE)
  { "stop", cmd_movenow, "stop thinking", 8, CMD_NOPARAM},
#endif
  { "time", cmd_time, "set own clock", 7, CMD_PARAM_NUMBER},
  // TODO: Check if setboard does not have to be present in XBoard 1 protocol
  { "setboard", cmd_setboard, "setup new position", 6, CMD_PARAM_STRING},
#if defined(FEATURE_UCI_INTERFACE)
  { "uci", cmd_uci, "enter uci mode", 7, CMD_NOPARAM},
  { "ucinewgame", cmd_ucinewgame, "prepare for new game", 8, CMD_NOPARAM},
#endif
  { "undo", cmd_undo, "undo move", 7, CMD_NOPARAM},
  { "usermove", cmd_usermove, "enter moves in algebraic notation", 6, CMD_PARAM_STRING},
#if defined(FEATURE_CHESS_VARIANT)
  { "variant", cmd_variant, "set chess variant", 7, CMD_PARAM_STRING | CMD_NOTSUPPORTED},
#endif
  { "white", cmd_white, "set computer to play white", 1, CMD_NOPARAM},
#if defined(FEATURE_XBOARD_INTERFACE)
  { "xboard", cmd_xboard, "switch to xboard mode", 7, CMD_PARAM_SWITCH | CMD_OPT},
#endif
  { NULL, NULL, NULL, 0, CMD_NOTSUPPORTED}
};

/**
 * show help on screen, information shown depends on protocol level
 */
static void cmd_help(UNUSED const char *szCommand, const char *szArg) {
  command_t *c;
  char szBuffer[MAX_DATANAME_LENGTH];
  t_boolean cFound = myFALSE;

  if (strlen(szArg)) {
    // request more details on specific command
    c = list_commands;
    do {
      if ((c->cmd_name != NULL)&&(!strcmp(szArg, c->cmd_name))) {
        // command found
        printf("%s\t%s\n", c->cmd_name ? c->cmd_name : "", c->cmd_help);
        cFound = myTRUE;
        if (c->cmd_paramdata == CMD_NOPARAM)
          printf("\tthe function does not take any parameters\n");
        else if (c->cmd_paramdata & CMD_OPT)
          printf("\tthe function does take optional parameters\n");
        if (c->cmd_paramdata & CMD_PARAM_STRING)
          printf("\t%s <string>", c->cmd_name ? c->cmd_name : "");
        else if (c->cmd_paramdata & CMD_PARAM_NUMBER)
          printf("\t%s <n>", c->cmd_name ? c->cmd_name : "");
        else if (c->cmd_paramdata & CMD_PARAM_SWITCH)
          printf("\t%s <on|off> (default is on)", c->cmd_name ? c->cmd_name : "");
        if (c->cmd_paramdata & CMD_NOTSUPPORTED)
          printf("\n\tUnsupported!\n");
        if (!(c->cmd_interfacelevel & uimode.protover))
          printf("\n\tNot available in this protover!\n");
      }
    } while ((c++)->cmd_name != NULL);
    if (cFound == myFALSE) {
      snprintf(szBuffer, sizeof (szBuffer), "Command <%s> not found!", szArg);
      cmdresp_snd(szBuffer);
    }
  } else {
    // print all commands
#if defined(FEATURE_UCI_INTERFACE)
    snprintf(szBuffer, sizeof (szBuffer),
            "Available UCI commands are:");
#else
    snprintf(szBuffer, sizeof (szBuffer),
            "Available protover %d commands are: (* unimplemented)",
            uimode.protover);
#endif
    cmdresp_snd(szBuffer);
    c = list_commands;
    do {
      if (c->cmd_interfacelevel & uimode.protover) {
        // is this command supported in this interface mode ?
        printf("%-10s - %s %s\n", c->cmd_name ? c->cmd_name : "", c->cmd_help,
                c->cmd_paramdata & CMD_NOTSUPPORTED ? "*" : "");
      }
    } while ((c++)->cmd_name != NULL);
    cmdresp_snd("\nType help <command> for more information.");
#if defined(FEATURE_XBOARD_INTERFACE)
    if (uimode.protover != 4)
      cmdresp_snd("Type protover <n> to change mode.");
#endif
  }
  cmdresp_snd("");
}

/**
 * execute all commands in file, lines starting with # are ignored
 */
int usercmd_execute(const char *szCommand, const char *szArg) {
  char szArgFile[MAX_DATANAME_LENGTH];
  char szCommandFile[MAX_DATANAME_LENGTH];
  char szLog[MAX_DATANAME_LENGTH];
  FILE *fpRes;
  int iCmd;

  fpRes = fopen(szArg, "rt");
  if (fpRes != NULL) {
    cmd_force("", "");
    snprintf(szLog, sizeof (szLog), "Reading file %s.", szArg);
    stat_logadd(szLog, 2);
    while (!feof(fpRes)) {
      file_readline(szCommandFile, fpRes, sizeof (szCommandFile));
      usercmd_readarg(szCommandFile, szArgFile, sizeof (szArgFile));

      if (strlen(szCommandFile) == 0) break;
      snprintf(szLog, sizeof (szLog), "%s %s", szCommandFile, szArgFile);
      cmdresp_snd_info(szLog);

      if (szCommandFile[0] != '#') {
        for (iCmd = 0; list_commands[iCmd].cmd_name != NULL; iCmd++) {
          // compare to all except last one
          if (0 == strcmp(list_commands[iCmd].cmd_name, szCommandFile)) {
            break;
          }
        }
        if (list_commands[iCmd].cmd_ptr != NULL)
          list_commands[iCmd].cmd_ptr(szCommandFile, szArgFile);
      }
    }
    fclose(fpRes);
    snprintf(szLog, sizeof (szLog), "Finished reading %s", szArg);
    if (strcmp(szCommand, "-i") != 0) {
      cmdresp_snd_info(szLog);
    } else {
      stat_logadd(szLog, 2);
    }
  } else {
    if (strcmp(szCommand, "-i") != 0) {
      snprintf(szLog, sizeof (szLog), "Unable to load file %s.", szArg);
      cmdresp_snd_info(szLog);
    }
    return 0;
  }
  return 1;
}

/**
 * read input command from xboard
 * return codes not used now
 */
int usercmd_readcommand(void) {
  char szArg[MAX_DATANAME_LENGTH];
  char szCommand[MAX_DATANAME_LENGTH];
  char szLog[MAX_DATANAME_LENGTH];

  int iCmd;

  for (;;) {
    // fflush(stdout);  // not required since there is a setbuf
    cmdresp_snd_prompt();
    file_readline(szCommand, stdin, sizeof (szCommand));
    usercmd_readarg(szCommand, szArg, sizeof (szArg));

    snprintf(szLog, sizeof (szLog), "%s %s", szCommand, szArg);
    stat_logadd(szLog, 0);

    for (iCmd = 0; list_commands[iCmd].cmd_name != NULL; iCmd++) {
      // compare to all except last one
      if (0 == strcmp(list_commands[iCmd].cmd_name, szCommand)) {
        if (list_commands[iCmd].cmd_interfacelevel & uimode.protover) {
          break;
        } else {
          snprintf(szLog, sizeof (szLog),
                  "Command not supported in protover %d!", uimode.protover);
          cmdresp_snd(szLog);
        }
      }
    }
    if (list_commands[iCmd].cmd_ptr) {
      list_commands[iCmd].cmd_ptr(szCommand, szArg);
    } else {
      if (szCommand[0] == '@') {
        // execute command special handling
        cmd_execute(szCommand, szArg);
      } else if (strlen(szCommand) != 0) {
        // TODO: bounds checking of input
        snprintf(szLog, sizeof (szLog), "Command [%s] not supported!", szCommand);
        cmdresp_snd(szLog);
      }
    }
    uistate.showprompt = myTRUE;

    break;
  }

  return 0;
}

/**
 * split arguments on line
 */
static int usercmd_readarg(char * szLine, char * szArg, const size_t nSizeArg) {
  int i;
  int nSizeLine = (int) strlen(szLine);

  if (nSizeArg < 1) return 0;

  szArg[0] = '\0';

  for (i = 0; i < nSizeLine; i++) {
    if (szLine[i] == ' ') {
      strncpy(szArg, szLine + i + 1, nSizeArg);
      szArg[nSizeArg - 1] = '\0';
      szLine[i] = '\0';
      // TODO: left trim arguments so there are no leading spaces
      break;
    }
  }
  return 1;
}

// eof
